package PureMessage::Message::ComputedFields;

use strict;

package PureMessage::Message;

use ActiveState::I18N::UTF8 qw(utf8_encode utf8_header utf8_length);
use PureMessage::MIME::Header qw(header_utf8);
use PureMessage::UTF8::Config qw(%utf8_config);

sub c_content_type {
    my $self = shift;
    return $self->{cache_content_type} ||= do {
	my $ct = $self->header("Content-Type");
	if ($ct) {
	    $ct =~ s/;.*//s;
	    $ct =~ s/\s+//g;
	    $ct = lc($ct);
	}
	else {
	    $ct = "text/plain";
	}
	$ct;
    };
}

sub c_charset {
    my $self = shift;
    return $self->{cache_charset} if exists $self->{cache_charset};
    return $self->{cache_charset} = do {
	my $charset;
	my $ct = $self->header("Content-Type");
	if ($ct) {
	    $ct = lc($ct);
	    if ($ct =~ /^\s*text\s*\//) {
		if ($ct =~ /;\s*charset\s*=\s*([^;]*)/) {
		    $charset = $1;
		    $charset =~ s/\"//g;
		    $charset =~ s/\'//g;
		}
		else {
		    $charset = "us-ascii";
		}
	    }
	}
	else {
	    # the default content-type is "text/plain; charset=us-ascii"
	    $charset = "us-ascii";
	}
	
	$charset;
    };    
}

sub c_date_isoz {
    my $self = shift;
    return $self->{cache_date_isoz} if exists $self->{cache_date_isoz};
    return $self->{cache_date_isoz} = do {
	my $d = $self->header("Date");
	if ($d) {
	    $d =~ s/\([^)]*\)/ /g;
	    $d = HTTP::Date::str2time($d);
	    $d = HTTP::Date::time2isoz($d) if $d;
	}
	$d;
    }
}

sub c_date_time {
    my $self = shift;
    return $self->{cache_date_time} if exists $self->{cache_date_time};
    return $self->{cache_date_time} = do {
	my $d = $self->header("Date");
	if ($d) {
	    $d =~ s/\([^)]*\)/ /g;
	    $d = HTTP::Date::str2time($d);
	}
	$d;
    }
}

sub c_hops {
    my $self = shift;
    return $self->{cache_hops} ||= do {
	my @r = $self->header("Received");
	scalar(@r);
    };
}

sub c_subj15 {
    my $self = shift;
    return $self->{cache_subj15} if exists $self->{cache_subj15};
    return $self->{cache_subj15} = do {
	my $subj15 = $self->header("Subject");
	if (defined $subj15) {
	    $subj15 = lc($subj15);
	    $subj15 =~ s/^[\d\s]+//;
	    $subj15 =~ s/[\d\s]+$//;
	    $subj15 =~ s/\s+/ /g;
	    $subj15 = substr($subj15, 0, 15);
	}
    	$subj15;
    }
}

## renamed from c_subject_utf8

sub c_subject {
    my $self = shift;
    return $self->{cache_subject_utf8} if exists $self->{cache_subject_utf8};
    return $self->{cache_subject_utf8} = do {
	my $subject = $self->header("Subject");
	if (defined $subject) {
	    $subject = header_utf8($subject);
	    # trim space
	    $subject =~ s/^\s+//;
	    $subject =~ s/\s+$//;
	    $subject =~ s/\s+/ /g;

	    # NUL values confuse databases like Postgres
	    $subject =~ s/\0//g;
	}
	$subject;
    };
}

sub c_subject_utf8 {
    my $self = shift;
    return $self->{cache_subject_utf8}
	if exists $self->{cache_subject_utf8};

    my $subject = $self->header('Subject');
    return unless defined $subject;

    utf8_header($subject, \%utf8_config);

    $subject =~ s/^\s+//;
    $subject =~ s/\s+$//;
    $subject =~ s/\s+/ /g;
    $self->{cache_subject_utf8} =
	substr($subject, 0, utf8_length($subject, 80));
}

sub c_subject_ascii {
    my $self = shift;
    return $self->{cache_subject_ascii} if exists $self->{cache_subject_ascii};
    return $self->{cache_subject_ascii} = do {
	my $subject = $self->c_subject_utf8;
	if (defined $subject) {
	    $subject =~ s/([^\x20-\x7e]+)/_ascii_repl($1)/ge;
	}
	$subject;
    };
}

sub _ascii_repl {
    my $orig = shift;
    return "*" x length($orig) if length($orig) <= 2;
    return "***";
}

# XXX There appears to be a bug in perl that causes the nested closures
# to not be deallocated properly.  c_subject_utf8_helper() has been made
# into a real sub to avoid this bug.  GSAR will investigate the perl bug
# to understand the ramifications (see bug 30541).

my $text;
my $len;

sub c_subject_utf8_helper {
    my($self, undef, $ct, $ct_param) = @_;
    if ($ct eq "text/html") {
        $text .= " << HTML: ";
        require HTML::Parser;
        my $p = HTML::Parser->new(api_version => 3);
        $p->ignore_elements(qw(script style));
        $p->handler(text => sub {
            $text .= shift;
            $text =~ s/\s+/ /g;
            die "DONE" if length($text) > $len;
        }, "dtext");
        $p->handler(start => sub {
            $text .= " " unless $text =~ / \z/;
        });

        $self->decode_te(1);
        $self->handler(data => sub { $p->parse($_[1]) });
        $self->handler(part_end => sub { $p->eof; die "DONE" });
    }
    elsif ($ct =~ /^text\//) {
        $text .= " << ";
        $self->decode_te(1);
        $self->handler(data => sub {
            $text .= $_[1];
            $text =~ s/\s+/ /g;
            die "DONE" if length($text) > $len;
        });
        $self->handler(part_end => sub { die "DONE" });
    }
    elsif ($ct !~ /^multipart\//) {
        $text .= " << $ct";
        die "DONE";
    }
}

## renamed from c_subject_body_utf8

sub c_subject_body {
    my $self = shift;
    $len = shift || 80;
    my $cache_key = "cache_subject_body_utf8_$len";
    return $self->{$cache_key} if exists $self->{$cache_key};
    return $self->{$cache_key} = do {
	$text = $self->c_subject_utf8;
	$text = "" unless defined $text;

	my $ct = $self->c_content_type;
	if ($ct =~ /^(?:text|multipart)\//) {

	    my $old_pos = $self->tell;
	    $self->seek(0, 0);

	    require PureMessage::MIME::Parser;
	    my $p = PureMessage::MIME::Parser->new;

            $p->handler(part_start => \&c_subject_utf8_helper);

	    # start feeding data to the parser.  This is wrapped
	    # in and eval block so that the callbacks can abort
	    # the parsing when they have seen enough.
	    eval {
		my $buf;
		while ($self->read($buf, 512)) {
		    $p->parse($buf);
		}
		$p->eof;
	    };
	    warn $@ if $@ && $@ !~ /^DONE/;

#	    if (my $charset = $ct_param->{charset}) {
#		# XXX this might ensocde the subject part twice
#		$text = _encode_utf8($charset, $text);
#	    }

	    # restore old position
	    $self->seek($old_pos, 0);
	}
	else {
	    $text .= " << $ct";
	}
	substr($text, $len) = "" if length($text) > $len;
	$text;
    };
}

sub c_subject_body_utf8 {
    my $self = shift;

    $len = shift || 80;
    my $cache_key = "cache_subject_body_utf8_$len";
    return $self->{$cache_key} if exists $self->{$cache_key};

    $self->{$cache_key} = $self->c_subject_utf8;
    return $self->{$cache_key} if length($self->{$cache_key}) >= $len;

    $self->{$cache_key} .= do {
	$text = '';

	my $ct = $self->c_content_type;
	if($ct =~ /^(?:text|multipart)\//) {
	    my $old_pos = $self->tell;
	    $self->seek(0, 0);

	    require PureMessage::MIME::Parser;
	    my $p = PureMessage::MIME::Parser->new;

	    $p->handler(part_start => \&c_subject_utf8_helper);

	    eval {
		my $buf;
		while($self->read($buf, 512)) {
		    $p->parse($buf);
		}
		$p->eof;
	    };
	    warn $@ if $@ && ($@ !~ /^DONE/);

	    utf8_encode($text, $self->c_charset, \%utf8_config);

	    $self->seek($old_pos, 0);
	}
	else {
	    $text .= " << $ct";
	}
	$text;
    };
    $self->{$cache_key} =
	substr($self->{$cache_key}, 0, utf8_length($self->{$cache_key}, $len));
}

package PureMessage::AnnotatedMessage;

require HTTP::Date;

sub c_qdate_iso {
    my $self = shift;
    return $self->{cache_qdate_iso} if exists $self->{cache_qdate_iso};
    return $self->{cache_qdate_iso} = do {
	my $d = $self->minfo_value("Date");
	if ($d) {
	    $d = HTTP::Date::str2time($d);
	    $d = HTTP::Date::time2iso($d) if $d;
	}
	$d;
    }
}

sub c_qdate_time {
    my $self = shift;
    return $self->{cache_date_time} if exists $self->{cache_date_time};
    return $self->{cache_date_time} = do {
	my $d = $self->minfo_value("Date");
	if ($d) {
	    $d = HTTP::Date::str2time($d);
	}
	$d;
    }
}

sub c_idate_onlyz {
    my $self = shift;
    return $self->{cache_date_onlyz} if exists $self->{cache_date_onlyz};
    return $self->{cache_date_onlyz} = do {
	my $d = $self->minfo_value("Date");
	if ($d) {
	    $d =~ s/\([^)]*\)/ /g;
	    $d = HTTP::Date::str2time($d);
	    $d = HTTP::Date::time2isoz($d) if $d;
	    $d = sprintf("%4d%02d%02d0000", 
			 (HTTP::Date::parse_date($d))[0,1,2]) if $d;
	}
	$d;
    }
}

sub c_idate_only {
    my $self = shift;
    return $self->{cache_date_only} if exists $self->{cache_date_only};
    return $self->{cache_date_only} = do {
	my $d = $self->minfo_value("Date");
	if ($d) {
	    $d =~ s/\([^)]*\)/ /g;
	    $d = HTTP::Date::str2time($d);
	    $d = HTTP::Date::time2iso($d) if $d;
	    $d = sprintf("%4d%02d%02d0000", 
			 (HTTP::Date::parse_date($d))[0,1,2]) if $d;
	}
	$d;
    }
}

sub c_idate_hourz {
    my $self = shift;
    return $self->{cache_date_hourz} if exists $self->{cache_date_hourz};
    return $self->{cache_date_hourz} = do {
	my $d = $self->minfo_value("Date");
	if ($d) {
	    $d =~ s/\([^)]*\)/ /g;
	    $d = HTTP::Date::str2time($d);
	    $d = HTTP::Date::time2isoz($d) if $d;
	    $d = sprintf("%4d%02d%02d%02d00", 
			 (HTTP::Date::parse_date($d))[0,1,2,3]) if $d;
	}
	$d;
    }
}

sub c_idate_hour {
    my $self = shift;
    return $self->{cache_date_hour} if exists $self->{cache_date_hour};
    return $self->{cache_date_hour} = do {
	my $d = $self->minfo_value("Date");
	if ($d) {
	    $d =~ s/\([^)]*\)/ /g;
	    $d = HTTP::Date::str2time($d);
	    $d = HTTP::Date::time2iso($d) if $d;
	    $d = sprintf("%4d%02d%02d%02d00", 
			 (HTTP::Date::parse_date($d))[0,1,2,3]) if $d;
	}
	$d;
    }
}

sub c_idate_10minz {
    my $self = shift;
    return $self->{cache_date_10minz} if exists $self->{cache_date_10minz};
    return $self->{cache_date_10minz} = do {
	my $d = $self->minfo_value("Date");
	if ($d) {
	    $d =~ s/\([^)]*\)/ /g;
	    $d = HTTP::Date::str2time($d);
	    $d = HTTP::Date::time2isoz($d) if $d;
	    $d = sprintf("%4d%02d%02d%02d%02d", 
			 (HTTP::Date::parse_date($d))[0,1,2,3,4]) if $d;
	    $d -= $d%10 if $d;
	}
	$d;
    }
}

sub c_idate_10min {
    my $self = shift;
    return $self->{cache_date_10min} if exists $self->{cache_date_10min};
    return $self->{cache_date_10min} = do {
	my $d = $self->minfo_value("Date");
	if ($d) {
	    $d =~ s/\([^)]*\)/ /g;
	    $d = HTTP::Date::str2time($d);
	    $d = HTTP::Date::time2iso($d) if $d;
	    $d = sprintf("%4d%02d%02d%02d%02d", 
			 (HTTP::Date::parse_date($d))[0,1,2,3,4]) if $d;
	    $d -= $d%10 if $d;
	}
	$d;
    }
}

sub c_idate_minz {
    my $self = shift;
    return $self->{cache_date_minz} if exists $self->{cache_date_minz};
    return $self->{cache_date_minz} = do {
	my $d = $self->minfo_value("Date");
	if ($d) {
	    $d =~ s/\([^)]*\)/ /g;
	    $d = HTTP::Date::str2time($d);
	    $d = HTTP::Date::time2isoz($d) if $d;
	    $d = sprintf("%4d%02d%02d%02d%02d", 
			 (HTTP::Date::parse_date($d))[0,1,2,3,4]) if $d;
	}
	$d;
    }
}

sub c_idate_min {
    my $self = shift;
    return $self->{cache_date_min} if exists $self->{cache_date_min};
    return $self->{cache_date_min} = do {
	my $d = $self->minfo_value("Date");
	if ($d) {
	    $d =~ s/\([^)]*\)/ /g;
	    $d = HTTP::Date::str2time($d);
	    $d = HTTP::Date::time2iso($d) if $d;
	    $d = sprintf("%4d%02d%02d%02d%02d", 
			 (HTTP::Date::parse_date($d))[0,1,2,3,4]) if $d;
	}
	$d;
    }
}

sub c_reason {
    my $self = shift;
    unless (exists $self->{cache_reason}) {
	$self->{cache_reason} = [ $self->minfo_value("Reason") ];
    }
    return $self->{cache_reason}[0] unless wantarray;
    return @{$self->{cache_reason}};
}

sub c_reason_csv {
    my $self = shift;
    return $self->{cache_reason_csv} if exists $self->{cache_reason_csv};
    return $self->{cache_reason_csv} =
	join(", ", $self->minfo_value("Reason"));
}

sub c_from {
    my $self = shift;
    if (exists $self->{cache_from}) {
	return $self->{cache_from}[0] unless wantarray;
	return @{$self->{cache_from}};
    }

    my @from;
    if (my $env_from = $self->minfo_value("Envelope-From")) {
	push(@from, lc($env_from));
    }

    for my $h (qw(From Sender Reply-To Resent-From Resent-Sender)) {
	my $v = $self->header($h);
	next unless $v;
	for my $addr (_clean_mail_addresses($v)) {
	    next if grep $_ eq $addr, @from;
	    push(@from, $addr);
	}
    }

    $self->{cache_from} = \@from;
    return $from[0] unless wantarray;
    return @from;
}

sub c_to {
    my $self = shift;
    if (exists $self->{cache_to}) {
	return $self->{cache_to}[0] unless wantarray;
	return @{$self->{cache_to}};
    }

    my @to;
    for my $env_to ($self->minfo_value("Envelope-To")) {
	push(@to, lc($env_to));
    }

    for my $h (qw(To Cc Bcc Resent-To Resent-Cc Resent-Bcc)) {
	my $v = $self->header($h);
	next unless $v;
	for my $addr (_clean_mail_addresses($v)) {
	    next if grep $_ eq $addr, @to;
	    push(@to, $addr);
	}
    }

    $self->{cache_to} = \@to;
    return $to[0] unless wantarray;
    return @to;
}

sub c_received_ip {
    my $self = shift;
     if (exists $self->{cache_received_ip}) {
	return $self->{cache_received_ip}[0] unless wantarray;
	return @{$self->{cache_received_ip}};
    }
  
    my @received = $self->header("Received");
    my @r_ip = ();
    unless ( @received && defined($received[0]) ) { # check $received[0] != undef
	@received = ();
    }
    for my $r (@received) { 
	# is there a more elegant way to write an ipv4 regex?
	if ( my @ip = $r =~ m/(\d{1,3}(?:\.\d{1,3}){3})/g ) {
	    push @r_ip, @ip;
	}
    }
    $self->{cache_received_ip} = \@r_ip;
    return $r_ip[0] unless wantarray;
    return @r_ip;
}

sub _clean_mail_addresses {
    my $v = shift;
    require Mail::Address;
    local $Mail::Address::VERBOSE = 0;  # silence "Unmatched () ..." warnings
    my @addr = grep length, map lc($_->address), Mail::Address->parse($v);
    return @addr;
}

1;# there may be a more elegant way to write 

__END__

=head1 NAME

PureMessage::Message::ComputedFields - Additional methods for message objects

=head1 DESCRIPTION

Loading this module populate the C<PureMessage::Message> class with
additional methods that compute values based on the message headers
and content:

The following extra methods are provided for C<PureMessage::Message> objects:

=over

=item $msg->c_content_type

Returns the normalized content type of the message.  It will always be
a simple string like "text/plain" or "multipart/alternative".

=item $msg->c_charset

Returns the character set used for the body of the message. It will be
a simple string like "us-ascii", "iso-8859-1", "big5".  This returns
as C<undef> if the body is not of type "text/*".

=item $msg->c_date_time

Returns the message's C<Date> header as number of seconds since epoch
(as for time).

=item $msg->c_date_isoz

Returns the message's C<Date> header as a ISO date.

=item $msg->c_idate_onlyz

Returns the message's C<Date> header as an integer C<YYYYMMDD0000> 
representation of the date in GMT to be used for date range comparisons.

=item $msg->c_idate_only

Returns the message's C<Date> header as an integer C<YYYYMMDD0000> 
representation of the date in localtime to be used for date range 
comparisons.

=item $msg->c_idate_hourz

Returns the message's C<Date> header as an integer C<YYYYMMDDHH00>
representation of the date in GMT to be used for date range comparisons.

=item $msg->c_idate_hour

Returns the message's C<Date> header as an integer C<YYYYMMDDHH00>
representation of the date in localtime to be used for date range 
comparisons.

=item $msg->c_idate_10minz

Returns the message's C<Date> header as an integer C<YYYYMMDDHHM0>
representation of the date in GMT to be used for date range comparisons.

=item $msg->c_idate_10min

Returns the message's C<Date> header as an integer C<YYYYMMDDHHM0>
representation of the date in localtime to be used for date range 
comparisons.

=item $msg->c_idate_minz

Returns the message's C<Date> header as an integer C<YYYYMMDDHHMM>
representation of the date in GMT to be used for date range comparisons.

=item $msg->c_idate_min

Returns the message's C<Date> header as an integer C<YYYYMMDDHHMM>
representation of the date in localtime to be used for date range 
comparisons.

=item $msg->c_from

Returns all the sender addresses given for the message.  All addresses
are normalized in lower case.

=item $msg->c_hops

Returns the number of hops this message has traveled.  Computed by
counting the number of Received headers.

=item $msg->c_subj15

This returns the normalized 15 first chars of the subject header.  It
is mainly useful for grouping similar messages together.

=item $msg->c_subject_utf8

This returns the subject with all MIME encoding undone and white space
normalized.

=item $msg->c_subject_ascii

This returns the subject with all MIME encoding undone and white space
normalized.  Any characters outside of the printable ASCII range is
replaces by "*".  This guarantees that the string is printable on any
terminal device.

=item $msg->c_subject_body_utf8

This returns up to the 80 first characters if the subject and the
message body combined with normalized whitespace.

=item $msg->c_subject_body_utf8( $len_limit )

Passing an argument to c_subject_body_utf8 might influence how much to
extract.

=item $msg->c_qdate_iso

This returns the date the message was stored.

=item $msg->c_qdate_time.

This returns the date the message was stored as number of seconds
since epoch (as for time).

=item $msg->c_reason_csv

This returns the all the reasons the message was quarantined as a
single comma separated string.

=item $msg->c_reason

This returns the all the reasons the message was quarantined as an
array in array context or the first reason in scalar context.

=item $msg->c_to

Returns all the recipient addresses given for the message.  All
addresses are normalized in lower case.

=item $msg->c_received_ip

Returns all ip addreses found in Received headers in text format.

=back

=head1 SEE ALSO

L<PureMessage::Message>

