package PureMessage::MessageFile;

use strict;
use Fcntl qw(O_RDONLY O_RDWR);
use Carp qw(croak carp);
use base 'PureMessage::AnnotatedMessage';

my %HEADER_RE;  # cache of regular expressions that match headers

sub _MINFO_SIZE_GUESS () { 2048 }

sub new {
    my($class, $fname, %opts) = @_;
    my $writable = delete $opts{writable};

    if (%opts) {
	carp "Unknown option '$_'" for keys %opts;
    }

    my $mode = $writable ? O_RDWR : O_RDONLY;
    CORE::sysopen(my $f, $fname, $mode) || die "Can't open '$fname': $!";

    # Try to read the last block from the file, if we are luck we read the
    # whole minfo with it.
    my $size;
    if ($^O eq "freebsd") {
	# On FreeBSD it is possible to seek to a position before 0 with
	# the result that the file pointer wraps around and the following
	# sysread fails with "File too large".  Need an extra stat(2) for
	# this platform so that we only seek if the file is large enough.
	$size = -s $f;
	CORE::sysseek($f, - _MINFO_SIZE_GUESS, 2) if $size > _MINFO_SIZE_GUESS;
    }
    else {
	$size = CORE::sysseek($f, - _MINFO_SIZE_GUESS, 2);
	if (defined $size) {
	    $size += _MINFO_SIZE_GUESS;
	}
	else {
	    $size = -s $f;
	}
    }
    my $m_info;
    unless (defined(sysread($f, $m_info, _MINFO_SIZE_GUESS))) {
	die "Can't read PureMessage message file signature from '$fname': $!";
    }

    # Check the signature (last line) of the file
    my $sig_size = 26;
    my $sig = substr($m_info, -$sig_size);
    my $m_offset;
    if ($sig =~ /^Message-Info-Size: 0x([\da-f]{4})\n\z/ ||
        $sig =~ /^Msg-Info-Size: 0x([\da-f]{8})\n\z/) {
	 $m_offset = hex($1)
    }
    else {
	# try to the PerlMx-2.1 Quarantine Format as well
	$sig_size = 29;
	$sig = substr($m_info, -$sig_size);
	if ($sig =~ /^Quarantine-Info-Size: 0x([\da-f]{4})\n\z/) {
	    $m_offset = hex($1);
	}
	else {
	    die "Not a PureMessage message file (bad signature for '$fname')";
	}
    }

    my $missing_bytes = $m_offset - length($m_info);
    if ($missing_bytes > 0) {
	# need to read more to get all of the minfo
        CORE::sysseek($f, -$m_offset, 2) || die "Can't seek in '$fname': $!";
	my $buf;
	my $n = sysread($f, $buf, $missing_bytes);
	die "Can't read message info from '$fname': $!" unless defined $n;
	die "Failed to read full message info ($n/$missing_bytes) from '$fname'" unless $n == $missing_bytes;
	substr($m_info, 0, 0) = $buf;
    }
    elsif ($missing_bytes < 0) {
	# throw away the extra bytes we read
	substr($m_info, 0, -$missing_bytes) = "";
    }
    $m_info =~ s/^\n-- \n// || die "Bad minfo header in '$fname'";
    substr($m_info, -$sig_size) = '';

    CORE::sysseek($f, 0, 0) || die "Can't reset file pointer for '$fname': $!";

    bless {
	pmxmf_fname   => $fname,
	pmxmf_fh      => $f,
        pmxmf_ro      => !$writable,
	pmxmf_minfo   => $m_info,
	pmxmf_end     => $size - $m_offset,
	pmxmf_fsize   => $size,
	pmxmf_pos     => 0,
    }, $class;
}

sub filename {
    return shift->{pmxmf_fname};
}

sub _minfo_block {
    my $minfo = shift;
    my $buf = join("", "\n-- \n", $minfo->as_hdr,"Msg-Info-Size: 0xXXXXXXXX\n");
    substr($buf, -9, 8) = sprintf "%08x", length($buf);
    return $buf;
}

sub _print {
    my ($self,$buf) = @_;
    my ($f,$offset,$len,$written);

    $f   = $self->{pmxmf_fh};
    $len = length($buf);
    $offset = 0;
    while ( $len > 0 ) { 
	if ( defined($written = CORE::syswrite($f,$buf,$len,$offset)) ) {
	    $offset += $written;
	    $len    -= $written; 
	}
	else {
	    $len = -1;
	}
    }

    return ( ($len!=-1) ? 1 : 0 );
}

sub make {
    my $class = shift;
    my $fname = shift;
    my $minfo = @_ == 1 ? shift : do {
	require PureMessage::MessageInfo;
        PureMessage::MessageInfo->new(@_);
    };
    open(my $f, ">>", $fname) || die "Can't append to '$fname': $!\n";
    print $f _minfo_block($minfo);
    close($f) || die "Can't append message info to '$fname': $!";
    1;
}

sub minfo {
    my $self = shift;
    unless (@_) {
	my $minfo = $self->{pmxmf_minfo};
	return $minfo if ref($minfo);

	require PureMessage::MessageInfo;
	$minfo = eval { PureMessage::MessageInfo->new($minfo) };
	die "message file $self->{pmxmf_fname}: $@" if $@;

	return $self->{pmxmf_minfo} = $minfo;
    }

    croak("Can't update minfo for readonly " . ref($self))
	if $self->{pmxmf_ro};

    my $fh = $self->{pmxmf_fh};
    my $pos = $self->{pmxmf_pos};
    my $end = $self->{pmxmf_end};
    my $size = $self->{pmxmf_fsize};

    my $minfo = shift;
    my $block = _minfo_block($minfo);
    my $blklen = length($block);

    if ($pos != $end) {
	CORE::sysseek($fh, $end, 0) || die "Can't seek: $!";
    }

    $self->_print($block) || die "Can't write minfo: $!\n";

    if ($size - $end > $blklen) {
	truncate($fh, $end + $blklen) || die "Can't truncate: $!";
    }

    $self->{pmxmf_fsize} = $end + $blklen;
    $self->{pmxmf_pos}   = $self->{pmxmf_fsize};
    $self->{pmxmf_minfo} = $minfo->as_hdr;
}

sub minfo_remove {
    my $self = shift;

    croak("Can't remove minfo for readonly " . ref($self))
	if $self->{pmxmf_ro};

    my $fh = $self->{pmxmf_fh};
    my $end = $self->{pmxmf_end};
    truncate($fh, $end) || die "Can't truncate minfo at $end: $!";
    $self->{pmxmf_fsize} = $end;
}

sub minfo_value {
    my $self = shift;
    if (@_ > 1) {
	my $minfo = $self->minfo;
	my @old = $minfo->value(@_);
	$self->minfo($minfo);  # set
	return wantarray ? @old : $old[0];
    }
    return $self->minfo->value(shift);
}

sub header_size {
    return length(shift->header);
}

sub message_size {
    return shift->{pmxmf_end};
}

sub file_size {
    return shift->{pmxmf_fsize};
}

sub message {
    my $self = shift;
    my $buf;
    my $pos = $self->{pmxmf_pos};
    my $end = $self->{pmxmf_end};
    my $fh = $self->{pmxmf_fh};
    if ($pos) {
	CORE::sysseek($fh, 0, 0) || die "Can't reset file pointer: $!";
        $self->{pmxmf_pos} = 0;
    }
    my $n = $self->_read($buf, $end);
    return undef unless defined($n);
    CORE::sysseek($fh, $pos, 0) || die "Can't reset file pointer: $!";
    $self->{pmxmf_pos} = $pos;
    return undef if $n != $end;
    return $buf;
}

sub _file_copy {
    my($self, $file, $copy_bytes) = @_;
    my $opened;
    
    unless (ref($file) || ref(\$file) eq "GLOB") {
	open(my $fh, ">", $file) || croak("Can't open $file: $!");
	$file = $fh;
	$opened++;
    }

    my $fh = $self->{pmxmf_fh};
    my $pos = $self->{pmxmf_pos};
    if ($pos) {
	CORE::sysseek($fh, 0, 0) || die "Can't reset file pointer: $!";
        $self->{pmxmf_pos} = 0;
    }

    # make sure the 'print' method used below
    # works for old file handles.
    require IO::Handle;
    
    my $remaining = $copy_bytes;
    my $buf;

    while (1) {
	my $len = $remaining > 4*1024 ? 4*1024 : $remaining;
	last unless $len;

	my $n = $self->_read($buf, $len);
	if ($n) {
	    $file->print($buf) or die "Can't print: $!";
	    $remaining -= $n;
	}
	else {
	    die "Read failed: $!" unless defined $n;
	    die "Early EOF";
	}
    }

    if ($opened) {
	close($file) || die "Close failed: $!";
    }

    CORE::sysseek($fh, $pos, 0) || die "Can't set back file pointer: $!";

    $self->{pmxmf_pos} = 0;

    1;
}

sub message_copy {
    my $self = shift;
    my $file = shift;
    _file_copy($self, $file, $self->{pmxmf_end});
}

sub file_copy {
    my $self = shift;
    my $file = shift;
    _file_copy($self, $file, $self->{pmxmf_fsize});
}

sub header {
    my $self = shift;
    unless (exists $self->{pmxmf_hdr}) {
	my $old_pos = $self->{pmxmf_pos};
	my $fh = $self->{pmxmf_fh};
	if ($old_pos) {
	    CORE::sysseek($fh, 0, 0) || die "Can't reset file pointer: $!";
	}
	my @hdr;
        my $eoh = 0;
	local $_;
	while (<$fh>) {
            if ( /^\r?\n\z/ ) {
                $eoh = 1;     # information is confirmed as type 'header'
                last;
            }
	    last if /^-- $/;  # safety
	    push(@hdr, $_);
	}
	CORE::sysseek($fh, $old_pos, 0) || die "Can't move back file pointer: $!";
        if ( !($eoh && scalar(@hdr)) ) {
            return undef;
        }
	$self->{pmxmf_hdr} = join("", @hdr);
    }

    for ($self->{pmxmf_hdr}) {
	return $_ unless @_;
	my $field_name = lc(shift);
	my $re = ($HEADER_RE{$field_name} ||=
		  qr/^\Q$field_name\E[ \t]*:[ \t]?/mi);
	my @h;
	pos($_) = 0;
	while (/$re/gc) {
	    my $start = pos;
	    my $end = $start;
	    $end = pos if /\G[^\r\n]*/gc; # get first line
	    $end = pos while /\G\r?\n[\t ][^\r\n]*/gc; # get rest
	    return substr($_, $start, $end-$start) unless wantarray;
	    push @h, substr($_, $start, $end-$start);
	}
	return wantarray ? @h : undef;
    }
}


sub tell {
    my $self = shift;
    return $self->{pmxmf_pos};
}

sub seek {
    my($self, $offset, $whence) = @_;
    $whence ||= 0;
    my $pos = $self->{pmxmf_pos};
    my $end = $self->{pmxmf_end};
    if ($whence == 0) {
	$pos = $offset;
    }
    elsif ($whence == 1) {
	$pos += $offset;
    }
    elsif ($whence == 2) {
	$pos = $end + $offset;
    }
    else {
	croak("Illegal whence value '$whence'");
    }

    if ($pos < 0 || $pos > $end) {
	$! = "Illegal seek";  #XXX
	return undef;
    }

    unless (CORE::sysseek($self->{pmxmf_fh}, $pos, 0)) {
	return undef;
    }

    $self->{pmxmf_pos} = $pos;
    return 1;
}

sub read {
    croak("Wrong number of arguments") if @_ != 3;
    my $self = shift;
    my $len  = $_[1];
    my $pos  = $self->{pmxmf_pos};
    my $end  = $self->{pmxmf_end};

    $len = $end - $pos if $len + $pos > $end;

    return ($self->_read($_[0],$len));
}

sub _read {
    my $self = shift;
    my $len = $_[1];

    unless ($len) {
        $_[0] = "";
        return 0;
    }

    my ($lenread,$n) = (0,0);
    while ( $len > 0 ) {
        if (defined($n = CORE::sysread($self->{pmxmf_fh}, $_[0], $len, $lenread))) {
            $self->{pmxmf_pos} += $n;
            $lenread           += $n;
            $len               -= $n;
        }
        else {
            $len = -1;
        }
    }

    return ( ($len!=-1) ? $lenread : undef );
}

1;

__END__

=head1 NAME

PureMessage::MessageFile - RFC822 message with MessageInfo

=head1 DESCRIPTION

Objects of this class represent a file containing an RFC822 message
with attached MessageInfo (minfo).

This class is a subclass of C<PureMessage::AnnotatedMessage>.

The following methods are provided in addition to those provided by C<PureMessage::AnnotatedMessage>:

=over

=item $f = PureMessage::MessageFile->new( $filename, %opts )

Constructor.  Will croak if the file is not of the apropriate format
or can't be opened.

If no options are given the file is opened in readonly mode. The
'writable => 1' option opens the file such that the attached minfo
object can be updated.

=item PureMessage::MessageFile->make( $rfc822_filename, $minfo )

=item PureMessage::MessageFile->make( $rfc822_filename, %kw )

This will transform a file that contains a RFC822 message into a
MessageFile by attaching the minfo to it.  The file will be modified
in-place.  Will croak if it gets into trouble.

Instead of passing in a minfo object as the second argument this
method can take key/value pairs and it will create a new minfo
internally and store that with the file.

=item $f->filename

Returns back the filename passed in during construction.

=item $f->minfo_remove

Kills the minfo block from the message.  What is left is the plain
RFC822 message.  This modifies the current mfile in-place. This will
fail unless the mfile object was created with the C<writable> option.

=back

=head1 SEE ALSO

L<PureMessage::AnnotatedMessage>, L<PureMessage::Message>,
L<PureMessage::MessageInfo>, L<PureMessage::MessageStore>

