package PureMessage::MessageStore::Dir;

use strict;
use Carp qw(croak carp);

use ActiveState::Handy qw(run);
use PureMessage::MessageFile ();
use PureMessage::MessageInfo ();
use PureMessage::_sys;
use PureMessage::MessageLog qw(join_line);
use PureMessage::MessageStore::Dir::Index;
use PureMessage::MessageStore::Dir::Fields;
use PureMessage::MessageStore::Dir::Config;
use File::Find;

BEGIN { eval "sub TEST_MODE() { 0 }" unless defined(&TEST_MODE) }

sub FIELDS() { "PureMessage::MessageStore::Dir::Fields" }

sub new {
    my $class = shift;
    my $dir = shift;
    my %opts = @_;

    my $create = delete $opts{create};
    my $verbose = delete $opts{verbose} || 0;
    my $store_class = delete $opts{class};
    my $dirmode = delete $opts{dirmode};
    my $logger = delete $opts{logger};
    my $idx_type = delete $opts{'index-type'};
    my $msg_writable => delete $opts{message_writable} || 0;

    # silently accept this option even though we don't use it (bug 30612)
    delete $opts{'connect'}; 
    

    if (%opts) {
	if ($logger) {
	    $logger->log("WARNING", "Unknown message store option '$_'") for keys %opts;
	}
	else {
	    carp "Unknown option '$_'" for keys %opts;
	}
    }

    my $self = {
	pmxdir_dir => $dir,
	pmxdir_verbose => $verbose,
	pmxdir_dirmode => $dirmode,
	pmxdir_logger => $logger,
	pmxdir_writable => $msg_writable,
	pmxdir_disk_used_delta => 0,
	pmxdir_num_msg_delta => 0,
    };
    bless($self, $class);

    if ($create) {
        print "Initializing message quarantine in $dir...\n" if $verbose;
	for (qw(new cur tmp mset)) {
	    next if -d "$dir/$_";
	    $class->_mk_path("$dir/$_", $verbose, $dirmode) ||
		croak("Can't create directory '$dir/$_': $!");
	}

	$class->init_cfg_file(dir => $dir, verbose => $verbose)
            unless (-f "$dir/db.conf");

	if ( $idx_type ) { 
	    if (-f "$dir/index.conf") {
		$self->_edit_index_type($idx_type);
	    } else {
		my $index_conf = "$dir/index.conf";
		$logger->log("DEBUG", "Setting up $index_conf") if $logger;
		print "Setting up $index_conf\n" if $verbose;
		open(my $f, ">", $index_conf) || 
		    croak ("Can't create file '$index_conf': $!");
		print $f "index_type = $idx_type\n";
		close($f) || croak("Can't write '$index_conf': $!");

                # fields indexed by default
                eval {
                    for my $field (qw(h_message_id m_queue_id c_from c_to
                                      m_enduser m_mapped_recipients)) {
                        $self->add_index($field);
                        print "Added index for '$field'.\n";
                    }
                };
                if ($@) {
                    my $msg = "Error while adding default index information into $index_conf: $@";
                    $self->log("WARNING", $msg);
                    carp $msg;
                }
	    }
	}
    }
    else {
	if (-d $dir) {
	    my @missing;
	    for (qw(new cur tmp mset)) {
		push(@missing, $_) unless -d "$dir/$_";
	    }
	    if (@missing) {
		croak("Malformed database directory '$dir': missing subdirs: " .
		      join(",", map qq('$_'), @missing));
	    }
	    croak("Malformed database directory '$dir': missing db.conf")
		unless -f "$dir/db.conf";
	}
	else {
	    croak("No such directory '$dir'");
	}
    }

    my $conf = $class->get_config($dir);
    $self->{pmxdir_location} = $conf->value("location");
    $self->{pmxdir_location_id} = $conf->value("location_id");
    return $self;
}

# retrieve the config and fix it if necessary (upgrade path)
sub get_config {
    my $class = shift;
    my $dir = shift;
    my $config = PureMessage::MessageStore::Dir::Config->new("$dir/db.conf");
    unless ($config->value("location") and $config->value("location_id")) {

        # get writable cfg object
        $config = PureMessage::MessageStore::Dir::Config->new(
            "$dir/db.conf", 
	    writable => 1,
            timeout => 5
        );
        my ($location_name, $location_id) = gen_location();
        $config->value("location" => $location_name)
            unless $config->value("location");
        $config->value("location_id" => $location_id)
            unless $config->value("location_id");
        $config->save;

        # now we need a read-only config
        $config = PureMessage::MessageStore::Dir::Config->new("$dir/db.conf");
    }
    return $config;
}

sub init_cfg_file {
    my ($store_class, %opts) = @_;

    my $dir = $opts{dir};
    my $location_name = $opts{location_name};
    my $location_id = $opts{location_id};
    my $verbose = $opts{verbose};

    ($location_name, $location_id) = gen_location()
        unless ($location_name and $location_id);

    my $db_conf = "$dir/db.conf";
    print "Setting up $db_conf\n" if $verbose;
    open(my $f, ">", $db_conf) || croak ("Can't create file '$db_conf': $!");
    print $f "class = $store_class\n";
    print $f "location = ", $location_name, "\n";
    print $f "location_id = ", $location_id, "\n";
    print $f "insert_rule_hits = 0\n";
    close($f) || croak("Can't write '$db_conf': $!");
}

sub DESTROY {
    my $self = shift;
    $self->_status_cache_sync;
}

sub dir {
    shift->{pmxdir_dir};
}

sub location { "localhost" }

sub location_id {
    my $self = shift;
    return $self->{pmxdir_location_id};
}

sub store_id {
    "dir:" . shift->{pmxdir_dir};
}

sub gen_location {
    my $location_name = 'localhost'; 
    my $location_id = 0;
    eval {
        require PureMessage::Resources::Location;
        my $location = PureMessage::Resources::Location->new();
        if ($location->location_name and $location->location_id) {
            $location_name = $location->location_name();
            $location_id = $location->location_id();
        }
        else {
            warn "Using defaults for location_name, location_id " .
                 "($location_name, $location_id)";
        }
    };
    warn $@ if $@;
    return ($location_name, $location_id);
}

sub get_user_views {
    my $self = shift;
    my $addr = shift;

    warn "Requesting multiple views in scalar context"
	if !wantarray && @_ > 1 && $^W;

    require PureMessage::MessageStore::Dir::UserView;
    return PureMessage::MessageStore::Dir::UserView->new( 
	store => $self, address => $addr, %{$_[0]}) unless wantarray;
    
    my @views;
    foreach my $v (@_) {
	push(@views, PureMessage::MessageStore::Dir::UserView->new(
		store => $self, address => $addr, %$v));
    } 
    return @views;
}

# XXX This is also implemented in PM::QMan; these should use the
# same code.
sub record_user_action {
    my ($self, %opts) = @_;
    my $action = $opts{action};
    my $addr = $opts{address};
    
    foreach my $mid (@{$opts{ids}}) {
	my $msg = $self->message($mid);
	my @minfo = $msg->minfo_value('Enduser');
	@minfo = grep { $_ =~ /^$action $addr$/ } @minfo;
	push(@minfo, "$action $addr") unless $opts{remove};
	$msg->minfo_value("Enduser", @minfo ? @minfo : undef);
    }
}


sub log {
    my $self = shift;
    if (my $logger = $self->{pmxdir_logger}) {
	return $logger->log(@_);
    }
    my $prio = shift;
    if ($prio eq "DEBUG" || $prio eq "INFO" || $prio eq "NOTICE") {
	return unless $self->{pmxdir_verbose};
    }

    my $msg = shift;
    $msg =~ s/^\s+//;
    $msg =~ s/\s+\z//;
    $msg =~ s/\s+/ /g;
    warn "LOG_$prio $msg\n";
}

sub new_tempfile {
    my $self = shift;
    require PureMessage::MessageStore::Dir::TempFile;
    return PureMessage::MessageStore::Dir::TempFile->new($self, @_);
}

sub export {
    my($self, $id, $mfile) = @_;
    # we can "cheat" since the internal storage is
    # already on the mfile format
    my $m = $self->message($id) || return undef;
    require File::Copy;
    return File::Copy::copy($m->filename, $mfile);
}

sub import {
    my $self = shift;
    for my $fname (@_) {
	my $f = PureMessage::MessageFile->new($fname);
	my $tmp = $self->new_tempfile;
	$f->message_copy($tmp);
	$tmp->store($f->minfo);
    }
    scalar(@_);
}

sub store {
    my $self = shift;
    my $file = shift;
    open(my $fh, "<", $file) || die "Can't open '$file': $!";
    my $tmp = $self->new_tempfile;
    my $buf;

    while (1) {
	my $n = read($fh, $buf, 4*1024);
	if ($n) {
	    $tmp->print($buf);
	}
	else {
	    die "Read from '$file' failed: $!" unless defined $n;
	    last;
	}
    }

    $tmp->store(@_);
}

sub _new_index {
    my $self = shift;
    return PureMessage::MessageStore::Dir::Index->new(
	@_,
	alt_config => "$self->{pmxdir_dir}/index.conf");
}

sub _edit_index_type {
    my $self   = shift;
    my $i_type = shift;
    my $dir    = $self->dir;

    require PureMessage::MessageStore::Dir::Config;
    my $conf = PureMessage::MessageStore::Dir::Config->new("$dir/index.conf", 
	writable => 1, timeout => 5);
    my $old_i_type = $conf->value("index_type");
    unless ( $old_i_type eq $i_type ) {
	$conf->value("index_type" => $i_type);
	$conf->save;
    }
    undef($conf);
 
    return $old_i_type;
}

sub get_index_type {
    my $self = shift;
    my $dir  = $self->dir;

    require PureMessage::MessageStore::Dir::Config;
    my $conf = PureMessage::MessageStore::Dir::Config->new("$dir/index.conf", 
	writable => 0, timeout => 5);

    return $conf->value("index_type");
}

sub add_index {
    my $self = shift;

    # XXX should happend in Index.pm
    require PureMessage::MessageStore::Dir::Fields;
    my $index_conf = "$self->{pmxdir_dir}/index.conf";
    for my $field (@_) {
	croak("Bad field '$field'")
	    unless PureMessage::MessageStore::Dir::Fields->field_ok($field);

	if (open(my $index_fh, "<", $index_conf)) {
	    local $_;
	    while (<$index_fh>) {
		if (/^\s*<index (\w+)/) {
		    croak("There is already an index for '$field'")
			if $1 eq $field;
		}
	    }
	}

	open(my $index_fh, ">>", $index_conf) ||
	    die "Can't open '$index_conf': $!";
	print $index_fh "\n<index $field>\n    field = $field\n</index>\n";
	close($index_fh) || die "Can't write to $index_conf: $!";
    }
}

sub index {
    my($self, %opts) = @_;
    my $progress_cb = delete $opts{progress};
    my $message_cb  = delete $opts{indexer};
    carp "Unknown option '$_'" for keys %opts;
    my $dir = $self->dir;
    my $new = "$dir/new";
    my $cur = "$dir/cur";

    # Numeric location identifier has to be non-zero
    unless ($self->{pmxdir_location_id}) {
        my $msg = "Unable to proceed with indexing: " .
                  "incorrect value for parameter 'location_id' in $dir/db.conf";
        $self->log("ERROR", $msg);
        die $msg;
    }

    require ActiveState::DiskUsage;
    require PureMessage::MessageStore::Dir::Config;

    # Open a global lock on the message store, timeout and croak if another 
    # process is already indexing
    my $conf = PureMessage::MessageStore::Dir::Config->new("$dir/db.conf", 
	writable => 1, timeout => 5);
    opendir(my $d, $new) || die "Can't read '$new' directory: $!";
    my @files = grep !/^\./, readdir($d);
    if (TEST_MODE) {
       # Sort by inode (required for unit tests to work correctly)
       @files = sort { (stat($new."/".$a))[1] <=> (stat($new."/".$b))[1] } @files;
    }
    closedir($d);

    unless (@files) {
	$conf->unlock;
	return 0;
    }

    my $id;
    my $first_id;
    my $count = 0;
    my $du = 0;
    my $last_index_file;
    eval {
	&$progress_cb(0.0) if $progress_cb;

	# If the last ID value doesn't exist, convert the id from the 
	# .last_index file that was used in previous versions. Default to 999
	# if .last_index is not available.
	$id = $conf->value("last_id");
	unless ($id) {
	    if (open(my $last_index, "<", "$dir/.last_index")) {
		$id = <$last_index>;
		chomp($id);
		close($last_index);
		$last_index_file = "$dir/.last_index";
	    }
	    $id ||= 999;
	}

	# Each file in 'new' gets assigned a new index ID and is renamed into
	# the 'cur' folder. The file's meta-data is not added to the indexes.
	for my $f (@files) {
	    my $mf;
	    unless ($mf = eval { PureMessage::MessageFile->new("$new/$f", writable => 1) }) {
		$self->log("ERR", "Can't index '$f': $@");
		next;
	    }

            next if ($message_cb && $message_cb->($mf));

	    my $fname;
	    do {
		$id++;
		$fname = "$cur/" . $self->_id_fname($id);
	    } while -e $fname;
	    unless (rename("$new/$f", $fname)) {
		my $err = "$!";
		(my $dirname = $fname) =~ s,/[^/]+\z,,;
		die "Can't rename as $fname: $err" if -d $dirname;
		$self->_mk_path($dirname);
		unless (rename("$new/$f", $fname)) {
		    die "Can't rename as $fname: $!";
		}
	    }

            my $global_id = $self->_new_global_id($id);
            $mf->minfo_value('Global-ID', "$global_id");

	    my $store_meta = 'yes';
	    $store_meta = lc($mf->minfo_value('Store-Meta'))
		if defined $mf->minfo_value('Store-Meta');

	    $self->_idx_log_insert('cur', $id) if $store_meta eq 'yes';

	    $first_id ||= $id;
	    $du += ActiveState::DiskUsage::du($fname);
	    $count++;

	    &$progress_cb($count/@files) if $progress_cb;
	}
    };
    warn $@ if $@;

    if (defined $id) {
	$conf->value("first_id" => $first_id) if $first_id && !$conf->value("first_id");
	$conf->value("last_id" => $id);
	$conf->value("disk_used", ($conf->value("disk_used") || 0) + $du);
	$conf->value("num_msg", ($conf->value("num_msg") || 0) + $count);
    }

    # Commit the changes and release the global lock
    $conf->save;
    undef($conf);
    if ($last_index_file) {
	unlink($last_index_file) || $self->log("WARNING", "Can't unlink '$last_index_file': $!");
    }

    $self->log("INFO", "Indexed $count new messages");

    return $count;
}

# Returns the list of files with scheduled index updates. In concurrent
# mode (which will be in effect if the underlying storage/indexing
# mechanism supports it) the list will contain at most one file, which
# is the consolidation of all the work files that are currently not being
# processed. In "exclusive" mode (non-concurrent) the list will contain
# all the pending work files. Croaks on erors.
sub _get_work_files {
    my $self = shift;
    my $folder = shift;
    unless ($folder) {
        my $msg = "_get_work_files(): required parameter missing.";
	$self->log("ERR", $msg);
        croak $msg;
    }
    my $curr_work_file = shift || "$folder/index_log";
    my $pending_dir = "$folder/index_pending";
    unless (-d $pending_dir) {
	require File::Path;
	File::Path::mkpath($pending_dir);
    }

    my $my_pid = $$;
    my $pending_work_file = "$pending_dir/$my_pid";
    if (-e $curr_work_file) {
        _append($curr_work_file, $pending_work_file);
        unlink($curr_work_file);
    }

    # Get the list of pending files that we need to work on.
    opendir(my $DIR, $pending_dir)
	or croak "Failed to read the $pending_dir directory: $!";

    # work file names consist of digits only
    my @pending_files = sort { (-M $a) <=> (-M $b) }
                        map {"$pending_dir/$_"}
                        grep { /^\d+$/ } readdir($DIR);

    my @work_files;
    my $tmp_consolidated = "$pending_dir/tmp_consolidated_$my_pid";
    foreach my $file (@pending_files) {
        if ($self->{'pmxdir_concurrent_ok'}) {

	    # if other process is taking care of this file then skip it
            my $indexer_pid = $file;
            $indexer_pid =~ s|.*/||;
	    next if (($my_pid != $indexer_pid) and kill(0, $indexer_pid));

            eval {
                _append($file, $tmp_consolidated);
	        unlink $file
            };
            if ($@) {
		my $msg = "Error while consolidating work files: $@";
		$self->log("ERR", $msg);
		croak $msg;
            }
        } else {
            push @work_files, $file;
        }
    }
    closedir($DIR);

    if (($self->{'pmxdir_concurrent_ok'}) and (-f $tmp_consolidated)) {
	unless (rename($tmp_consolidated, $pending_work_file)) {
            my $msg = "Error while consolidating work files: $!";
            $self->log("ERR", $msg);
            croak $msg;
        }
	push @work_files, $pending_work_file;
    }

    return @work_files;
}

# Acquires qmeta-index lock that'll exclude other instances of the
# indexer for this folder and stores it into $self; croaks on error.
sub _qmx_lock_acquire {
    my $self = shift;
    my $folder = shift;
    $folder ||= 'cur';

    # currently this is how re-entrant locking works
    return if ($self->{'qmx_lock'});

    my $lock = $self->counter("meta-index-$folder", writable => 1,
                              create => 1, timeout => 5);
    die "Meta-indexing aborted: failed to acquire lock" unless $lock;
    $self->{'qmx_lock'} = $lock;
}

# releases the qmeta-index lock, updating the counter with the specified
# value or current time
sub _qmx_lock_release {
    my $self = shift;
    my $value = shift || time;

    # nothing to do if we don't have it
    $self->{'qmx_lock'} or return;

    $self->{'qmx_lock'}->value($value);
    $self->{'qmx_lock'}->save;
    delete $self->{'qmx_lock'};
}

# appends source file to destination file; croaks on errors
sub _append {
    my $source = shift;
    $source = shift if ref $source;
    my $destination = shift;
    open(my $SRC, "<$source")
        or croak "Unable to open $source for reading: $!";
    open(my $DEST, ">>$destination")
        or croak "Unable to open $destination for appending: $!";
    print $DEST $_ foreach (<$SRC>);
    close($SRC) or carp "Unable to close $source: $!";
    close($DEST) or carp "Unable to close $destination: $!";
}

# Update the meta-data indexes for all messages that have been added or 
# removed from this folder since the last time this was run.
sub meta_index {
    my($self, %opts) = @_;
    my $progress_cb = delete $opts{progress};
    my $dir = $self->dir;
    my $folder = 'cur';
    carp "Unknown option '$_'" for keys %opts;
    my $fq_folder = "$dir/$folder";
    
    # Acquire a lock on the meta-index counter or croak if another
    # process already has it. Depending on whether the message store
    # allows concurrent indexing or not we either release the lock after
    # we're done with work file manipulation or keep it during the entire
    # course of the indexer opertion, respectively.
    $self->_qmx_lock_acquire($folder);

    # Open the index writable. The underlying index system may be using
    # its own use locks so this is the only API that should open them for
    # writing to avoid blocking. Verify that we can update the indexes
    # by aquiring an opaque handle to the indexer.
    my $idx = $self->_new_index($fq_folder, writable => 1);
    die "Meta-indexing aborted: failed to initialze the index" unless $idx;
    
    # work files effectively contain "index" or "unindex" commands
    my @work_files = $self->_get_work_files($fq_folder, $opts{index_log});

    $self->_qmx_lock_release() if ($self->{'pmxdir_concurrent_ok'});

    unless (@work_files) {
	$self->log("DEBUG", "Exiting meta-index. No pending index transactions.");
	return 0;
    }

    # If the caller wants a progress callback then we need to figure out how 
    # much work there is to do. This is a bit of a performance hit so code that
    # cares should not pass in a callback.
    my $to_do = 0;
    if ($progress_cb) {
	foreach my $fname (@work_files) {
	    open(my $tmp_file, "<$fname")
		or die "Failed to read the file $fname: $!";
	    $to_do++ while <$tmp_file>;
	    close($tmp_file);
	}
    }

    # Open and process each pending transaction. This loop may be aborted by 
    # the callback procedure
    my $count = 0;
    my @processed_work_files;
    eval {
	foreach my $fname (@work_files) {
	    open(my $fh, "<$fname") or 
		die "Can't open index log $fname: $!";
	    while (<$fh>) {
		chomp;
		$count++;
		next unless $_;
		my ($a, $id, $f, @attr) = split;
		eval {
		    if ($a eq "+" && !$opts{no_insert}) {
			if (my $m = $self->message($id)) {
			    $self->_index($m, $idx);
			    $self->log("DEBUG", "Added message id $id to the meta-index in folder $f");
			}
			else {
			    $self->log("WARNING", "Message $id can't be found in folder $f");
			}
		    }
		    elsif ($a eq "-" && !$opts{no_remove}) {
			my $m = PureMessage::MessageStore::Dir::IdxLogMessage->new($id, \@attr);
			$self->_forget($m, $idx);
			$self->log("DEBUG", "Removed the message id $id from the meta-index in folder $f");
		    }
		    else {
			$self->log("ERR", "Invalid index log line, skipping: [$a:$id:$f]");
		    }
		};
		$self->log("ERR", "Error updating an index with message $id: $@") if $@;

		# The callback may throw an exception to abort the outer loop.
		&$progress_cb($count/$to_do) if $progress_cb;
	    }
	    # schedule work file for removal
            if (UNIVERSAL::can($self, '_flush_indexed')) {
                $self->_flush_indexed();
                $self->_remove_files($fname);
            }
            else {
                push @processed_work_files, $fname;
            }
	}

        # cleanup if necessary
	$idx->done() if $idx->can('done');
	$self->done() if $self->can('done');

        # remove successfully processed work files; it's highly desirable
        # that we have the lock while doing that
        eval {$self->_qmx_lock_acquire($folder)};
        if ($@) {
            $self->log("WARNING", "$@");
            carp "$@";
        }
        $self->_remove_files(@processed_work_files);
    };
    if ($@) {
	$self->log("DEBUG", "Aborting the index update loop.: $@");
    }

    $self->_qmx_lock_release();
    return $count;
}

# returns the count of removed files
sub _remove_files {
    my $self = shift;
    my $count;
    while (my $file = shift) {
        unless (unlink("$file")) {
            my $msg = "Failed to unlink file $file: $!";
	    $self->log("ERR", $msg);
            carp $msg;
	    next;
        }
        $count++;
    }
    return $count;
}

# _index() & _forget() are the methods where the index is actually updated
# to add/remove a message.
sub _index {
    my ($self, $msg, $idx) = @_;
    $idx->index($msg);
}

sub _forget {
    my ($self, $msg, $idx) = @_;
    $idx->forget($msg);
}

sub reindex {
    my($self, %opts) = @_;
    my $progress_cb  = delete $opts{'progress'};
    my $add_new      = delete $opts{'add-new'};
    my $forget_old   = delete $opts{'forget-old'} || 0;
    my $field        = delete $opts{'forget-field'};
    my $purge        = delete $opts{'purge'} || 0;
    my $start_id     = delete $opts{'start-id'};
    my $max_messages = delete $opts{'max-messages'};
    my $first        = delete $opts{'first'} || 0;
    my $folder       = delete $opts{'folder'} || "cur";
    my $update_stats = delete $opts{'update-stats'} || 0;
    my $load_direct  = delete $opts{'load-direct'} || 0;
    my $index_type   = delete $opts{'index-type'} || 0;
    my $num_ids = 10000; # how many ids to check at once if forget-old
    carp "Unknown option '$_'" for keys %opts;
    if ( $index_type ) {
	my $old_i_type = $self->_edit_index_type($index_type);
	$self->log("INFO", "Changed indexer type to $index_type")
	    unless ( $index_type eq $old_i_type );
    }
    $add_new ||= !$forget_old;

    # do $add_new || $forget_old || $update_stats

    $self->_qmx_lock_acquire($folder);

    # prepare the db.conf file for editing
    my $dir = $self->dir;
    require PureMessage::MessageStore::Dir::Config;
    my $conf = PureMessage::MessageStore::Dir::Config->new("$dir/db.conf", 
							   writable => 1);
 
    my ($min_id, $max_id, $count, $reverse, $min_found_id, $max_found_id);
    unless ( $self->_folder_name_ok($folder) || ($folder =~ /^all$/i) ) {
	$folder = "cur";
    }
    $count = 0;
    my ($rem_msg, $rem_rec) = (0) x 2;
    $reverse = !$first;
    # set min and max if applicable
    if ( $reverse ) { # default
	if ( defined($start_id) ) {
	    $max_id = $start_id;
	    $min_id = $start_id - $max_messages if defined($max_messages);
	}
    } else
    {
	if ( defined($start_id) ) {
	    $min_id = $start_id;
	    $max_id = $start_id + $max_messages if defined($max_messages);
	}
    }
    # $max_messages may still be defined when $start_id is not
    # update-stats just 
    eval {
	&$progress_cb(0.0) if $progress_cb;

	# figure out how much we have to index
	my @folders = ($folder =~ /^all$/i) ? ($self->folders()) : ($folder);
	# does above get more folders than we really want?
	my $total = 0;
	for my $fldr (@folders) {
	    # run _scan_dir without a callback method
	    my $c = $self->message_count($fldr, $min_id, $max_id);
	    if ( defined($max_messages) ) { # get at most max_messages
		$c = ( $c < $max_messages ) ? $c : $max_messages;
	    }
	    $total += $c;
	}
	# Do we just want to do this for every id from max indexed id -> min id ?
	if ( $forget_old ) {
	    $purge = 1 unless $total; # purge if empty quarantine
	    for my $fldr (@folders) {
		my $last_id;
		my $idx = $self->_new_index($self->_foldername($fldr),
					    truncate => $purge,
					    writable => 1);
		# get the count of messages indexed
		# get a filename from an id
		my $cb = sub { $self->_filename($fldr, shift); };

		# min/max, first and reverse factors?
		unless ( $purge ) {
		    ($rem_msg, $rem_rec) = $idx->forget_old($cb, $progress_cb, 
							    $min_id, $max_id);
		}
		undef $idx;
	    }
	    &$progress_cb(1.0) if $progress_cb;
	} elsif ( $add_new ) {
	    if ($total) {
		# time to create the indices
		$self->message_writable(1);
		for my $fldr (@folders) {
		    my $idx;
		    if ( $update_stats ) { # traverse directory w/o indexing
			$min_id = $max_id = undef;
			$self->log("INFO", "Updating Quarantine Statistics");
		    } 
		    elsif ( $load_direct || $purge ) {
			$idx = $self->_new_index($self->_foldername($fldr),
						 truncate => $purge,
						 writable => 1);
			undef($idx) unless (  $load_direct && $idx->indexed_fields );
		    } 
                    my $map;
		    eval { $map = $self->get_alias_map(1) };
		    $self->log("WARNING", "Unable to retrieve aliases: $@")
                        if $@;
		    if ( $add_new || $update_stats ) {
			# add all the messages in the current folder to the index
			# for all message ids > $min_id and < $max_id
			my $global_id_count = 0;
			$self->message_scan(sub {
			    my $msg = shift;
			    return if ( defined($max_messages) and $count >= $max_messages );
			    my $id = $msg->id;
                            my $global_id = $self->_new_global_id($id);
			    $msg->minfo_value("Global-ID", $global_id);

			    # The EUI requires Mapped-Recipients be set
			    if ($map and
                                     !scalar $msg->minfo_value("Mapped-Recipients") ) {
				$msg->minfo_value("Mapped-Recipients", 
						 map { lc $_ } $map->transform($msg->minfo_value("Envelope-To")));
			    }
			    
			    $min_found_id = $id if !$min_found_id || $id < $min_found_id;
			    $max_found_id = $id if !$max_found_id || $id > $max_found_id;

			    my $store_meta = 'yes';
			    $store_meta = lc($msg->minfo_value('Store-Meta'))
				if defined $msg->minfo_value('Store-Meta');

			    if ($fldr ne 'cur' or ($store_meta eq 'yes')) {
				$self->_idx_log_insert($fldr, $id) 
				    unless ($load_direct || $update_stats);
				$idx->index($msg) if ( $load_direct && $idx );
			    }

			    $count++;
			    &$progress_cb($count/$total) if $progress_cb;
			}, $fldr, $min_id, $max_id, $reverse);
			$idx->done() if ( $load_direct && $idx );
		    }
		}
	    } else { # no messages in quarantine
		# purge database if requested
		for my $fldr (@folders) {
		    my $idx = $self->_new_index($self->_foldername($fldr),
						truncate => $purge,
						writable => 1);
		    $min_found_id = $max_found_id = 0;
		}
		# there are no messages to index, just signal that we are done
		&$progress_cb(1.0) if $progress_cb;
	    }
	}
    };
    if ($@) {
	$self->log("WARNING", $@);
    }
    else {
	$conf->value(first_id => $min_found_id) if $min_found_id;

        # we _never_ want to decrement the counter...
        my $curr_last_id = $conf->value('last_id');
        $conf->value(last_id  => $max_found_id)
            if ($max_found_id and ($curr_last_id < $max_found_id));

        $conf->value(num_msg  => $count);
    }

    # also update disk usage information
    # this could be optimized a bit - may be slow !!!
    if ( 1 || $update_stats ) { # enabled for now
	require ActiveState::DiskUsage;
	require Cwd;
	$conf->value("disk_used", ActiveState::DiskUsage::du(Cwd::realpath($dir)) - 
		     ActiveState::DiskUsage::du("$dir/tmp"));
    }
    $conf->save;
    undef($conf);

    # $max_found_id is assumed to be the last id
    if ($max_found_id and $load_direct) {
        $self->_qmx_lock_release($max_found_id);
    } 

    $self->log("INFO", "Reindexed $count messages") if $add_new;;
    $self->log("INFO", "Removed $rem_rec records from $rem_msg messages") if $forget_old;

    return ($count || $rem_msg);
}

# Record the insertion of a message into a folder in an index log. meta_index()
# uses this information to update the meta-data indexes. 
sub _idx_log_insert {
    my $self   = shift;
    my $folder = shift || croak "Missing required folder parameter";
    my $id     = shift || "-";
 
    $folder =~ s,.*/,,;
    return if ($folder ne 'cur');

    my $log = $self->dir . "/$folder/index_log";
    PureMessage::_sys::append($log, "+ $id $folder\n");
}

# Record the removal of a message from a folder in an index log. In addition
# to logging the id this logs all the fields that this message has indexed.
# This is because the message file may have been removed by the time the
# data is to be removed from the index and the index needs this data in order
# to find the id's to remove.
sub _idx_log_remove {
    my $self   = shift;
    my $folder = shift || croak "Missing required folder field";
    my $id     = shift || "-";
    my $msg    = shift || croak "Missing required message field";

    $folder =~ s,.*/,,;

    return if ($folder ne 'cur');
    
    eval {
	my @fields = ($self->_indexed_fields($folder), "m_global_id");
	my $attr;
	if (ref($msg)) {
	    $attr = _encode_msg_attr($msg, @fields);
	}
	else {
	    $msg = PureMessage::MessageFile->new($msg);
	    $attr = _encode_msg_attr($msg, @fields);
	}
	$self->log("DEBUG", "No attribute values for message $id") 
	    unless $attr;
	$attr = "-" unless $attr;

	my $log = $self->dir . "/$folder/index_log";
	PureMessage::_sys::append($log, "- $id $folder $attr\n");
    };
    $self->log("ERR", "Error recording the removal of message $id from folder $folder: $@") if $@;
}

sub _encode_msg_attr {
    my ($msg, @fields) = @_;
    my @kv;
    foreach (@fields) {
	my $f = $_;
	my @vals = FIELDS->field($msg, $_);
	push(@kv, $f, $_) foreach @vals;
    }
    return join_line(undef, @kv);
}

sub _indexed_fields {
    my ($self, $folder) = @_;
    return PureMessage::MessageStore::Dir::Index::get_indexed_fields($self, $folder);
}


# This returns the path to the message file that corresponds to an id.  This is
# (unfortunately?) needed since there is no way to update the minfo in a message
# file directly in pmdb-mode, so we occasionally need to create a
# PureMessage::MessageFile for the message file.
sub id_to_filename {
    my $self = shift; 
    my $id = shift;
    my $folder = shift || 'cur';
    $id = $self->local_from_global($id) unless $self->id_is_local($id);
    
    return $self->dir . "/$folder/" . $self->_id_fname($id);
}


sub _id_fname {
    my $self = shift; 
    my $id = shift;
    my @dir;
    my $tmp = $id;
    while (length $tmp) {
        push(@dir, substr($tmp, -3, 3, ""));
    }
    shift(@dir);
    my $len = length($id);
    if ($len > 9) {
	$len -= 10;
	die "Can't map $id to filename (too big)"
	    if $len >= 26;
	$len = chr(ord('a') + $len);
    }
    join("/", $len, reverse(@dir), $id);
}

sub _filename {
    my($self, $folder, $id) = @_;
    join("/", $self->{pmxdir_dir}, $folder, $self->_id_fname($id));
}

sub _foldername {
    my($self, $folder) = @_;
    join("/", $self->{pmxdir_dir}, $folder);
}

sub _mk_path {
    my $self = shift;
    require File::Path;
    File::Path::mkpath(@_);
}

sub message {
    my $self = shift;
    return map scalar($self->message($_)), @_ if wantarray;

    die "Can't return multiple messages in scalar context" if @_ > 1;
    my $id = shift;
    $id = $self->local_from_global($id) unless $self->id_is_local($id);
    require PureMessage::MessageStore::Dir::Message;
    my $m = PureMessage::MessageStore::Dir::Message->new($self, $id, "cur",
						    $self->_filename("cur", $id),
						    $self->{pmxdir_writable},
						    );
    return $m if $m;

    for my $f ($self->folders) {
	next if $f eq "cur";  # already tried;
	$m = PureMessage::MessageStore::Dir::Message->new($self, $id, $f,
						     $self->_filename($f, $id),
						     $self->{pmxdir_writable},
						     );
	return $m if $m;
    }
    return undef;
}

sub message_scan {
    my($self, $cb, $folder, $min_id, $max_id, $reverse) = @_;

    $folder = "cur" unless defined $folder;
    croak("Illegal folder name '$folder'")
	unless $self-> _folder_name_ok($folder);

    my $dir = $self->dir . "/" . $folder;
    croak("No such folder '$folder'")
	unless -d $dir;

    require PureMessage::MessageStore::Dir::Message;

    $min_id ||= 0;
    return _scan_dir($dir, $cb, $folder, $self, $min_id, $max_id, $reverse);
}

sub message_writable {
    my $self = shift;
    my $old = $self->{pmxdir_writable};
    $self->{pmxdir_writable} = shift if @_;
    $old;
}

sub message_count {
    my $self = shift;
    return $self->message_scan(undef, @_);
}

# always true for file-system based index
sub is_message_body_inserted { return 1; }
sub rule_hits_inserted { return 1; }

sub _scan_dir {
    my($dir, $cb, $folder, $store, $min, $max, $reverse) = @_;
    #print "SCAN DIR $dir\n";
    return 0 if ($dir =~ /\bfi_\w+$/);
    my $count = 0;
    if (opendir(my $d, $dir)) {
	# The number of files here cannot exceed 1000.  See _id_fname()
	# and spec/3.0/qdir.html for why.
	my @files = sort grep /^\d+\z/ || /^[a-z]\z/, readdir($d);
	closedir($d);
	@files = reverse @files if $reverse;
	if (@files) {
	    if (-d "$dir/$files[0]") {
		for my $d (@files) {
		    $count += _scan_dir("$dir/$d", $cb, $folder, $store, $min, $max, $reverse);
		}
	    }
	    elsif ($cb) {
		for my $f (@files) {
		    if ($reverse) {
			next if defined($max) &&
			        (length($f) > length($max) ||
                                 (length($f) == length($max) && $f gt $max));
			last if length($f) < length($min) ||
			        (length($f) == length($min) && $f lt $min);
		    }
		    else {
			next if length($f) < length($min) ||
			        (length($f) == length($min) && $f lt $min);
			last if defined($max) &&
			        (length($f) > length($max) ||
                                 (length($f) == length($max) && $f gt $max));
		    }
		    if (my $m = PureMessage::MessageStore::Dir::Message->new($store, $f, $folder, "$dir/$f", $store->{pmxdir_writable})) {
			&$cb($m);
			$count++;
		    }
		    else {
			$store->log("ERR", "Can't open message $f: $@");
		    }
		}
	    }
	    else {
		$count += @files;
	    }
	}
    }
    return $count;
}

sub _folder_name_ok {
    my $self = shift;
    my $f = shift;
    $f eq "cur" || $f eq "trash" || $f =~ /^[A-Z][-\w]*\z/;
}

sub folders {
    my $self = shift;
    my $dir = $self->dir;
    opendir(my $d, $dir) || die "Can't opendir '$dir': $!";
    my @folders = grep $self->_folder_name_ok($_), readdir($d);
    closedir($d);
    @folders = sort @folders;
    return @folders;
}

sub empty_trash {
    my $self = shift;
    my $dir = $self->dir;
    my $trash = "$dir/trash";

    if (-d $trash) {
	# first remove all the messages and decrement usage counts
	$self->message_scan(sub {
	    my $m = shift;
	    _unlink_message_file($self, $m->filename);
	}, "trash");

	# then blow away any indexes that might be left
	require File::Path;
        File::Path::rmtree($trash, $self->{pmxdir_verbose}, 1);
    }
}

sub _unlink_message_file {
    my $self = shift;
    require ActiveState::DiskUsage;
    for my $f (@_) {
	my $du = ActiveState::DiskUsage::du($f);
	unlink($f) || die "Can't unlink '$f': $!";
	$self->{pmxdir_disk_used_delta} -= $du;
	$self->{pmxdir_num_msg_delta}--;
    }
}

sub tidy {
    my $self = shift;
    my $dir = $self->dir;
    for my $f ($self->folders) {
	my $dir = $self->_foldername($f);
	_tidy($dir, $self->{pmxdir_verbose});
    }
    my $day = 60 * 60 * 24;
    $self->_cleanup_old_sto("$dir/mset", $day);
}

sub _cleanup_old_sto {
    my ($self, $dir, $expiry) = @_;
    my $current_time = time;
    File::Find::find({ 
        no_chdir => 1,
        wanted => sub {
                if ($File::Find::name =~ /\.sto$/) {
                    my $atime = (stat($File::Find::name))[8] || $current_time;
                    if ($atime + $expiry < $current_time) {
                        unlink $File::Find::name or
                           warn "Cannot remove $File::Find::name: $!\n";
                    }
                }
        }, 
    }, $dir );
}

sub _tidy {
    my($dirname, $verbose) = @_;
    opendir(my $dir, $dirname) or return 0;
    my @files = readdir($dir);
    closedir($dir);

    my $tidy = 1;
    for my $f (@files) {
	next if $f eq "." || $f eq "..";
	$tidy = 0 unless _tidy("$dirname/$f", $verbose);
    }
    if ($tidy && $dirname =~ m,/\d+$,) {
	print "Removing $dirname\n" if $verbose;
	return 1 if rmdir($dirname);
	warn "Can't delete $dirname: $!";
    }
    return 0;
}

sub message_set {
    my $self = shift;
    require PureMessage::MessageStore::Dir::MessageSet;
    PureMessage::MessageStore::Dir::MessageSet->new(verbose => $self->{pmxdir_verbose},
					       @_,
					       message_store => $self);
}

sub counter {
    require PureMessage::MessageStore::Dir::Counter;
    PureMessage::MessageStore::Dir::Counter->new(@_);
}

sub counters {
    require PureMessage::MessageStore::Dir::Counter;
    PureMessage::MessageStore::Dir::Counter->Counters(@_);
}

sub status {
    my $self = shift;
    my %opts = @_;
    my $folder = delete($opts{folder}) || "cur";
    my %status;
    
    my $dir = $self->dir;
    $status{"quarantine_directory"} = $dir;
    my $idx = $self->_new_index("$dir/$folder");
    if ($idx->isa("PureMessage::MessageStore::Dir::Index::cdb") ) {
	$status{"quarantine_indexer"} = "cdb";
	$status{"quarantine_messages_indexed"} = $idx->get_count();
	$status{"quarantine_report"} = sub { $idx->gen_report(flush => 2) } if $opts{full};
    }

    require ActiveState::DiskUsage;
    require ActiveState::Unix::DiskFree;

    require PureMessage::MessageStore::Dir::Config;
    my $conf = PureMessage::MessageStore::Dir::Config->new("$dir/db.conf");
    for (qw(first_id last_id num_msg disk_used)) {
	if (defined(my $x = $conf->value($_))) {
	    $status{$_} = $x;
	}
    }

    if (exists $status{num_msg}) {
	$status{num_msg} += $self->{pmxdir_num_msg_delta};
    }

    if (exists $status{disk_used}) {
	$status{disk_used} += ActiveState::DiskUsage::du("$dir/tmp");
	$status{disk_used} += $self->{pmxdir_disk_used_delta};
    }

    $status{disk_free} = ActiveState::Unix::DiskFree::df($dir)->{free};
    $status{num_msg}  = 0 unless (defined $status{num_msg});
    $status{first_id} = 0 unless (defined $status{first_id});
    $status{last_id}  = 0 unless (defined $status{last_id});

    return \%status;
}

sub _status_cache_sync {
    my $self = shift;
    return unless $self->{pmxdir_disk_used_delta} || $self->{pmxdir_num_msg_delta};

    require PureMessage::MessageStore::Dir::Config;
    my $dir = $self->dir;
    my $conf = PureMessage::MessageStore::Dir::Config->new("$dir/db.conf", writable => 1);

    if (my $delta = $self->{pmxdir_disk_used_delta}) {
	$conf->value("disk_used", $conf->value("disk_used") + $delta);
	$self->{pmxdir_disk_used_delta} = 0;
    }
    if (my $delta = $self->{pmxdir_num_msg_delta}) {
	$conf->value("num_msg", $conf->value("num_msg") + $delta);
	$self->{pmxdir_num_msg_delta} = 0;
    }
    $conf->save;
}

sub get_alias_map {
    my $self = shift;
    my $refresh = shift;
    require PureMessage::Map;
    $self->{alias_map} = PureMessage::Map->new('recipient-aliases')
        unless $self->{alias_map};
    my $map = $self->{alias_map};
    $map->refresh() if $refresh;
    return $map;
}

# misc global/local id manipulation routines

# non-digit character that separates local id from location part in the
# global id.
sub _separator { return '-'; }

sub _new_global_id {
    my $self = shift;
    my $local_id = shift;
    return join _separator(), $local_id, $self->{pmxdir_location_id};
}

sub id_is_local {
    my $id = shift;
    $id = shift if ref $id;

    # local ids consist of digits only
    return 1 if $id =~ /^\d+\z/;
    return 0;
}

sub id_is_global {
    my $id = shift;
    $id = shift if ref $id;

    # two numbers separated by the separator
    return 1 if $id =~ /^\d+-\d+\z/;
    return 0;
}

sub local_from_global {
    my $id = shift;
    $id = shift if ref $id;
    $id =~ s/[^\d].*//;
    return $id;
}

# constructs global id from the specified local id and numeric location id
sub global_from_local {
    my $local_id_part = shift;
    $local_id_part = shift if ref $local_id_part;
    my $location_part = shift;
    return "$local_id_part" . _separator() . "$location_part";
}

# returns the location part of the global message identifier
sub location_part {
    my $id = shift;
    $id = shift if ref $id;
    $id =~ s/[\d]+[^\d]//;
    return $id;
}

# Returns the value that the specified global key can be sorted on so
# that we get the order of quarantining (more or less).
# In this module we're dealing mostly with local context, so the sort key
# can be the numeric local identifier.
sub sort_key {
    my $self = shift;
    my $id = shift;
    return $self->local_from_global($id);
}

package PureMessage::MessageStore::Dir::IdxLogMessage;
use strict;
use PureMessage::MessageLog qw(decode_value);
use PureMessage::MessageInfo;

our $AUTOLOAD;

sub new {
    my $p = shift;
    my $class = ref($p) || $p;
    my $id = shift;
    my $attr = shift;
    my %headers;
    my %computed_fields;
    my @m_info;
    foreach my $kv (@$attr) {
	my ($k, $v) = split('=', $kv, 2);
	$v = decode_value($v);
	if ($k =~ /^h_/) {
	    if (exists $headers{$k}) {
		if (ref $headers{$k}) {
		    push @{$headers{$k}}, $v;
		}
		else {
		    $headers{$k} = [$headers{$k}, $v];
		}
            }
	    else {
		$headers{$k} = $v;
	    }
	}
	elsif ($k =~ /^m_/) {
	    push(@m_info, $k, $v);
	}
	else {
	    if (exists $computed_fields{$k}) {
		if (ref $computed_fields{$k}) {
		    push @{$computed_fields{$k}}, $v;
		}
		else {
		    $computed_fields{$k} = [$computed_fields{$k}, $v];
		}
            }
	    else {
		$computed_fields{$k} = $v;
	    }
	}
    }
    my $minfo = PureMessage::MessageInfo->new(@m_info);
    return bless {
	id => $id, 
	h => \%headers, 
	c => \%computed_fields, 
	m => $minfo
	}, $class;

}

sub header {
    my ($self, $header) = @_;
    $header =~ s/-/_/g; # reverse Fields changes
    $header = "h_$header" unless $header =~ /^h_/;
    my $v = $self->{h}{$header};
    if (ref($v)) {
        return @$v if wantarray;
        return $v->[0];
    }
    return $v;
}

sub minfo_value {
    my ($self, $field) = @_;
    $field =~ s/-/_/g; # reverse Fields changes
    $field = "m_$field" unless $field =~ /^m_/;
    return $self->{m}->value($field);
}

# Use AUTOLOAD to implement the computed fields functions
AUTOLOAD {
    my $self = shift;
    return if $AUTOLOAD =~ /::DESTROY$/;
    $AUTOLOAD =~ /::(c_\w+)$/;
    my $v = $self->{c}{$1};
    if (ref($v)) {
	return @$v if wantarray;
	return $v->[0];
    }
    return $v;
}

sub id { shift->{id} }

1;

__END__

=head1 NAME

PureMessage::MessageStore::Dir - File system storage

=head1 DESCRIPTION

The directory format is inspired by the I<maildir> format
designed for I<qmail>; <http://cr.yp.to/proto/maildir.html>.  There is
a single file per message, and each of these files live in
sub-directories of the quarantine directory.

A quarantine directory will contain at least the following
sub-directories.  There might be more:

=over

=item F<tmp>

This directory is used for temporary files.  Messages that might
potentially be stored are collected here first.  If the message
is to be stored, then it renamed so that it ends up in the F<new>
directory.  If not it is unlinked.  Old files left in the F<tmp>
directory can be safely deleted.

=item F<new>

Messages that are to be stored by PureMessage end up here.
The file name is ugly and long to make sure different milter
processes generate unique names.

=item F<cur>

When C<pmx-qman> starts up it moves new messages to this folder.  At
the same time it renames the messages so that the file names are plain
sequential integer numbers (like mh uses).

=back

=head2 Message file format

There is a single file per stored message.  The file contains the raw
e-mail message containing headers and a body, followed by the meta
information.

The meta information is a plain text block appended to the
message.  The size of this block is determined by reading the last 5
bytes of the file.  The last of these 5 bytes will be "\n".  It can be
ignored.  The 4 before should be interpreted as an ASCII hex number.

The internal format of the meta information block is
similar to that of email headers.  The first line contains the string
"-- \n" and should be ignored.

=head1 COPYRIGHT

Copyright (c) 2002 ActiveState Corp.  All rights reserved.

=head1 SEE ALSO

L<PureMessage::MessageStore>

=cut
