package PureMessage::MessageStore::Dir::DB;

use strict;
use Carp qw(carp croak);

use Fcntl qw(LOCK_EX LOCK_NB O_CREAT O_RDWR);
use CDB_File;
use Time::HiRes qw (usleep);

# The purpose of this module is to provide support for 
# CDB databases to PureMessage::MessageStore::Dir::Index.pm

# flush to cdb where there are more than this many ids in memdb
my $MAXCOUNT_MDB = $ENV{PMX_MAXCOUNT_MDB} || 16383;
my $MAXCOUNT_CDB = $ENV{PMX_MAXCOUNT_CDB} || 82000;
my $MAXCOUNT_CDB10 = $ENV{PMX_MAXCOUNT_CDB10} || 500000;
my $MAXAGE = 0.002; # stale age (days) a file must be older than this process
my $MAXWAITTIME = 1000000; # max time to wait to flush when done
my $WAITTIME = 1000; # microseconds
my %HIDDEN = ( "__COUNT__" => 1, "__MSGIDS__" => 1 ); # hidden keys

sub LEXICAL { $a cmp $b }; # lexical sorting
sub NUMERIC { $a <=> $b }; 
sub DBG ()  { 0 }; # if we are debugging

sub new {
    my($class, %opts) = @_;

    my $flush    = delete $opts{flush} || 0;
    my $truncate = delete $opts{truncate};
    my $writable = delete $opts{writable} || 0;

    my $self = {
        dir      => delete $opts{dir},
	db_name  => delete $opts{db_name},
	verbose  => delete $opts{verbose},
	rdonly   => !$writable,
	ordered  => delete $opts{ordered},
	lex_val  => delete $opts{lex_val} || 0,
	num_key  => delete $opts{num_key} || 0,
	newcount => 0,
	oldcount => 0,
	cdbcountrem => 0,
        msgidsrem   => 0,
	cdbcount    => 0,
        msgids      => 0,
	cdbcount10  => 0,
        msgids10    => 0,
	cdbcount100 => 0,
        msgids100   => 0,
    };
    #print STDERR "RDONLY = $self->{rdonly}\n";

    if ($^W) {
	for (keys %opts) {
	    carp("Unrecognized option '$_'");
	}
    }

    bless $self, $class;

    # zero all databases? 
    my @all = (1) x 7;
    $self->_truncate_db(@all) if $truncate;
    # recent, removed, cdb_removed, cdb, cdb10, cdb100, clean_all_tempfiles
    # check for broken index files?
    
    $self->_get_db([1,1,1,1]); # keep filehandles?

    $self->flush(1, 0) if ($flush == 1); # load to big db
    $self->flush(1, 1) if ($flush  > 1); # be sure to flush
    $self->flush(1, 1) if ($flush  > 2); # sure flush twice?
    
    unless ( $self->{rdonly} ) {
	delete $self->{cdbrem};
	delete $self->{cdb};
	delete $self->{cdb10};
	delete $self->{cdb100};
	delete $self->{lock};
	# Adjust file sizes for big quarantines
	if ( $self->{msgids100} > 5000000 ) {
	    $MAXCOUNT_CDB10 *= 2;
	    if ( $self->{msgids100} > 10000000 ) {
		$MAXCOUNT_CDB10 *= 2;
		$MAXCOUNT_CDB *= 2;
		$MAXCOUNT_MDB *= 2 + 1;
	    }
	}
    }

    return $self;
}

sub DESTROY {
    my $self = shift;

    # if we are killed without cleaning up, try to clean up
    $self->done() if ( $self->{newcount} || $self->{oldcount} );
}

sub done {
    my $self = shift;
    return if $self->{rdonly};
    my $wait = 0;
    # try hard to save everything to cdb
    while ( ($self->{newcount} || $self->{oldcount}) ) {
	print "$$ flushing -> old=" .$self->{oldcount} . " new=" .  $self->{newcount} . " ---\n"
	    if $self->{verbose};
	$self->flush(0, 1); # may take $MAXWAITTIME
	if ( $self->{newcount} || $self->{oldcount} ) {
	    usleep $WAITTIME;
	    $wait += $WAITTIME;
	}
	# note: wait may be as much as forever
    }
    # throw an error if we can't save everything to cdb!
    if ( $self->{newcount} || $self->{oldcount} ) {
	my (@newmsg, @oldmsg);
	@newmsg = $self->{recent}->get_range() if $self->{newcount}; # get min and max values
	@oldmsg = $self->{removed}->get_range() if $self->{oldcount};
	my $err = "";
	$err .= "Couldn't save details for message range [@newmsg]\n" if @newmsg;
	$err .= "Couldn't delete messages for range [@oldmsg]\n" if @oldmsg;
	warn $err;
	print STDERR "WARNING ONLY $$: IS DYING without saving DATA!\n" . $err . "old=" .$self->{oldcount} . 
	    " new=" .  $self->{newcount} . " ---\n";
    }
    # clean up any temp files?
    $self->_truncate_db(!$self->{newcount}, !$self->{oldcount}, 0, 0, 0, 0, 0, 1);
    
    # drop all tie's ;)
    delete $self->{recent};
    delete $self->{removed};
    delete $self->{cdbrem};
    delete $self->{cdb};
    delete $self->{cdb10};
    delete $self->{cdb100};
    delete $self->{lock};
}

# return the number of message ids in the cdb's
sub get_count {
    my $self = shift;
    my $count = shift;
    my ($keycount, $msgcount) = (0, 0);
    $self->_get_db([1,1,1,1]);
    if ( defined($count) ) {
	return 0 unless ( $count =~ /^(cdbcount|msgids)/ );
	$keycount = $self->{$count};
	$msgcount = $self->{$count};
    } else {
	$keycount = ($self->{cdbcount100} + $self->{cdbcount10} + $self->{cdbcount}
		     - $self->{cdbcountrem});
	$msgcount =  ($self->{msgids100} + $self->{msgids10} + $self->{msgids} 
		      - $self->{msgidsrem});
    }
    #print STDERR $self->{db_name} . " has $keycount, $msgcount\n" . $self->{cdbcount100} . "\n";
    # block while there is a lock or tmp file?
    return wantarray ? ($keycount, $msgcount) : $msgcount;
}

# Assumes keys are integers
# return an array ref to a sorted list of missing keys
sub find_missing {
    my $self = shift;
    die "$$ Cannot find missing non-numeric keys" unless $self->{num_key};
    die "$$ Cannot find missing keys for index with lexical values" 
	if $self->{lex_val};
    my @missing = ();
    my $indx = 'cdb100';
    my $idx = $self->{$indx};
    $self->_get_db([0,0,0,1]);
    my $rdonly =  $self->{rdonly};
    my $max_tries = 2;
    $self->{rdonly} = 0; # if you do this are a writer now!
    while ( $max_tries && ($self->{cdbcount10} || $self->{cdbcount} || 
			   $self->{cdbcountrem}) ) {
	$max_tries--;
	$self->flush(1, 1); # push everything into cdb100
    }
    $self->{rdonly} = $rdonly; # restore reader status if req'd
    # get first (least) key
    my $f_key = $idx->FIRSTKEY();
    my $l_key = $f_key;
    my $n_key = defined($f_key) ? $idx->NEXTKEY($f_key) : undef;
    return ([], 0) unless ( defined($f_key) && defined($n_key) and not
			    (exists($HIDDEN{$f_key}) || exists($HIDDEN{$n_key})) );
    # iterate through keys
    while ( defined($n_key) and not exists($HIDDEN{$n_key}) ) {
	my $m_key = $l_key + 1;
	while ( $m_key < $n_key ) {
	    push @missing, $m_key;
	    $m_key++;
	}
	die "$$ has non-integer keys -> $m_key != $n_key" unless ($m_key == $n_key);
	$l_key = $n_key;
	$n_key = $idx->NEXTKEY($l_key);
    }

    return (\@missing, $l_key - $f_key) ;
}

# you should call flush before calling this
# count the number of sortedness breaks in keys and vals
sub chk_ordr {
    my ($self, $indx) = @_;
    my($ooky, $ooid, $dupky, $dupid) = (0) x 4;
    my($key, $lkey, $v, @v);
    $self->_get_db([1,1,1,1]);

    $indx = 'cdb100' 
	unless ( defined($indx) && ($indx =~ /^cdb(rem|10|100)?$/) );
    $indx = $self->{$indx};
    #($key, $v) = _get_first_keyval($indx);
    #$lkey = $key;
    #@v = ();
    # if we have lexical values then we have repeated keys by design
    if ( $self->{lex_val} ) {
	($key, $lkey, $v) = $self->_get_first_multi_keyvals($indx);
	$dupky += $#{$v};
	#$| = 1;
	#print "We have lexical values for ", $self->{db_name}, "\n";
	while ( defined($key) && !exists($HIDDEN{$key}) ) {
	    #($key, $v) = _get_next_keyval($indx, $lkey);
	    #last unless defined($key);
	    #if ( exists($HIDDEN{$key}) ) { $lkey = $key; next; }
	    ($key, $lkey, $v) = $self->_get_next_multi_keyvals($self->{cdb100}, $key);
	    #last unless @v;
	    $ooky++ if ( !exists($HIDDEN{$key}) && 
			 _le($key, $lkey, $self->{num_key}) );
	    $dupky += $#{$v};

	    ($ooid, $dupid) = $self->chk_ids($v, $ooid, $dupid);
	}
    } else {
	($key, $v) = _get_first_keyval($indx);
	return ( (0) x 4 ) unless ( defined($key) && !exists($HIDDEN{$key}) );
	$lkey = $key;
	@v =  length($v) ? unpack("w*", $v) : ();
	($ooid, $dupid) = $self->chk_ids(\@v, $ooid, $dupid);
	
	while ( defined($key) ) {
	    ($key, $v) = _get_next_keyval($indx, $lkey);
	    last unless defined($key);
	    if ( exists($HIDDEN{$key}) ) { $lkey = $key; next; }
	    $ooky++ unless _gt($key, $lkey, $self->{num_key});
	    $dupky++ if _eq($key, $lkey, $self->{num_key});
	    $lkey = $key;
	    @v = length($v) ? unpack("w*", $v) : (); # $self->{lex_val}?
	    ($ooid, $dupid) = $self->chk_ids(\@v, $ooid, $dupid);
	}
    }
    
    return ($ooky, $ooid, $dupky, $dupid);
}

sub chk_ids {
    my($self, $vals, $ooval, $dupval) = @_;

    while ( $#{$vals} ) {
	my $lv =  shift(@{$vals});
	last unless @{$vals};
	$ooval++ unless _lt($lv, $vals->[0], !$self->{lex_val});
	$dupval++ if _eq($lv, $vals->[0], !$self->{lex_val});
    }

    return ($ooval, $dupval);
}

# walk through index and return [ { val => [ key, ... ] }, ...]
sub get_multi_vals {
    my ($self, $vals) = @_; # val is probably an id
    my $result = [];
    my($key, $lkey, $v, @v);
    my $indx = 'cdb100';
    my $max_tries = 2;
    while ( $max_tries && ($self->{cdbcount10} || $self->{cdbcount} || 
			   $self->{cdbcountrem}) ) {
	$max_tries--;
	$self->flush(1, 1); # push everything into cdb100
    }
    $self->_get_db([0,0,0,1]);

    if ( $self->{lex_val} ) {
	foreach my $val ( @{$vals} ) {
	    my @vkey = ();
	    ($key, $lkey, $v) = $self->_get_first_multi_keyvals($self->{$indx});
	    $lkey =~ s/\A___/__/; # un-escape keys!
	    push @vkey, $lkey if ( grep $_ eq $val, @{$v} );
	    #$| = 1;
	    #print "We have lexical values for ", $self->{db_name}, "\n";
	    while ( defined($key) && !exists($HIDDEN{$key}) ) {
		($key, $lkey, $v) = $self->_get_next_multi_keyvals($self->{$indx}, $key);
		$lkey =~ s/\A___/__/; # un-escape keys!
		push @vkey, $lkey if ( grep $_ eq $val, @{$v} );
	    }
	    push @{$result}, { $val => \@vkey };
	}
    } else {
	foreach my $val ( @{$vals} ) {
	    my @vkey = ();
	    ($key, $v) = _get_first_keyval($self->{$indx});
	    unless ( defined($key) && !exists($HIDDEN{$key}) ) {
		push @{$result}, { $val => [] };
		next;
	    }
	    $lkey = $key;
	    @v =  length($v) ? unpack("w*", $v) : ();
	    $lkey =~ s/\A___/__/; # un-escape keys!
	    push @vkey, $lkey if ( grep $_ == $val, @v );
	    
	    while ( defined($key) ) {
		($key, $v) = _get_next_keyval($self->{$indx}, $lkey);
		last unless defined($key);
		if ( exists($HIDDEN{$key}) ) { $lkey = $key; next; }
		$lkey = $key;
		@v = length($v) ? unpack("w*", $v) : (); # $self->{lex_val}?
		$lkey =~ s/\A___/__/; # un-escape keys!
		push @vkey, $lkey if ( grep $_ == $val, @v );
	    }
	    push @{$result}, { $val => \@vkey };
	}
    }

    return $result;
}

# get all key/vals for the keys provided
sub get_multi_keys {
    my ($self, $keys) = @_;
    my $result = []; # [ { key => [ val, ... ] }, ...]
    $self->_get_db([1,1,1,1]);

    foreach my $item ( @{$keys} ) {
	#$item =~ s/\A___/__/; # un-escape keys! ??
	if ( $self->{lex_val} ) {
		push @{$result}, { $item => $self->get_multi_lex($item) };
	} else {
		push @{$result}, { $item => $self->get_multi($item) };
	}
    }

    return $result;
}

# This function is intended to help clean out old keys and values
# you should call flush before calling this
# iterate through all keys returning $count key/vals at a time
# return keys greater than or equal to min and less than max
sub get_keys {
# WARNING: An error may be returned if calls to this are intermixed with call to
# write to this index
    my($self, $count, $nextkey, $min, $max) = @_;
    my $result = [];
    my ($v, $lkey, $key);
    my @val = ();
    $self->_get_db([0,0,0,1]);
    return ($result, undef) unless ( ref($self->{cdb100}) eq "CDB_File" );
    my $c = 0;
    $count ||= 1000;
    # iterate through that cdb returning $count [ key => [vals] ]
    
    if ( !defined($nextkey) ) { # first time through?
	if ( $self->{lex_val} ) {
	    ($key, $lkey, $v) = $self->_get_first_multi_keyvals($self->{cdb100});
	    
	    if ( defined($lkey) && !exists($HIDDEN{$lkey}) ) {
		if ( ((defined($min) && ($lkey >= $min)) || !defined($min)) &&
		     ((defined($max) && ($lkey <  $max)) || !defined($max)) ) {
		    $c++;
		    $lkey =~ s/\A___/__/; # un-escape keys!
		    push @{$result}, { $lkey => $v };
		}
		#$lkey = $key;
		#($key, $v) = $self->_get_key_vals($self->{cdb100}, $key);
	    } else {
		return ($result, undef);
	    }
	} else { # numeric values
	    ($key, $v) = _get_first_keyval($self->{cdb100});
	    
	    if ( defined($key) && !exists($HIDDEN{$key}) ) {
		if ( ((defined($min) && ($lkey >= $min)) || !defined($min)) &&
		     ((defined($max) && ($lkey <  $max)) || !defined($max)) ) {
		    $c++;
		    # unpack $v
		    $lkey = $key;
		    $lkey =~ s/\A___/__/; # un-escape keys!
		    push @{$result}, { $lkey => (length($v) ? [ unpack("w*", $v) ] : []) };
		}
	    } else {
		return ($result, undef);
	    }
	}
    } else {
	$key = $nextkey;
    }

    # spin through to where we need to be
    if ( defined($min) ) {
	while ( defined($key) && !exists($HIDDEN{$key}) && 
		_lt($key, $min, $self->{num_key}) ) {
	    # wouldn't it be faster to just get the keys? (no FETCH)
	    if ( $self->{lex_val} ) {
		($key, $lkey, $v) = $self->_get_next_multi_keyvals($self->{cdb100}, $key);
	    } else {
		($key, $v) = _get_next_keyval($self->{cdb100}, $key);
	    }
	}
    }

    if ( defined($max) ) { 
	while ( ($c < $count) && defined($key) && !exists($HIDDEN{$key}) &&
		_lt($key, $max, $self->{num_key}) ) {
	    if ( $self->{lex_val} ) {
		($key, $lkey, $v) = $self->_get_next_multi_keyvals($self->{cdb100}, $key);
		$c++;
		$lkey =~ s/\A___/__/; # un-escape keys!
		push @{$result}, { $lkey => $v };
	    } else {
		($key, $v) = _get_next_keyval($self->{cdb100}, $key);
		last unless ( defined($key) && !exists($HIDDEN{$key}) );
		$c++;
		$lkey = $key;
		$lkey =~ s/\A___/__/; # un-escape keys!
		push @{$result}, { $lkey => (length($v) ? [ unpack("w*", $v) ] : []) };
	    }
	}
	$key = undef if ( defined($key) && _ge($key, $max, $self->{num_key}) );
    } else {
	# $key is the next key, $lkey is the last key, @$v are the $key values
	while ( ($c < $count) && defined($key) && !exists($HIDDEN{$key}) ) {
	    if ( $self->{lex_val} ) {
		($key, $lkey, $v) = $self->_get_next_multi_keyvals($self->{cdb100}, $key);
		$c++;
		$lkey =~ s/\A___/__/; # un-escape keys!
		push @{$result}, { $lkey => $v };
		#$lkey = $key;
	    } else {
		#$lkey = $key;
		# get next key and its values
		($key, $v) = _get_next_keyval($self->{cdb100}, $key);
		last unless ( defined($key) && !exists($HIDDEN{$key}) );
		$c++;
		$lkey = $key;
		$lkey =~ s/\A___/__/; # un-escape keys!
		push @{$result}, { $key => (length($v) ? [ unpack("w*", $v) ] : []) };
	    }
	}
    }
    $key = undef unless ( defined($key) && !exists($HIDDEN{$key}) );

    return ($result, $key);
}

sub get {
    my $self = shift;
    $self->_get_db([1,1,1,1]);
    # return $self->get_multi(@_);
    if ( $self->{lex_val} ) {
	return $self->get_multi_lex(@_); # change to get_fast
    } else {
	return $self->get_multi(@_); 
    }
}

# get value(s) associated with @$key or $key
#sub get_multi {
sub get_multi {
    my($self, $key) = @_;
    my @rem = ();
    my $rem = "";
    my @v   = ();
    my $v   = "";
    $key =~ s/\A__/___/; # escape hidden keys!

    if ( $self->{msgids100} && 
	 (@v = @{$self->{cdb100}->multi_get($key)}) ) {
	# @v is an array of packed values, each of which is an array
	$v = combine_vals( \@v );
    }
    if ( $self->{msgids10} &&
	 (@v = @{$self->{cdb10}->multi_get($key)}) ) {
	$v = combine_vals( [$v, @v] ); 
    }
    if ( $self->{msgids} &&
	 (@v = @{$self->{cdb}->multi_get($key)}) ) {
	$v = combine_vals( [$v, @v] );
    }
    @v = length($v) ? unpack("w*", $v) : ();

    if ( $self->{msgidsrem} ) {
	$rem = combine_vals( $self->{cdbrem}->multi_get($key) ) || "";
    }
    @rem = length($rem) ? unpack("w*", $rem) : ();
    
    return _and_not_vals(\@v, \@rem);
}

sub get_multi_lex {
    my($self, $key) = @_;
    my @rem = ();
    my $rem = [];
    my @v   = ();
    my $v   = "";
    $key =~ s/\A__/___/; # escape hidden keys!

    if ( $self->{msgids100} ) { 
	$v = combine_lex( [ $self->{cdb100}->multi_get($key) ] );
    }
    if ( $self->{msgids10} ) {
	$v = combine_lex( [$v, $self->{cdb10}->multi_get($key)] ); 
    }
    if ( $self->{msgids} ) {
	$v = combine_lex( [$v, $self->{cdb}->multi_get($key)] );
    }

    if ( $self->{msgidsrem} ) {
	$rem = combine_lex( $self->{cdbrem}->multi_get($key) );
    }
    
    return _and_not_lex($v, $rem);
}

# this assumes no duplicate keys
# get value(s) associated with @$key or $key
sub get_fast {
#sub get {
    my($self, $key) = @_;
    my $rem = [];
    my $val = [];
    my ($v, $lv, @v);
    $key =~ s/\A__/___/; # escape hidden keys!

    $self->_get_db([1,1,1,1]);
    # if we had $self->{lex_val} we wouldn't have packed values
    if ( $self->{msgids100} && ($v = $self->{cdb100}->FETCH($key)) ) {
	$val = [ unpack("w*", $v) ];
    }
    if ( $self->{msgids10}  && ($v = $self->{cdb10}->FETCH($key)) ) {
	if ( @{$val} ) {
	    @v = unpack("w*", $v);
	    $lv = [ shift @{$val}, shift @v ];
	    $val = _ior_vals($lv, [$val, \@v]);
	} else {
	    $val = [ unpack("w*", $v) ];
	}
    }
    if ( $self->{msgids}    && ($v = $self->{cdb}->FETCH($key)) ) {
	# push @val, unpack("w*", $v);
	if ( @{$val} ) {
	    @v = unpack("w*", $v);
	    $lv = [ shift @{$val}, shift @v ];
	    $val = _ior_vals($lv, [$val, \@v]);
	} else {
	    $val = [ unpack("w*", $v) ];
	}
    }

    if ( $self->{msgidsrem} && ($v = $self->{cdbrem}->FETCH($key)) ) {
	$rem = [ unpack("w*", $v) ];
	# for my $rem ( @rem ) { @val = grep $_ != $rem, @val; }
    }

    return _and_not_vals($val, $rem);
}

sub dump {
    my $self = shift;
    my $form = shift;
    my $indx = shift;
    require Data::Dump;
    $| =1; # print right away
    delete $self->{cdb100};
    my $rdonly = $self->{rdonly};

#    print STDERR "My cdb100=", ref($self->{cdb100}), "\n";
    $indx = 'cdb100' 
	unless ( defined($indx) && ($indx =~ /^cdb(rem|10|100)?$/) );
    if ( $indx =~ "cdb100" ) { # flush all db's into cdb100
	$self->{rdonly} = 0; # if you dump you are a writer now!
	my $max_tries = 2;
	while ( $max_tries && ($self->{cdbcount10} || $self->{cdbcount} || 
			       $self->{cdbcountrem}) ) {
	    $max_tries--;
	    $self->flush(1, 1); # push everything into cdb100
	}
    }
    $self->{rdonly} = $rdonly; # become what you were before 

    $self->_get_db([1,1,1,1]);
    $indx = $self->{$indx};
    $form ||= "keyval_dump"; 
    
    # dump all key, values in cdb(s)
    #print("My index is $self->{cdb100} ", ref($self->{cdb100}), "\n");
    return unless (ref($indx) eq 'CDB_File');
    #return; # unless ref($self->{cdb100});
    #_get_first_keyval( $self->{cdb100} );
    #print("My first key/val is ", _get_first_keyval( $self->{cdb100} ), "-\n");
    if ( $self->{lex_val} ) {
	for ( my ($key, $lastk, $val) = $self->_get_first_multi_keyvals($indx);
	      defined($key);
	      ($key, $lastk, $val) = $self->_get_next_multi_keyvals($indx, $key) ) {
	    next if exists($HIDDEN{$key});
	    my $lkey = $key;
	    $lkey =~ s/\A___/__/; # un-escape keys!
	    my $dkey =  Data::Dump::dump($lkey);
	    if ( $form eq "keyval_dump" ) {
		print (" $dkey => ", Data::Dump::dump($val), "\n");
		#$next_val = Data::Dump::dump($val);
	    } elsif ( $form eq "key_dump" ) {
		print ($dkey, "\n");
	    } elsif ( $form eq "key" ) {
		print ($lkey, "\n");
	    } elsif ( $form eq "key_count" ) {
		my $count = scalar(@{$val});
		print ("$key => $count\n");
	    } elsif ( $form eq "keydump_count" ) {
		my $count = scalar(@{$val});
		print ("$dkey => $count\n");
	    } else {
		print "Do not recognize index dump form = $form\n";
	    }
	}
    } else {
	for ( my ($key, $val) = _get_first_keyval($indx);
	      defined($key);
	      ($key, $val) = _get_next_keyval($indx, $key) ) {
	    next if exists($HIDDEN{$key});
	    my $lkey = $key;
	    $lkey =~ s/\A___/__/; # un-escape keys!
	    my $dkey =  Data::Dump::dump($lkey);
	    
	    if ( $form eq "keyval_dump" ) {
		print (" $dkey => [", join(", ", unpack("w*", $val)), "],\n"); 
	    } elsif ( $form eq "key_dump" ) {
		print ($dkey, "\n");
	    } elsif ( $form eq "key" ) {
		print ($lkey, "\n");
	    } elsif ( $form eq "key_count" ) {
		my $count = scalar( my @values = unpack("w*",$val) );
		print ("$key => $count\n");
	    } elsif ( $form eq "keydump_count" ) {
		my $count = scalar( my @values = unpack("w*",$val) );
		print ("$dkey => $count\n");
	    } else {
		print "Do not recognize index dump form = $form\n";
	    }
	}
    }
}

# add @[$key => $val] to the database 
# there may be repeated keys with (hopefully) different values
# does not insert duplicate values with the same key
sub put_multi {
    my($self, $aref) = @_;
    my $result = 0;

    my $i = 0;
    while ( my ($key, $val) = ($aref->[$i], $aref->[$i+1]) ) {
	$result += $self->{recent}->put($key, $val);
	$i += 2;
	last if ( $i > $#{$aref} );
    }
    $self->{newcount} += $result;

    $self->flush(0) if ( $self->{newcount} > $MAXCOUNT_MDB );
    return $result;
}

# add $val to values associated with $key in the (recent) database
# LIMITATION - if $val is in removed db it will not be added #
# need to detect that case and remove it from removed :)
sub put {
    my($self, $key, $val) = @_;
    
    my $result = $self->{recent}->put($key, $val);
    $self->{newcount} += $result;
    $self->{oldcount} -= $self->{removed}->del($key, $val) if $self->{oldcount};
    
    $self->flush(0) if ( $self->{newcount} > $MAXCOUNT_MDB );
    return $result;
}

# delete this value associated with this key from the database
# by adding $val to values associated with $key in the (removed) database
sub del {
    my($self, $key, $val) = @_;

    my $result = $self->{removed}->put($key, $val);
    $self->{oldcount} += $result;
    $self->{newcount} -= $self->{recent}->del($key, $val) if $self->{newcount};

    # do a flush to truncate $self->{removed}
    $self->flush(0) if ( $self->{oldcount} > $MAXCOUNT_MDB );
    return $result;
}

# add $id => [ @val ] to the database 
# does not insert duplicate values with the same key
sub put_rev {
    my($self, $id, $vals) = @_;

    # use the MemDB put_many
    my $result = $self->{recent}->put_many($id, $vals);
    $self->{newcount} += $result;
    $self->{oldcount} -= $self->{removed}->del_many($id, $vals) if $self->{oldcount};

    $self->flush(0) if ( $self->{newcount} > $MAXCOUNT_MDB );
    return $result;
}

# delete $id => [ @val ] from the database 
# does not insert duplicate values with the same key
sub del_rev {
    my($self, $id, $vals) = @_;

    # use the MemDB put_many
    my $result = $self->{removed}->put_many($id, $vals);
    $self->{oldcount} += $result;
    $self->{newcount} -= $self->{recent}->del_many($id, $vals) if $self->{newcount};

    # do a flush to truncate $self->{removed}
    $self->flush(0) if ( $self->{oldcount} > $MAXCOUNT_MDB );
    return $result;
}

# add $key => [ @val ] to the database 
# does not insert duplicate values with the same key
sub put_many {
    my($self, $key, $ids) = @_;

    # use the MemDB put_many
    my $result = $self->{recent}->put_many($key, $ids);
    $self->{newcount} += $result;
    $self->{oldcount} -= $self->{removed}->del_many($key, $ids) if $self->{oldcount};

    $self->flush(0) if ( $self->{newcount} > $MAXCOUNT_MDB );
    return $result;
}

# delete $key => [ @val ] from the database 
# does not insert duplicate values with the same key
sub del_many {
    my($self, $key, $ids) = @_;

    # use the MemDB put_many
    my $result = $self->{removed}->put_many($key, $ids);
    $self->{oldcount} += $result;
    $self->{newcount} -= $self->{recent}->del_many($key, $ids) if $self->{newcount};

    # do a flush to truncate $self->{removed}
    $self->flush(0) if ( $self->{oldcount} > $MAXCOUNT_MDB );
    return $result;
}

sub remove_me {
    my @all = (1) x 7;
    shift->_truncate_db(@all);
}

# get lock - serialize indexers
sub _get_lock {
    my ($self, $lock_file) = @_;
    if ( defined($lock_file) ) {
	$lock_file .= "l";
    } else {
	$lock_file = $self->{dir} . "/" . $self->{db_name} . ".lock";
    }
    my $lock = _open($lock_file, O_CREAT|O_RDWR);
    _flock( $lock, LOCK_EX ) || die "$$: Can't flock '$lock_file': $!";

    return $lock;
}

# get lock - serialize indexers
sub _get_nb_lock {
    my ($self, $lock_file) = @_;
    if ( defined($lock_file) ) {
	$lock_file .= "l";
    } else {
	$lock_file = $self->{dir} . "/" . $self->{db_name} . ".lock";
    }
    my $lock = _open($lock_file, O_CREAT|O_RDWR);

    if ( _flock( $lock, LOCK_EX|LOCK_NB ) ) {
	return $lock;
    }

    return;
}

# empty dbnames provided
sub _truncate_db {
    my($self, $del_new, $del_rem, $del_cdbrem, $del_cdb, 
       $del_cdb10, $del_cdb100, $clean_tmp) = @_;
    my $file = "";
    return if $self->{rdonly};

    if ( $del_new ) { # delete recent
	delete $self->{recent} if defined($self->{recent});
	$self->{newcount} = 0;
    }

    if ( $del_rem ) { # delete removed
	delete $self->{removed} if defined($self->{removed});
	$self->{oldcount} = 0;
    }

    if ( $del_cdbrem ) { # delete cdbrem?
	$file = $self->{dir} . "/" . $self->{db_name} . "_removed.cdb";
	unlink($file) if ( -e $file );
	$self->{cdbcountrem} = 0;
	$self->{msgidsrem} = 0;
	$self->_get_a_cdb($file, 1) or die "Can't replace $file with empty cdb: $!";
	if ( exists($self->{lock}) ) {
	    _close($self->{lock});
	    delete($self->{lock});
	}
    }
    if ( $del_cdb ) { # delete cdb?
	$file = $self->{dir} . "/" . $self->{db_name} . ".cdb";
	unlink($file) if ( -e $file );
	$self->{cdbcount} = 0;
	$self->{msgids} = 0;
	$self->_get_a_cdb($file, 1) or die "Can't replace $file with empty cdb: $!";
	if ( exists($self->{lock}) ) {
	    _close($self->{lock});
	    delete($self->{lock});
	}
    }
    if ( $del_cdb10 ) { # delete cdb10?
	$file = $self->{dir} . "/" . $self->{db_name} . "10.cdb";
	unlink($file) if ( -e $file );
	$self->{cdbcount10} = 0;
	$self->{msgids10} = 0;
	$self->_get_a_cdb($file, 1) or die "Can't replace $file with empty cdb: $!";
	if ( exists($self->{lock}) ) {
	    _close($self->{lock});
	    delete($self->{lock});
	}
    }
    if ( $del_cdb100 ) { # delete cdb100?
	$file = $self->{dir} . "/" . $self->{db_name} . "100.cdb";
	unlink($file) if ( -e $file );
	$self->{cdbcount100} = 0;
	$self->{msgids100} = 0;
	$self->_get_a_cdb($file, 0) or die "Can't replace $file with empty cdb: $!";
	if ( exists($self->{lock}) ) {
	    _close($self->{lock});
	    delete($self->{lock});
	}
    }


    if ( $clean_tmp ) {
	opendir(my $DIR, $self->{dir}) or die "Can't read the $self->{dir} directory: $!";
	my @tfile = grep /\.(?:tmp)\z/, readdir($DIR);
	my @nfile = grep /\.(?:news?cdb)(\.\d+)?\z/, readdir($DIR);
	closedir($DIR);
	for $file ( @tfile ) {
	    $file = $self->{dir} . "/" . $file;
	    my $time_waited = 0;
	    $self->_check_if_broken(0, [$file]);
	    if (-e $file && (-M _ > $MAXAGE) ) {
		unlink($file); 
	    } 
	}
	$self->flush(1, 1) if scalar(@nfile);
    }
}

# get databases - set handles
sub _get_db {
    my $self = shift;
    my $fhdl = shift;           # [cdbrem, cdb, cdb10, cdb100]
    my $keep_lock = shift || 0; # don't release global lock if this is set
    my $db_name   = $self->{db_name};

    # the MemDB_File's should usually have a count of "0" when this is called
    unless ( $self->{rdonly} || defined($self->{recent}) ) {
	$self->{recent}   = PureMessage::MessageStore::Dir::DB::MemDB->new(num_key => $self->{num_key},
									   lex_val => $self->{lex_val},
									   verbose => $self->{verbose},
									   );
	$self->{newcount} = $self->{recent}->{COUNT};
	print("$$: Opened my New MemDB with count ", $self->{newcount}, " -\n") 
	    if  $self->{verbose};
    }
    unless ( $self->{rdonly} || defined($self->{removed}) ) {
	$self->{removed}  = PureMessage::MessageStore::Dir::DB::MemDB->new(num_key => $self->{num_key},
									   lex_val => $self->{lex_val},
									   verbose => $self->{verbose},
									   );
	$self->{oldcount} = $self->{removed}->{COUNT};
	print("$$: Opened my Old MemDB with count ", $self->{oldcount}, " -\n") 
	    if  $self->{verbose};
   }

    if ( $fhdl->[0] && (ref($self->{cdbrem}) ne 'CDB_File') ) {
	my $file  = $self->{dir} . "/" .  $db_name . "_removed.cdb";
	$file .= "r" if $self->{rdonly};
	if ( $self->_get_a_cdb($file, 1) ) {
	    $self->{cdbrem} = CDB_File->TIEHASH($file)
		or die "$$: Can't tie $file $!";
	    if ( exists($self->{lock}) and not $keep_lock ) {
		_close($self->{lock});
		delete($self->{lock});
	    }
	    $self->{cdbcountrem} = $self->{cdbrem}->FETCH('__COUNT__');
	    $self->{msgidsrem}   = $self->{cdbrem}->FETCH('__MSGIDS__');
	} elsif ( $self->{rdonly} ) {
	    $self->{cdbcountrem} = 0;
	    $self->{msgidsrem}   = 0;	    
	} else {
	    die "$$: Can't make an empty cdb file at $file";
	}

	print(("$$: Opened " . $db_name . "_removed.cdb with "), $self->{cdbcountrem}, 
	      " records and ", $self->{msgidsrem}, " ids\n") if $self->{verbose};
    }

    if  ( $fhdl->[1] && (ref($self->{cdb}) ne 'CDB_File') ) {
	my $file  = $self->{dir} . "/" .  $db_name . ".cdb";
	$file .= "r" if $self->{rdonly};
	if ( $self->_get_a_cdb($file, 1) ) {
	    $self->{cdb} = CDB_File->TIEHASH($file)
		or die "$$: Can't tie $file $!";
	    if ( exists($self->{lock}) and not $keep_lock ) {
		_close($self->{lock});
		delete($self->{lock});
	    }
	    $self->{cdbcount} = $self->{cdb}->FETCH('__COUNT__');
	    $self->{msgids}   = $self->{cdb}->FETCH('__MSGIDS__');
	} elsif ( $self->{rdonly} ) {
	    $self->{cdbcount} = 0;
	    $self->{msgids}   = 0;
	} else {
	    die "$$: Can't make an empty cdb file at $file";
	}

	print(("$$: Opened " . $db_name . ".cdb with "), $self->{cdbcount}, 
	      " records and ", $self->{msgids}, " ids\n") if $self->{verbose};
    }

    if ( $fhdl->[2] && (ref($self->{cdb10}) ne 'CDB_File') ) {
	my $file = $self->{dir} . "/" .  $db_name . "10.cdb";
	$file .= "r" if $self->{rdonly};
	if ( $self->_get_a_cdb($file, 1) ) {
	    $self->{cdb10} = CDB_File->TIEHASH($file)
		or die "$$: Can't tie $file $!";
	    if ( exists($self->{lock}) and not $keep_lock ) {
		_close($self->{lock});
		delete($self->{lock});
	    }
	    $self->{cdbcount10} = $self->{cdb10}->FETCH('__COUNT__');
	    $self->{msgids10}   = $self->{cdb10}->FETCH('__MSGIDS__');
	} elsif ( $self->{rdonly} ) {
	    $self->{cdbcount10} = 0;
	    $self->{msgids10}   = 0;
	} else {
	    die "$$: Can't make an empty cdb file at $file";
	}

	print(("$$: Opened " . $db_name . "10.cdb with "), $self->{cdbcount10}, 
	      " records and ", $self->{msgids10}, " ids\n") if $self->{verbose};
    }

    if ( $fhdl->[3] && (ref($self->{cdb100}) ne 'CDB_File') ) {
	my $file  = $self->{dir} . "/" .  $db_name . "100.cdb";
	if ( $self->_get_a_cdb($file, 0) ) {
	    $self->{cdb100} = CDB_File->TIEHASH($file)
		or die "$$: Can't tie $file $!";
	    if ( exists($self->{lock}) and not $keep_lock ) {
		_close($self->{lock});
		delete($self->{lock});
	    }
	    $self->{cdbcount100} = $self->{cdb100}->FETCH('__COUNT__');
	    $self->{msgids100}   = $self->{cdb100}->FETCH('__MSGIDS__');
	} elsif ( $self->{rdonly} ) {
	    $self->{cdbcount100} = 0;
	    $self->{msgids100}   = 0;
	} else {
	    die "$$: Can't make an empty cdb file at $file";
	}

	print(("$$: Opened " . $db_name . "100.cdb with "), $self->{cdbcount100}, 
	      " records and ", $self->{msgids100}, " ids\n") if $self->{verbose};
    }
}

# make sure we have a cdb file - make an empty one if not!
# $safe is true if we have the critical section lock already
sub _get_a_cdb {
    my ($self, $dbname, $mkreadlink) = @_;

    my $tmpcdb = $dbname . ".tmp";
    my ($rdbn, $safe);
    return (-e $dbname) if $self->{rdonly};

    $rdbn = $dbname . "r";
    $safe = exists($self->{lock}) && $self->{lock};

    # get critical section lock?
    $self->{lock} = $self->_get_lock() unless $safe;

    if ( -e $tmpcdb ) {
	$self->_check_if_broken(1, [$tmpcdb]);
	
	if ( -e $tmpcdb ) {
	    return -e $dbname;
	}
    }
    if ( -e $dbname ) {

	link($dbname, $rdbn) unless ( !$mkreadlink or -e $rdbn );

	return (1); # true if the db exists
    }

    my $cdb = CDB_File->new($dbname, $tmpcdb) 
	or die "Can't create new database $tmpcdb $!";
    $cdb->insert("__COUNT__", 0);
    $cdb->insert("__MSGIDS__", 0);
    $cdb->finish or die "$0: CDB_File finish failed for $dbname: $!";
    $cdb = undef; # drop database reference
    link($dbname, $rdbn) unless ( !$mkreadlink or -e $rdbn );
    unless ( $safe ) {
	_close($self->{lock}) unless $safe;
	delete($self->{lock});
    }
    return (-e $dbname); # extra system call?
}

# this subroutine expects only 1 process is modifying the index
sub _check_if_broken {
    my ($self, $safe, $files) = @_;
    # try to flock files and delete
    $self->{lock} = $self->_get_lock() unless $safe;

    print "$$: checking for broken files in: @{$files}\n" if $self->{verbose};
    foreach my $file ( @{$files} ) {
	if ( -e $file ) {
	    my $lock;
	    my $lockfile = $file . "l";
	    if ( ($lock =_open($lockfile, O_CREAT|O_RDWR)) && 
		 _flock($lock, LOCK_EX|LOCK_NB) ) {
		_close($lock); 
	        unlink($file); 
		print "$$: $file was broken - deleted XXX\n" if $self->{verbose};
	    } else {
		print "$$: $file NOT broken - NOT deleting\n" if $self->{verbose};
	    }
	}
    }
    unless ( $safe ) {
	_close($self->{lock});
	delete($self->{lock});
    }
    return;
}

# check if a given table is stale (can be flocked)
# this is ONLY called in a critical section
sub _check_if_stale {
    my ($self, $file) = @_;
    # try to flock files and delete
    #return if $self->{rdonly};
    my $lock;
    print "$$: checking for stale file= $file\n" if $self->{verbose};

    if ( -e $file ) {
	my $lockfile = $file . "l";
	if ( ($lock = _open($lockfile, O_CREAT|O_RDWR)) && 
	     _flock($lock, LOCK_EX | LOCK_NB) ) {
	    print "$$: $file was stale - including ZZZ\n" if $self->{verbose};
	    return($lock);
	} else {
	    print "$$: $file NOT stale - ignoring\n" if $self->{verbose};
	}
    }
    return;
}

sub flush {
    # update cdb's  - we won't worry about big values for now
    my($self, $fullflush, $tryhard) = @_;
    my $dbname = $self->{db_name};
    my $cdbremfile    = $self->{dir} . "/" . $dbname . "_removed.cdb";
    my $newcdbremfile = $self->{dir} . "/" . $dbname . "_removed.newcdb";
    my $tmpcdbrem     = $self->{dir} . "/" . $dbname . "_removed.cdb" . ".tmp";
    my $newcdbfile    = $self->{dir} . "/" . $dbname . ".newcdb";
    my $newcdbfile2   = $self->{dir} . "/" . $dbname . ".newscdb";
    my $newcdbfile10  = $self->{dir} . "/" . $dbname . "10.newcdb";
    my $cdbfile       = $self->{dir} . "/" . $dbname . ".cdb";
    my $tmpcdb        = $self->{dir} . "/" . $dbname . ".cdb" . ".tmp";
    my $cdbfile10     = $self->{dir} . "/" . $dbname . "10.cdb";
    my $tmpcdb10      = $self->{dir} . "/" . $dbname . "10.cdb" . ".tmp";
    my $cdbfile100    = $self->{dir} . "/" . $dbname . "100.cdb";
    my $tmpcdb100     = $self->{dir} . "/" . $dbname . "100.cdb" . ".tmp";
    my $cdb10rlock    = $self->{dir} . "/" . $dbname . "10.cdb" . "u";
    my $cdb100rlock   = $self->{dir} . "/" . $dbname . "100.cdb" . "u";
    my($count, $msgids, $cdbrem, $cdb, $newcdb);

    return if $self->{rdonly};
    # check if we need to do anything first!
    $fullflush = 0 unless ( $self->{cdbcount} || $self->{cdbcount10} || 
			    $self->{cdbcountrem} );
    return unless ( $self->{newcount} or $self->{oldcount} or $fullflush );
    
    if ( $self->{newcount} ) {

	if ( my $lk = $self->_get_nb_lock() ) { ### critical section lock
	    $self->{lock} = $lk;
	} else {
	    return;
	}

	# if cdb->finish was called but its still not updated - wait (or return?)
	my $time_waited = 0;

	### critical section
	while ( -e $tmpcdb ) {
	    # check if its broken?
	    if ( $tryhard and ($time_waited < $MAXWAITTIME) ) {
		usleep $WAITTIME;
		$time_waited += $WAITTIME;
		print "$$: is WAITING for $tmpcdb\n" if $self->{verbose};
	    } else {
		$self->_check_if_broken(1, [$tmpcdb]);
		if ( -e $tmpcdb ) {
		    print "$$: is NOT-WAITING for $tmpcdb - NOT flushing\n" if $self->{verbose};
		    _close($self->{lock});
		    delete($self->{lock});
		    print "$$: closed lock\n" if $self->{verbose};
		    return;
		}
		last;
	    }
	}
	# ASSERT: $tmpcdb does NOT exist

	# get the latest handle on this
	delete($self->{cdb});
	$self->_get_db([0,1,0,0], 1);

	my $lock;
	unless ( $lock = $self->_get_nb_lock($tmpcdb) ) {
	    _close($self->{lock});
	    delete($self->{lock});
	    print "$$: closed lock at line __LINE__ but - NOT flushing\n" 
		if $self->{verbose};
	    return;
	}

	# now we flush $self->{recent} to $self->{cdb}
	$cdb = CDB_File->new($cdbfile, $tmpcdb)
	    or die "Can't create new database $tmpcdb $!";
	
	### critical section end
	_close($self->{lock});
	delete($self->{lock});
	print "$$: closed lock\n" if $self->{verbose};

	($count, $msgids) = (0, 0);
	# for each key in cdb, merge vals with vals in _recent, remove vals

	# ASSERT $self->{cdb} exists and at least one key ("__COUNT__") is defined
	# ASSERT $self->{recent} is defined and is a MemDB reference
	# ASSERT $self->{removed} is defined and is a MemDB reference

	my ($k, $thisk, $v);
	if ( $self->{lex_val} ) {
	    ($k, $thisk, $v)   = $self->_get_first_multi_keyvals($self->{cdb});
	    $thisk = undef if exists($HIDDEN{$thisk})
	} else {
	    ($thisk, $v)   = _get_first_keyval($self->{cdb});
	}

	my ($kr, $vr) = _get_first_keyval($self->{recent});

	while( defined($thisk) || defined($kr) ) {

	    #print("Opened memdb with ", $self->{newcount}, " ids\n");
	    while ( defined($thisk) && exists($HIDDEN{$thisk}) ) { # $k is NOT an empty key
		#print STDERR ("I've seen $k before\n");
		if ( $self->{lex_val} ) {
		    my $lostk; # eq __COUNT__ or __MSGIDS__
		    ($thisk, $lostk, $v) = $self->_get_next_multi_keyvals($self->{cdb}, $thisk);
		} else {
		    ($thisk, $v) = _get_next_keyval($self->{cdb}, $thisk);
		}
	    }
	    #print STDERR ("Now kr is $kr\n") if defined($kr);

	    # add $self->{recent} keys that precede any $self->{cdb} keys # <-- lt
	    if ( defined($kr) && !exists($HIDDEN{$kr}) && 
		    (!defined($thisk) || _lt($kr, $thisk, $self->{num_key})) ) {
		
		if ( @{$vr} ) {
		    if ( $self->{lex_val} ) {
			foreach my $vl ( @{$vr} ) {
			    $cdb->insert($kr, $vl);
			}
		    } else {
			$cdb->insert($kr, pack("w*", @{$vr}));
		    }
		    $count++;
		    $msgids += scalar(@{$vr});
		    #print STDERR ("Now Inserted $msgids msgids for key=$kr number $count\n");
		}
		($kr, $vr) = _get_next_keyval($self->{recent}, $kr);
	    } elsif ( defined($thisk) ) {
		# print("Saving cdb key $k\n");
		# ASSERT $kr ge $k if defined($kr), 
		my $vals  = [];
		if ( $self->{lex_val} ) {
		    $vals = $v;
		    ($k, $thisk, $v) = $self->_get_next_multi_keyvals($self->{cdb}, $thisk);
		    #($k, $v) = _get_next_keyval($self->{cdb}, $k); # get next cdb key, val
		} else {
		    $vals  = [ unpack("w*", $v) ];
		    ($k, $v) = _get_next_keyval($self->{cdb}, $thisk); # get next cdb key
		    # get all values with this key from $self->{cdb}
		    while ( defined($k) && !exists($HIDDEN{$k}) && 
			    _eq($thisk, $k, $self->{num_key}) ) { # check for dups?
			my @newv = unpack("w*", $v);
			my $lvl = [shift @{$vals}, shift @newv];
			$vals = $self->{lex_val} ? _ior_lex($lvl, [$vals, \@newv]) : 
			                           _ior_vals($lvl, [$vals, \@newv]);
			($k, $v) = _get_next_keyval($self->{cdb}, $k); # get next cdb key
		    }
		} # $vals has all values from $self->{cdb} where $thisk eq $k

		# get all values in $self->{recent} with this key
		while ( defined($kr) && !exists($HIDDEN{$kr}) &&
			_eq($thisk, $kr, $self->{num_key}) ) {
		    my $lvl = [shift @{$vals}, shift @{$vr}];
		    $vals = $self->{lex_val} ? _ior_lex($lvl, [$vals, $vr]) : 
			                       _ior_vals($lvl, [$vals, $vr]);
		    ($kr, $vr) = _get_next_keyval($self->{recent}, $kr);
		}

		if ( @{$vals} ) {
		    if ( $self->{lex_val} ) {
			foreach my $vl ( @{$vals} ) {
			    $cdb->insert($thisk, $vl);
			}
		    } else {
			$cdb->insert($thisk, pack("w*", @{$vals}));
		    }
		    $count++;
		    $msgids += scalar(@{$vals});
		    
		    #print("Now Inserted $msgids msgids for key=$thisk number $count\n");
		}
		$thisk = $k;
	    }
	}
	
	$self->{newcount} = 0; # all the keys are in the cdb
	$self->{cdbcount} = $count;
	$self->{msgids}   = $msgids;
	$cdb->insert("__COUNT__", $count);
	$cdb->insert("__MSGIDS__", $msgids);
	$cdb->finish or die "$0: CDB_File finish failed for $dbname: $!";
	undef($self->{cdb}); # drop reference ?
	# update the readlink for this file unless its undergoing consolidation
	if ( my $lk1 = $self->_get_nb_lock($cdb100rlock) ) {
	    if ( my $lk2 = $self->_get_nb_lock($cdb10rlock) ) {
		$self->update_readlink($cdbfile, 1);
		_close($lk2);
	    }
	    _close($lk1);
	}

	undef($cdb);
	_close($lock);

	# we CANNOT remove things from $self->{cdbrem} unless they are 
	# NOT IN ANY cdb - truncate recent and NOT removed NOT cdb 
	# (and reset database handles)
	delete $self->{recent};
	$self->_get_db([0,0,0,0]); # get new handles?
    }
    
    if ( $self->{oldcount} ) {
	
	if ( my $lk = $self->_get_nb_lock() ) {
	    $self->{lock} = $lk;
	} else {
	    return;
	}	

	# if cdb->finish was called but its still not updated - wait (or return?)
	my $time_waited = 0;

	### critical section
	while ( -e $tmpcdbrem ) {
	    # check_if _broken ?
	    if ( $tryhard and ($time_waited < $MAXWAITTIME) ) {
		usleep $WAITTIME;
		$time_waited += $WAITTIME;
	    } else {
		$self->_check_if_broken(1, [$tmpcdbrem]);
		if ( -e $tmpcdbrem ) {
		    print "$$: is TIRED-OF-WAITING for $tmpcdbrem - NOT flushing\n" if $self->{verbose};
		    _close($self->{lock});
		    delete($self->{lock});
		    print "$$: closed rem lock\n" if $self->{verbose};		    
		    return;
		}
		last;
	    }
	}
	# get the latest handle on this
	delete($self->{cdbrem});
	$self->_get_db([1,0,0,0], 1);

	my $lock;
	unless ( $lock = $self->_get_nb_lock($tmpcdbrem) ) {
	    _close($self->{lock});
	    delete($self->{lock});
	    print "$$: closed lock but - NOT flushing\n" if $self->{verbose};
	    return;
	}
	# ASSERT: $tmpcdbrem does NOT exist
	$cdbrem = CDB_File->new($cdbremfile, $tmpcdbrem) 
	    or die "Can't create new database $tmpcdbrem $!";

	### critical section end
	_close($self->{lock});
	delete($self->{lock});
	print "$$: closed rem lock\n" if $self->{verbose};

	($count, $msgids) = (0, 0);
	
	# for each key in cdbrem, merge vals with vals in _remove
	my ($k, $v)   = _get_first_keyval($self->{cdbrem});
	my ($kd, $vd) = _get_first_keyval($self->{removed});

	while ( defined($k) || defined($kd) ) {

	    # ignore __COUNT__ and __MSGIDS__ keys
	    while ( defined($k) && exists($HIDDEN{$k}) ) { # $k is NOT an empty key
		#print("I've seen $k before\n");
		($k, $v) = _get_next_keyval($self->{cdbrem}, $k);
	    }
	
	    # add $self->{removed} keys that precede any $self->{cdbrem} keys
	    while ( defined($kd) && !exists($HIDDEN{$kd}) && 
		    (!defined($k) || _lt($kd, $k, $self->{num_key})) ) {

		if ( @{$vd} ) {
		    if ( $self->{lex_val} ) {
			foreach my $vl ( @{$vd} ) {
			    #print $tmp "MemDB inserting $kd, $vl\n";
			    $cdbrem->insert($kd, $vl);
			}
		    } else {
			$cdbrem->insert($kd, pack("w*", @{$vd}));
		    }
		    $count++;
		    $msgids += scalar(@{$vd});
		    #print("Now Inserted $msgids msgids for key=$kd number $count\n");
		}
		($kd, $vd) = _get_next_keyval($self->{removed}, $kd);
	    }

	    # claim $kd ge $k if defined $kd
	    if ( defined($k) ) {

		# print("Saving cdb key $k\n");
		# ASSERT $kd ge $k if defined($kd)
		my $thisk = $k;
		my $vals = [];
		if ( $self->{lex_val} ) {
		    while ( defined($k) && !exists($HIDDEN{$k}) && 
			    _eq($k, $thisk, $self->{num_key}) ) {
			push @{$vals}, $v;
			($k, $v) = _get_next_keyval($self->{cdbrem}, $k); # get next cdb key, val
		    }
		} else {
		    $vals  = length($v) ? [ unpack("w*", $v) ] : [];
		    ($k, $v) = _get_next_keyval($self->{cdbrem}, $k); # get next cdb key, vals
		    # get all values with this key from $self->{cdbrem}
		    while ( defined($k)  && !exists($HIDDEN{$k}) && 
			    _eq($thisk, $k, $self->{num_key}) ) { 
			my $newv = [ unpack("w*", $v) ]; # $v not empty
			my $lvl = [shift @{$vals}, shift @{$newv}];
			$vals = $self->{lex_val} ? _ior_lex($lvl, [$vals, $newv]) : 
			                           _ior_vals($lvl, [$vals, $newv]);
			($k, $v) = _get_next_keyval($self->{cdbrem}, $k); # get next cdb key
		    }
		} # $vals has all values from $self->{cdbrem} where $thisk eq $k

		# make sure we get all values with this key from $self->{removed} (MemDB)
		while ( defined($kd) && !exists($HIDDEN{$kd}) && 
			_eq($thisk, $kd, $self->{num_key}) ) { 
		    my $lvl = [shift @{$vals}, shift @{$vd}];
		    $vals = $self->{lex_val} ? _ior_lex($lvl, [$vals, $vd]) :
			                       _ior_vals($lvl, [$vals, $vd]);
		    ($kd, $vd) = _get_next_keyval($self->{removed}, $kd);
		} # $vals has all values from $self->{removed} where $thisk eq $kd
		
		if ( @{$vals} ) {
		    if ( $self->{lex_val} ) {
			foreach my $vl ( @{$vals} ) {
			    #print $tmp "inserting $thisk, $vl\n";
			    $cdbrem->insert($thisk, $vl);
			}
		    } else {
			$cdbrem->insert($thisk, pack("w*", @{$vals}));
		    }
		    $count++;
		    $msgids += scalar(@{$vals});
		    # print("Now Inserted $msgids msgids for key=$k number $count\n");
		}
	    }
	}
	#close($tmp);
	$self->{oldcount} = 0; # all the keys are in cdbrem
	$self->{cdbcountrem} = $count;
	$self->{msgidsrem}   = $msgids;
	$cdbrem->insert("__COUNT__", $count);
	$cdbrem->insert("__MSGIDS__", $msgids);
	$cdbrem->finish or die "$0: CDB_File finish failed for $dbname: $!"; 
	undef($self->{cdbrem}); # drop reference ?
	# update the readlink for this file unless its undergoing consolidation
	if ( my $lk1 = $self->_get_nb_lock($cdb100rlock) ) {
	    $self->update_readlink($cdbremfile, 1);
	    _close($lk1);
	}

	undef($cdbrem);
	_close($lock); 
	
	# truncate removed and reset database handles
	delete $self->{removed};
	# don't know if next 2 lines necessary or not
	$self->_get_db([0,0,0,0]); # get new handles
    }

    # build new $cdbfile100 and clean out everything else?
    if ( ($self->{msgids10} > $MAXCOUNT_CDB10) or 
	 (($self->{msgids} > $MAXCOUNT_CDB) && ($self->{msgids100} < $MAXCOUNT_CDB10)) or
	 ($self->{msgidsrem} > $MAXCOUNT_CDB) or $fullflush ) {

	# block anyone else from updating the readlinks
	my $lk100;
	if ( my $lk2 = $self->_get_nb_lock($cdb10rlock) ) {
	    unless ( $lk100 = $self->_get_nb_lock($cdb100rlock) ) {
		_close($lk2);
		print "$$: closed lock but - NOT flushing cdb100 - CANNOT lock readlink!\n" 
		    if $self->{verbose};
		return;
	    }
	    _close($lk2);
	} else {
		print "$$: closed lock but - NOT flushing cdb100 - CANNOT lock readlink!\n" 
		    if $self->{verbose};
		return;
	}

	if ( my $lk = $self->_get_nb_lock() ) {
	    $self->{lock} = $lk;
	} else {
	    _close($lk100);
	    return;
	}

	### critical section start

	my $time_waited = 0;
	# make sure cdbrem, cdb, cdb10 and cdb100 are not being rebuilt
	for my $tmpfile ( $tmpcdbrem, $tmpcdb, $tmpcdb10, $tmpcdb100 ) {
	    # check if its stale?
	    while ( -e $tmpfile ) {
		if ( $tryhard and ($time_waited < $MAXWAITTIME) ) {
		    usleep $WAITTIME;
		    $time_waited += $WAITTIME;
		} else {
		    $self->_check_if_broken(1, [$tmpfile]);
		    if ( -e $tmpfile ) {
			_close($self->{lock});
			delete($self->{lock});
			_close($lk100);
			print "$$: closed lock but - NOT flushing cdb100\n" if $self->{verbose};
			return;
		    }
		    $time_waited = 0;
		    last;
		}
	    }
	}
	# ASSERT: $tmpcdbrem, $tmpcdb, $tmpcdb10, $tmpcdb100 do NOT exist

	# if -s $cdbfile100 > $MAXFILESIZE rename it and get a new empty one
	# Need to ensure this new file is read and deleted from when necessary

	# idea for fast checking for intersection?

	# drop handles
	delete $self->{cdbrem};
	delete $self->{cdb};
	delete $self->{cdb10};
	delete $self->{cdb100};

	my $lock;
	unless ( $lock = $self->_get_nb_lock($tmpcdb100) ) {
	    _close($self->{lock});
	    delete($self->{lock});
	    _close($lk100);
	    print "$$: closed lock but - NOT flushing cdb100\n" 
		if $self->{verbose};
	    return;
	}
	$newcdb = CDB_File->new($cdbfile100, $tmpcdb100) 
	    or die "Can't create new database $tmpcdb100 $!";

	# check for stale files
	opendir(my $DIR, $self->{dir}) or die "Can't read the $self->{dir} directory: $!";
	my @ifile = map { "$self->{dir}/$_" } readdir($DIR);
	closedir($DIR);
	my @file = grep /(?:$newcdbfile|$newcdbfile10|$newcdbfile2)(?:\.\d+)?\z/, @ifile;
	my @remfile = grep /$newcdbremfile(?:\.\d+)?\z/, @ifile;
	my @lock;
	my $newfiles = [];
	my $remfile;
	foreach my $f ( @file ) {
	    if ( my $l = $self->_check_if_stale($f) ) {
		push @lock, $l;
		push @{$newfiles}, $f;
	    }
	}
	foreach my $f ( @remfile ) {
	    if ( my $l = $self->_check_if_stale($f) ) {
		push @lock, $l;
		$remfile = $f;
		last;
	    }
	}
	my @update = (0,0,0);
	my $tmplocks = [];
	# rename files and get list of files to consolidate from
	if ( $self->{cdbcountrem} and not $remfile ) { 
	    if ( my $lk = $self->_get_lock("$newcdbremfile.$$") ) {
		$update[0] = 1; # we are updating $self->{cdbrem}
		push @{$tmplocks}, $lk;
		$remfile = "$newcdbremfile.$$";
		link($cdbremfile, "$newcdbremfile.$$") 
		    or die "Can't link $cdbremfile to $newcdbremfile.$$: $!";
		print STDERR " $$ READONLY cdb trying to flush!" if $self->{rdonly};
		$self->_get_a_cdb($cdbremfile . "tp", 0) or die "$$: Can't get $cdbremfile\"tp\": $!";
		rename($cdbremfile . "tp", $cdbremfile)
		    or die "Can't rename $cdbremfile\"tp\" to $newcdbremfile.$$: $!";
		unlink($cdbremfile . "tp");
	    }
	}
	if ( $self->{cdbcount} ) { # consolidate cdb?
	    $update[1] = 1; # we are updating $self->{cdb}
	    if ( grep /$newcdbfile/, @{$newfiles} ) { # if file is stale
		if ( my $lk =  $self->_get_nb_lock("$newcdbfile.$$") ) {
		    link($cdbfile, "$newcdbfile.$$") 
			or die "Can't link $cdbfile to $newcdbfile.$$: $!";
		    $self->_get_a_cdb($cdbfile . "tp", 0) 
			or die "Can't get $cdbfile.\"tp\": $!";
		    rename($cdbfile . "tp", $cdbfile) 
			or die "Can't rename $cdbfile.\"tp\" to $cdbfile: $!";
		    push @{$tmplocks}, $lk;
		    push @{$newfiles}, ("$newcdbfile.$$");
		} else {
		    #$update[1] = 0; # update is being done on stale file
		    #die "Can't get lock on $newcdbfile.$$: $!";
		}
	    } elsif ( -e $newcdbfile ) { # file is NOT stale
		$update[1] = 0; # another process is doing an update with this file
	    } else {
		if ( my $lk =  $self->_get_nb_lock($newcdbfile) ) {
		    link($cdbfile, "$newcdbfile") 
			or die "Can't link $cdbfile to $newcdbfile: $!";
		    $self->_get_a_cdb($cdbfile . "tp", 0) 
			or die "Can't get $cdbfile\"tp\": $!";
		    rename($cdbfile . "tp", $cdbfile) 
			or die "Can't rename $cdbfile\"tp\" to $cdbfile: $!";
		    push @{$tmplocks}, $lk;
		    push @{$newfiles}, $newcdbfile;
		} else {
		    $update[1] = 0;
		    #die "Can't get lock on $newcdbfile: $!";
		}
	    }
	    unlink($cdbfile . "tp");
	}
	if ( $self->{cdbcount10} ) {
	    $update[2] = 1; # we are updating $self->{cdb10}
	    if ( grep /$newcdbfile10/, @{$newfiles} ) { # if file is stale
		if ( my $lk =  $self->_get_lock("$newcdbfile10.$$") ) {
		    link($cdbfile10, "$newcdbfile10.$$") 
			or die "Can't link $cdbfile10 to $newcdbfile10.$$: $!";
		    $self->_get_a_cdb($cdbfile10 . "tp", 0) 
			or die "Can't get $cdbfile10\"tp\": $!";
		    rename($cdbfile10 . "tp", $cdbfile10) 
			or die "Can't rename $cdbfile10\"tp\" to $cdbfile10: $!";
		    push @{$tmplocks}, $lk;
		    push @{$newfiles}, ("$newcdbfile10.$$");
		} else {
		    #$update[2] = 0;
		    #die "Can't get lock on $newcdbfile10.$$: $!";
		}
	    } elsif ( -e $newcdbfile10 ) {
		$update[2] = 0; # another process is doing an update with this file
	    } else {
		if ( my $lk =  $self->_get_lock($newcdbfile10) ) {
		    link($cdbfile10, $newcdbfile10) 
			or die "Can't link $cdbfile10 to $newcdbfile10: $!";
		    $self->_get_a_cdb($cdbfile10 . "tp", 0) 
			or die "Can't get $cdbfile10\"tp\": $!";
		    rename($cdbfile10 . "tp", $cdbfile10)
			or die "Can't rename $cdbfile10\"tp\" to $cdbfile10: $!";
		    push @{$tmplocks}, $lk;
		    push @{$newfiles}, $newcdbfile10;
		} else {
		    $update[2] = 0;
		    #die "Can't get lock on $newcdbfile10: $!";
		}
	    }
	    unlink($cdbfile10 . "tp");
	}

	### critical section end
	_close($self->{lock});
	delete($self->{lock});
	print "$$: closed lock\n" if $self->{verbose};
	
	unshift @{$newfiles}, $cdbfile100;
	consolidate($self, $newcdb, $newfiles, $remfile, $tmplocks);

	#$self->update_readlink("$newcdbremfile.$$");
	undef($newcdb);
	if ( defined($remfile) ) {
	    unlink($remfile); # or die "Can't delete $newcdbremfile.$$: $!";
	    unlink($remfile . "l");
	}
	foreach my $l ( @lock ) {
	    _close($l);
	}
	@lock = ();
	if ( $update[0] or $update[1] or $update[2] ) {
	    $self->{lock} = $self->_get_lock(); ### critical section
	    $self->update_readlink($cdbremfile, 1) if $update[0];
	    $self->update_readlink($cdbfile, 1)    if $update[1];
	    $self->update_readlink($cdbfile10, 1)  if $update[2];
	    _close($self->{lock}); ### critical section end
	    delete($self->{lock});
	}
	_close($lock); 
	_close($lk100);

	$self->_get_db([0,0,1,1]); # get handles - update counts?
	delete $self->{cdb10};
	delete $self->{cdb100};

    } elsif ( $self->{msgids} > $MAXCOUNT_CDB ) { # we need no fh 
	
	# block anyone else from updating the readlinks
	my $lk10;
	unless ( $lk10 = $self->_get_nb_lock($cdb10rlock) ) {
	    print "$$: NOT flushing cdb10 CANNOT lock readlink!\n" 
		if $self->{verbose};
	    return;
	}

	# make sure cdb and cdb10 are not being rebuilt
	if ( my $lk = $self->_get_nb_lock() ) {
	    $self->{lock} = $lk;
	} else {
	    _close($lk10);
	    return;
	}

	### critical section start

	my $time_waited = 0;
	for my $tmpfile ( $tmpcdb, $tmpcdb10 ) {
	    # check if its stale?
	    while ( -e $tmpfile ) {
		if ( $tryhard and ($time_waited < $MAXWAITTIME) ) {
		    usleep $WAITTIME;
		    $time_waited += $WAITTIME;
		} else {
		    $self->_check_if_broken(1, [$tmpfile]);
		    if ( -e $tmpfile ) {
			_close($self->{lock});
			delete($self->{lock});
			_close($lk10);
			print "$$: closed lock but - NOT flushing cdb10\n" 
			    if $self->{verbose};
			return;
		    }
		    $time_waited = 0;
		    last;
		}
	    }
	}

	# drop handles
	delete $self->{cdbrem};
	delete $self->{cdb};
	delete $self->{cdb10};
	delete $self->{cdb100}; # don't need this right now

	my $lock;
	unless ( $lock = $self->_get_nb_lock($tmpcdb10) ) {
	    _close($self->{lock});
	    delete($self->{lock});
	    _close($lk10);
	    print "$$: closed lock but - NOT flushing cdb10\n" 
		if $self->{verbose};
	    return;
	}
	$newcdb = CDB_File->new($cdbfile10, $tmpcdb10) 
	    or die "Can't create new database $tmpcdb10 $!";

	opendir(my $DIR, $self->{dir}) or die "Can't read the $self->{dir} directory: $!";
	my @ifile =  map { "$self->{dir}/$_" } readdir($DIR);
	closedir($DIR);
	my @file = grep /$newcdbfile2(?:\.\d+)?\z/, @ifile;
	my @lock;
	my $newfiles = [];
	foreach my $f ( @file ) {
	    if ( my $l = $self->_check_if_stale($f) ) {
		push @lock, $l;
		push @{$newfiles}, $f;
	    }
	}
	my $update = 0;
	my $tmplocks = [];
	# rename files - _get_lock
	if ( $self->{cdbcount} ) {
	    $update = 1;
	    my $tmpcdbfile = $cdbfile . "tp";
	    if ( grep /$newcdbfile2/, @{$newfiles} ) { # if file is stale
		if ( my $lk = $self->_get_nb_lock("$newcdbfile2.$$") ) {
		    # get a copy of this file
		    link($cdbfile, "$newcdbfile2.$$") 
			or die "Can't link $cdbfile to $newcdbfile2.$$: $!";
		    # create a new empty one for writers to see
		    $self->_get_a_cdb($tmpcdbfile, 0) 
			or die "Can't get $tmpcdbfile: $!";
		    # overwrite this file for writers, though readers should still
		    # be using the read link which won't be updated
		    rename($tmpcdbfile, $cdbfile)
			or die "Can't rename $tmpcdbfile to $cdbfile: $!";
		    push @{$tmplocks}, $lk;
		    push @{$newfiles}, "$newcdbfile2.$$";
		} else {
		    # $update = 0; # we still update the stale file
		}
	    } elsif ( -e $newcdbfile2 ) { # if another process is updating with this file
		$update = 0;
	    } else {
		if ( my $lk = $self->_get_nb_lock($newcdbfile2) ) {
		    link($cdbfile, $newcdbfile2) 
			or die "Can't link $cdbfile to $newcdbfile2: $!";
		    $self->_get_a_cdb($tmpcdbfile, 0)
			or die "Can't get $tmpcdbfile: $!";
		    rename($tmpcdbfile, $cdbfile)
		    or die "Can't rename $tmpcdbfile to $cdbfile: $!";
		    push @{$tmplocks}, $lk;
		    push @{$newfiles}, $newcdbfile2;
		} else {
		    $update = 0;
		}
	    }
	    unlink($tmpcdbfile);
	}

	### critical section end
	_close($self->{lock});
	delete($self->{lock});
	print "$$: closed lock\n" if $self->{verbose};

	unshift @{$newfiles}, $cdbfile10;
	consolidate($self, $newcdb, $newfiles, undef, $tmplocks);
	undef($newcdb); # drop handle

	if ( $update ) {
	    if ( my $lk100 = $self->_get_nb_lock($cdb100rlock) ) {
		$self->{lock} = $self->_get_lock();
		$self->update_readlink($cdbfile10, 1);
		$self->update_readlink($cdbfile, 1);
		_close($self->{lock});
		delete($self->{lock});
		_close($lk100);
	    }
	}
	foreach my $l ( @lock ) {
	    _close($l);
	}
	@lock = ();
	_close($lock);
	_close($lk10);

	$self->_get_db([0,0,1,0]); # update counts
	# drop handles
	delete $self->{cdb10};
    }
}

sub update_readlink {
    my($self, $cdbfile, $safe) = @_;
    my $readlink = $cdbfile . "r";
    my $tmplink = $cdbfile . "t";
    my $lock = $self->_get_lock() unless $safe;
    if (-e $readlink) {
	#warn "OVERWRITING LINK\n";
	unlink($tmplink); 
	if ( link($cdbfile, $tmplink) ) { 
	    rename($tmplink, $readlink) or warn "Can't overwrite $readlink with $tmplink: $!";
	    unlink($tmplink);
	} else {
	    #rename($tmplink, $readlink) or warn "Can't overwrite $readlink with $tmplink: $!";
	    unlink($tmplink); # if -e $tmplink;
	}

    } else {
	link($cdbfile, $readlink) or warn "Can't update $readlink: $!";
    }
    _close($lock) unless $safe;
}

# consolidate several cdb databases into one
sub consolidate {
    # 
    my($self, $newcdb, $olddb, $removed, $locks) = @_;
    my @cdb  = ();
    my $cdbfile;
    my $newrem;

    for $cdbfile ( @{$olddb} ) {
	# get handles to new cdb's
	push @cdb, CDB_File->TIEHASH($cdbfile)
	    or die "Can't tie $cdbfile $!";
    }

    if ( defined($removed) ) {
	$newrem = CDB_File->TIEHASH($removed) or die "Can't tie $removed $!";
    }
    _sort_in_vals($self, $newcdb, \@cdb, $newrem);
    
    shift @{$olddb}; # remove the name of the new index from list to delete

    for my $lk ( @{$locks} ) {
	_close($lk); # <------------------
    }
    undef($locks);
    for $cdbfile ( @{$olddb} ) {
	unlink($cdbfile); # or die "Can't delete $cdbfile $!";
	unlink($cdbfile . "l"); # or die "Can't delete $cdbfile $!";
    }
    $olddb = undef;
}

#each
sub _get_first_keyval { my $_k = $_[0]->FIRSTKEY(); 
			return (undef, undef) unless (defined($_k) && length($_k)); 
		        return ($_k, $_[0]->FETCH($_k)); }

sub _get_next_keyval  { my $_k = $_[0]->NEXTKEY($_[1]); 
			return (undef, undef) unless (defined($_k) && length($_k)); 
			return ($_k, $_[0]->FETCH($_k)); }
# _get_first_key
# get key and all vals 
sub _get_first_multi_keyvals {
    my $self = $_[0];
    my $_k = $_[1]->FIRSTKEY();
    return (undef, $_k, undef) unless (defined($_k) && length($_k));
    my $_v = [ $_[1]->FETCH($_k) ];
    my $_k2;
    #$| = 1;
    while ( ($_k2 = $_[1]->NEXTKEY($_k)) && defined($_k2) && 
	    !exists($HIDDEN{$_k2}) && _eq($_k2, $_k, $self->{num_key}) ) {
	#print STDERR "first fetched ", @$_v, " using $_[2]\n";
	push @{$_v}, $_[1]->FETCH($_k);
	#print STDERR "first now fetched ", @$_v, " using $_k\n";
    }
    # return next key and all vals with this key
    #print STDERR "returning $_k2, $_k, @$_v\n";
    return ($_k2, $_k, $_v);
}

# get all values with this key (for lex_val)
sub _get_next_multi_keyvals { # use this when we have duplicated keys 
    my $self = $_[0];
    my $_v = [ $_[1]->FETCH($_[2]) ];
    #$| = 1;
    #print STDERR "fetched ", @$_v, " using $_[2]\n";
    my $_k2;
    while ( ($_k2 = $_[1]->NEXTKEY($_[2])) && defined($_k2) && 
	     !exists($HIDDEN{$_k2}) && _eq($_k2, $_[2], $self->{num_key}) ) {
	#print STDERR "fetched ", @$_v, " using $_[2]\n";
	push @{$_v}, $_[1]->FETCH($_[2]);
	#print STDERR "now fetched ", @$_v, " using $_[2]\n";
    }
    # return next key and all vals with this key
    #print STDERR "returning $_k2, $_[2], @$_v\n";
    return ($_k2, $_[2], $_v); 
}

sub _get_first_bdb_keyval {
    return ($_[0]->seq($_[1], $_[2], R_FIRST()), $_[1], $_[2]);
}

sub _get_next_bdb_keyval {
    return ($_[0]->seq($_[1], $_[2], R_NEXT()), $_[1], $_[2]);
}

sub _get_next_val    { return shift @{ $_[0] }; }

# if lexical ? do a lexical compare : otherwise numeric
sub _lt { return $_[2] ? ($_[0] <  $_[1]) : ($_[0] lt $_[1]); }
sub _gt { return $_[2] ? ($_[0] >  $_[1]) : ($_[0] gt $_[1]); }
sub _eq { return $_[2] ? ($_[0] == $_[1]) : ($_[0] eq $_[1]); }
sub _ne { return $_[2] ? ($_[0] != $_[1]) : ($_[0] ne $_[1]); }
sub _le { return $_[2] ? ($_[0] <= $_[1]) : ($_[0] le $_[1]); }
sub _ge { return $_[2] ? ($_[0] >= $_[1]) : ($_[0] ge $_[1]); }

# get the least key or value from an array using lexical or numeric compare
sub _least { 
    my ($val, $num) = @_;
    my @least = (1, (0) x $#{$val});
    my $l = $val->[0];
    for my $i (1..$#{$val}) {
	if ( _lt($val->[$i], $l, $num) ) {
	    @least = (0) x $i;
	    $least[$i] = 1;
	    $l = $val->[$i];
	} elsif ( _eq($val->[$i], $l, $num) ) {
	    $least[$i] = 1;
	}
    }
    return ($l, @least);
}

# save a list of key/vals where keys may be duplicated and there is just one value per key
sub _sort_in_keyval { # only use this when the values have lexical sorting
    my ($self, $newcdb, $key, $val, $dbhdl, $remkey, $remval, $remhdl) = @_;
    my ($count, $msgids) = (0) x 2;
    my ($lkey, $lval, $remvals, @least);

    # $newcdb is the handle we are writing to
    # $key is ref to array of first keys
    # $val is a ref to array of first values
    # $dbhdl is a ref to array of handles
    #open(my $tmp, ">>/tmp/data.txt") or die "can't open /tmp/data.txt to append: $!";
    #print $tmp "OPENED file to append - Consolidating stuff\n";

    while ( @{$key} ) {
	# skip past hidden keys, remove indexes with no defined values
	foreach my $i ( 0..$#{$key} ) {
	    while ( defined($key->[$i]) && exists($HIDDEN{$key->[$i]}) ) {
		($key->[$i], $val->[$i]) = _get_next_keyval($dbhdl->[$i], $key->[$i]);
	    }
	    unless  ( defined($key->[$i]) ) {
		splice(@{$key}, $i, 1);
		splice(@{$val}, $i, 1);
		splice(@{$dbhdl}, $i, 1);
	    }
	}
	# get the least key(s)
	($lkey, @least) = _least($key, $self->{num_key});

	$lval = [];    # all values associated with this $lkey
	$remvals = []; # removed values associated with this $lkey

	# get all identical keys - sorting in by value
	foreach my $i ( 0..$#{$key} ) {
	    $lval->[$i] = [];
	    if ( $least[$i] ) {
		while ( defined($key->[$i]) && !exists($HIDDEN{$key->[$i]}) && 
			_eq($key->[$i], $lkey, $self->{num_key}) ) {
		    push @{$lval->[$i]}, $val->[$i];
		    ($key->[$i], $val->[$i]) = _get_next_keyval($dbhdl->[$i], $lkey);
		}
	    }
	}
	my $vals = combine_lex($lval);

	# get all the removed vals associated with this key
	while ( defined($remkey) && !exists($HIDDEN{$remkey}) && 
		_lt($remkey, $lkey, $self->{num_key}) ) {
	    ($remkey, $remval) = _get_next_keyval($remhdl, $remkey);
	}
	while ( defined($remkey) && !exists($HIDDEN{$remkey}) && 
		_eq($remkey, $lkey, $self->{num_key}) ) {
	    push @{$remvals}, $remval;
	    ($remkey, $remval) = _get_next_keyval($remhdl, $remkey);
	}
	# get all values not removed
	$vals = _and_not_lex($vals, $remvals);

	foreach my $vl ( @{$vals} ) {
	    $newcdb->insert($lkey, $vl);
	    #print $tmp "inserting $lkey, $vl\n";
	    $msgids++;
	}
	# if there is at least one value associated with this key, count it
	$count++ if @{$vals}; 
    }
    #close($tmp);
    $newcdb->insert("__COUNT__", $count);
    $newcdb->insert("__MSGIDS__", $msgids);
    $newcdb->finish or die "$0: CDB_File finish failed: $!";
    #undef($newcdb);
    
    return;
}

sub _sort_in_vals {
    # maintain list of non-empty db handles
    # $new = new cdb handle, $old = old cdb handles, $rem = removed cdb handle
    my ($self, $newcdb, $old, $remhdl) = @_;
    
    # maintain list of least keys for the dbs
    # @key is an array of most recent (least) keys from the "old" cdbs
    # @val is an array of corresponding packed values
    # $remkey, $remval = current least removed key, value
    # $lkey, $vref, $vrem - least key and array of packed values for cdb's
    my (@key, @val, $remkey, $remval, $lkey, $vref, $vrem);
    my (@v, @vrem, $lval, $lrem, $least, @new, $newv);
    my ($count, $msgids) = (0) x 2; # make sure these are initialized
    
    # get non-empty old db handles in @cdb
    my @cdb = grep ref($_) eq 'CDB_File', @{$old};
    
    # initialize @key and @val, $remkey and $remval
    my @empty;
    for my $i ( 0..$#cdb ) {
	($key[$i], $val[$i]) =_get_first_keyval($cdb[$i]);
	# $key[$i] = $cdb[$i]->FIRSTKEY();
	# $val[$i] = $cdb[$i]->FETCH($key[$i]) if ( defined($key[$i]) && length($key[$i]) );
	while ( defined($key[$i]) && exists($HIDDEN{$key[$i]}) ) { 
	    ($key[$i], $val[$i]) = _get_next_keyval($cdb[$i], $key[$i]);
	}
	# print("My first key/value is ", $key[$i], "=", $val[$i], "\n");
	$empty[$i] = !defined($key[$i]);
    }
    for (my $i = $#empty; $i > -1; $i-- ) {
	if ( $empty[$i] ) {
	    splice(@cdb, $i, 1);
	    splice(@key, $i, 1);
	    splice(@val, $i, 1);
	}
    }
    
    if ( defined($remhdl) && ref($remhdl) ) { # get first values from the removed db
	($remkey, $remval) = _get_first_keyval($remhdl);
	# $remkey = $remhdl->FIRSTKEY();
	# $remval = $remhdl->FETCH($remkey) if ( defined($remkey) && length($remkey) );
	while ( defined($remkey) && exists($HIDDEN{$remkey}) ) { 
	    ($remkey, $remval) = _get_next_keyval($remhdl, $remkey);
	}
	# print("My first key/value is ", $remkey, "=", $remval, "\n");
	undef($remhdl) unless ( defined($remkey) );
    } else {
	($remkey, $remval) = ( undef, undef );
    }

    return $self->_sort_in_keyval($newcdb, \@key, \@val, \@cdb, $remkey, $remval, $remhdl) 
	if $self->{lex_val};

    #print("Got key = $remkey ");
    $| = 1 if $self->{verbose}; # set me for printing stuff in a timely manner
    
    @new = ();
    # LOOP1
    while ( @key ) { # as long as we have keys
	
	# get the least key
	if ( defined($remhdl) ) {
	    ($lkey, $vref, $vrem) = $self->get_least_key_if_ok(\@key, \@val, \@cdb, 
							\$remkey, \$remval, $remhdl);
	    # print("*lkey*", $lkey, "*");
	    if ( defined($vrem) && length($vrem) ) {
		@vrem = unpack("w*", $vrem);
		$lrem = shift @vrem; # put first value in $lrem
		# print("^lrem^", $lrem, "^");
		if ( defined($vref)  && length($vref) ) {
		    @v = unpack("w*", $vref);
		    $lval = shift @v;
		    # print("%lval%", $lval, "%");
		} # LOOP2
		while ( defined($lval) ) {       # CHECK $vrem or $lrem?
		    # print("My lrem=$lrem and vrem=@vrem\n");
		    # if ( scalar(@vrem) and 
		     # print("%%lval%%", $lval, "%%");
		     # print("^^lrem^^", $lrem, "^^");
		    if ( defined($lrem) && 
			 ($newv = get_least_val_if_ok( \$lval, \@v , \$lrem, 
						       \@vrem )) ) {
			# DON'T FORGET to quit if @v is empty
			push @new, $newv; 
		    } else {
			# just load it all in (assume its sorted)
			push @new, $lval if ( defined($lval) and length($lval) );
			push @new, @v if (@v);
			
			@v = ();
			undef($lval);
		    }
		}
		$lrem = $lval = undef;
		@v = @vrem = ();
		
	    } else {
		next unless defined($vref);
		push @new, unpack("w*", $vref) if ( length($vref) );
	    }
	    # print("Got key =", $lkey, " with ", scalar(@new), 
		#  " values ", @new, "-!!!-\n");
	} else {
	    ($lkey, $vref) = $self->get_least_key(\@key, \@val, \@cdb);
	    next unless ( defined($lkey) and defined($vref));
	    push @new, unpack("w*", $vref) if ( length($vref) );
	    # print("Got key =", $lkey, " with ", scalar(@new), 
		#  " values @new -!-\n");
	}
	if ( scalar(@new) ) {
	    $newcdb->insert($lkey, pack("w*", @new));
	    $count++;
	    # print("+", $count, "+");
	    $msgids += scalar(@new);
	    # print("-", $msgids);
	}
	@new = ();
    }
    
    $newcdb->insert("__COUNT__", $count);
    $newcdb->insert("__MSGIDS__", $msgids);
    $newcdb->finish or die "$0: CDB_File finish failed: $!";
    undef($newcdb);

    @key = @val = @cdb = @v = @vrem = ();
    $lval = $lrem = $least = $remkey = $remval = $lkey = $vref = $vrem = undef;
    $old = undef;
}

# remove least value unless its in "rem"
sub get_least_val_if_ok {
    my ($dbval, $dbhdl, $remval, $remhdl) = @_;

    my $least = get_least_val($dbval, $dbhdl);
    
    while ( defined($$remval) && defined($least) && ($least >= $$remval) ) { # <-- >=
	# compare to remval
	while (  defined($$remval) && ($least > $$remval) ) { # <------------- >
	    # while remval less => pull $remhdl
	    $$remval = _get_next_val($remhdl);
	}
	
	my $rem_least = 0; # flag that says we need to get a new least value
	while ( defined($$remval) and ($$remval == $least) ) { # remove dups # <-- ==
	    $$remval = _get_next_val($remhdl);
	    $rem_least = 1;
	}
	if ( $rem_least ) {
	    # remval equal => pull remhdl, get new least value
	    $least = get_least_val($dbval, $dbhdl);
	} 
    }

    # (remval is now undef or greater than least) => return least
    return $least;
}

# remove least value given 1 array
# always update $dbval and pull or undef $dbhdl
sub get_least_val {
    my ($dbval, $dbhdl) = @_;
    
    unless ( defined($dbval) ) {
	undef($dbhdl);
	return undef;
    }
    #$sort ||= "NUM";
    my $least = $$dbval;

    # get next value / remove dups (and empty's)
    while ( defined($$dbval) && ( ref($dbhdl) eq 'ARRAY' )
	    && (($$dbval == $least) or !length($$dbval)) ) {
	$$dbval = _get_next_val($dbhdl);
    }

    undef($dbhdl) unless defined($$dbval);
    undef($$dbval) unless ( ref($dbhdl) eq 'ARRAY' );

    return $least;
}

# remove least value given 2 or more numerically sorted arrays (also works with 1)
sub get_least_val2 {
    my ($dbval, $dbhdl, $num_val) = @_;

    # clear empty values from @{$dbval} ? may not be needed !!!
    for ( my $i = $#{$dbval}; $i > -1; $i-- ) {
	# remove empty values
	while ( defined($dbval->[$i]) && !length($dbval->[$i]) ) {
	    $dbval->[$i] = _get_next_val($dbhdl->[$i]); # unlikely case
	}
	if ( !defined($dbval->[$i]) ) {
	    splice(@{$dbhdl}, $i, 1);
	    splice(@{$dbval}, $i, 1);

	    return undef unless @{$dbval};
	}
    }
    # now we know defined($dbval->[$i]) and length($dbval->[$i] for @{$dbval}

    my @least = (1, (0) x $#{$dbval}); 
    my $leastv = $dbval->[0];
    my $DONE = 0;
    
    for my $i (1..$#{$dbval}) {
	if ( _lt($dbval->[$i], $leastv, $num_val) ) {
	    @least = (0) x $i;
	    $least[$i] = 1;
	    $leastv = $dbval->[$i];
	} elsif ( _eq($dbval->[$i], $leastv, $num_val) ) {
	    $least[$i] = 1;
	}
    }
    for my $l ( 0..$#least ) {
	if ( $least[$l] ) {
	    # pull handle????
	    #$dbval->[$l] = _get_next_val($dbhdl->[$l]);
	    # get next value / remove dups (and empty's)
	    while ( defined($dbval->[$l]) && (!length($dbval->[$l]) ||
					      _eq($dbval->[$l], $leastv, $num_val)) ) {

		$dbval->[$l] = _get_next_val($dbhdl->[$l]);
		# last unless defined($dbval->[$l]);
	    }
	    # remove db handle from array of handles if its DONE 
	    $DONE ||= !defined($dbval->[$l]);
	}
    }
    if ( $DONE ) { # There has to be a simpler way to do this
	for ( my $l = $#least; $l > -1; $l-- ) {
	    unless ( defined($dbval->[$l]) ) {
		splice(@{$dbhdl}, $l, 1);
		splice(@{$dbval}, $l, 1);
	    }
	}
    }

    return $leastv;
}

# remove least key
sub get_least_key_if_ok { # $key is ref to array of keys
    my ($self, $key, $val, $dbhdl, $rem, $remval, $remhdl) = @_; 
    my $removed;

    # get least key in $db
    my ($lk, $vals) = $self->get_least_key($key, $val, $dbhdl);
    unless ( ref($remhdl) ) {
	# not sure if $$rem can be defined while ref($remhdl) is not true
	if ( defined($$rem) && defined($$remval) && !exists($HIDDEN{$$rem}) &&
	     _eq($$rem, $lk, $self->{num_key}) ) {
	    return ($lk, $vals, $remval);
	} else {
	    return ($lk, $vals, undef);
	}
    }
    
    # compare to rem
    while ( defined($$rem) ) {
	# next unless length($$rem);
	while ( defined($$rem) && (exists($HIDDEN{$$rem}) || !length($$rem)) ) {  
	    ($$rem, $$remval) = _get_next_keyval($remhdl,$$rem);
	}
	unless ( defined($$rem) ) {
	    undef($remhdl);
	    undef($$remval);
	    last;
	}
	# while remkey less => pull $remhdl
	if ( _lt($$rem, $lk, $self->{num_key}) ) {
	    ($$rem, $$remval) = _get_next_keyval($remhdl,$$rem);
	} else {
	    last;
	}
    }
    # remkey ($$rem) NOT less than least key ($lk)
    #my $removed = $$remval; 
    
    while ( defined($$rem) ) { # ($lk eq $$rem) ) {
	while ( defined($$rem) && exists($HIDDEN{$$rem}) ) {  
	    ($$rem, $$remval) = _get_next_keyval($remhdl,$$rem);
	}
	unless ( defined($$rem) ) {
	    undef($remhdl);
	    undef($$remval);
	    last;
	} 
	#$removed = $$remval; 
	if  ( _eq($$rem, $lk, $self->{num_key}) ) {
	    my $rvals = [ $removed, $$remval ];
	    $removed = combine_vals($rvals);
	    # remkey equal => pull remhdl, return least values and removed values
	    ($$rem, $$remval) = _get_next_keyval($remhdl,$$rem);
	} else {
	    last;
	}
    }
    return( $lk, $vals, $removed );
}

# NOTE: GET rid of EMPTY VALUE CHECKING when ok
#
# unpack and combine then repack vals for duplicated keys
sub combine_vals {
    my($vals) = shift; # ref to array of packed values
    my @lv = (); # least values
    my @v  = (); # array of value arrays
    my $cvref;

    return undef unless ( defined($vals) and ref($vals) );
    # setup @lv and @v
    for my $val ( @{$vals} ) {
	if ( defined($val) and length($val) ) {
	    push @v, [ unpack("w*", $val) ];
	    push @lv, shift @{$v[-1]};
	}
    }
    $cvref = _ior_vals(\@lv, \@v);
    #  return \@cv; # alternative return a reference to an array of values
    return scalar(@{$cvref}) ? pack("w*", @{$cvref}) : undef;
}

sub combine_lex {
    my($vals) = shift; # ref to array of refs to arrays of lexical values
    my @lv = (); # least values
    my @v  = (); # array of value arrays
    my $cvref;
    # If @{$val} == 1 should we return $val ?

    # setup @lv and @v
    for my $val ( @{$vals} ) {
	if ( defined($val) && (ref($val) eq "ARRAY") && scalar(@{$val}) ) {
	    push @lv, shift @{$val};
	    push @v, $val;
	}
    }
    $cvref = _ior_lex(\@lv, \@v);
    #  return \@cv; # alternative return a reference to an array of values
    return scalar(@{$cvref}) ? $cvref : [];
}

sub andnot_vals {
    my $self = shift;
    return $self->{lex_val} ?  _and_not_lex(@_) :_and_not_vals(@_);
}

sub and_vals {
    my $self = shift;
    return $self->{lex_val}  ? _and_lex(@_) : _and_vals(@_);
}

sub or_vals {
    my $self = shift;
    my ($ref1, $ref2) = @_;
    my @lv = (shift @$ref1, shift @$ref2);
    return $self->{lex_val} ? _ior_lex(\@lv, [$ref1, $ref2]) : _ior_vals(\@lv, [$ref1, $ref2]);
}

# this is a copy of _and_not_vals using lexical sorting <- is there a better way?
sub _and_not_lex {
    my ($vr, $rem) = @_;
    return [] unless @{$vr};
    return $vr unless @{$rem};
    my @clean = ();
    my $last = ""; # "" is an impossible value used for initialization
    my $not = shift @{$rem}; # $not is defined to start
    my $cur = shift @{$vr};
    # while there are values 
    while ( defined($cur) and defined($not) ) {
	while ( defined($cur) and ($cur eq $last) ) {
	    $cur = shift @{$vr}; # eliminate dups / get next $cur
	}
	last unless defined($cur); 
	$last = $cur;
	while ( defined($not) and ($not lt $cur) ) {
	    $not = shift @{$rem};
	}
	if ( defined($not) ) {
	    next if ( $not eq $cur );
	    push @clean, $cur;
	    $cur = shift @{$vr}; # <-- This line optional
	} else {
	    push @clean, $cur;
	}
    }
    push @clean, @{$vr} if @{$vr}; # add any values left after @{$rem} empty
    return \@clean;
}

sub _and_not_vals {
    my ($vr, $rem) = @_; # references to non-empty arrays?
    return [] unless @{$vr};
    return $vr unless @{$rem}; # checking they are non-empty anyway (for now) :|
    my @clean = ();
    my $last = -1; # -1 is an impossible value used for initialization
    my $not = shift @{$rem};
    my $cur = shift @{$vr};
    # while there are values 
    while ( defined($cur) and defined($not) ) {
	while ( defined($cur) and ($cur == $last) ) {
	    $cur = shift @{$vr}; # eliminate dups / get next $cur
	}
	last unless defined($cur);
	$last = $cur;
	while ( defined($not) and ($not < $cur) ) {
	    $not = shift @{$rem};
	}
	if ( defined($not) ) {
	    next if ( $not == $cur );
	    push @clean, $cur;
	    $cur = shift @{$vr}; # <-- This line optional
	} else {
	    push @clean, $cur;
	}
    }
    push @clean, @{$vr} if @{$vr}; # add any values left after @{$rem} empty
    return \@clean;
}

# inclusive or of lexical values using array refs
sub _ior_lex {
    my ($lvref, $vref) = @_;
    my @cv = (); # combined values
    my $nv;      # 

    while ( @{$lvref} ) { 
	if ( $#$lvref ) {
	    if ( $nv = get_least_val2($lvref, $vref, 0) ) { 
		# if we have more than one array of values
		push @cv, $nv;
		# print("-", $nv, "-");
	    }
	} else {
	    # just append values
	    
	    push @cv, $lvref->[0] 
		if ( defined($lvref->[0]) and length($lvref->[0]) );
	    push @cv, @{$vref->[0]} 
	        if ( defined($vref->[0]) and (ref($vref->[0]) eq 'ARRAY') );
	    $lvref = [];
	}
    }
    return \@cv;
}

# inclusive or of values using array refs
sub _ior_vals {
    my ($lvref, $vref) = @_;
    my @cv = (); # combined values
    my $nv;      # 

    while ( @{$lvref} ) { 
	if ( $#$lvref ) {
	    if ( $nv = get_least_val2($lvref, $vref, 1) ) { 
		# if we have more than one array of values
		push @cv, $nv;
		# print("-", $nv, "-");
	    }
	} else {
	    # just append values
	    
	    push @cv, $lvref->[0] 
		if ( defined($lvref->[0]) and length($lvref->[0]) );
	    push @cv, @{$vref->[0]} 
	        if ( defined($vref->[0]) and (ref($vref->[0]) eq 'ARRAY') );
	    $lvref = [];
	}
    }
    return \@cv;
}

# return values in A and B
sub _and_lex {
    my ($vref1, $vref2) = @_;
    my $v1 = shift @{$vref1};
    my $v2 = shift @{$vref2};
    my @both = ();
    my $last = "";
#    while ( @{$vref1} and @{$vref2} ) {
    while ( defined($v1) and defined($v2) ) {
	if ( $v1 eq $v2 ) {
	    push @both, $v1 unless ( $last eq $v1 );
	    $last = $v1;
	    $v1 = shift @{$vref1};
	    $v2 = shift @{$vref2};
	} elsif ( $v1 lt $v2 ) {
	    $v1 = shift @{$vref1};
	} else { # $v1 gt $v2
	    $v2 = shift @{$vref2};
	}
    }
    return \@both;
}

# return values in A and B
sub _and_vals {
    my ($vref1, $vref2) = @_;
    my $v1 = shift @{$vref1};
    my $v2 = shift @{$vref2};
    my @both = ();
    my $last = -1;
#    while ( @{$vref1} and @{$vref2} ) {
    while ( defined($v1) and defined($v2) ) {
	if ( $v1 == $v2 ) { # <------------ ==
	    push @both, $v1 unless ( $last == $v1 ); # <------------ ==
	    $last = $v1;
	    $v1 = shift @{$vref1};
	    $v2 = shift @{$vref2};
	} elsif ( $v1 < $v2 ) { # <------------ <
	    $v1 = shift @{$vref1};
	} else { # $v1 > $v2
	    $v2 = shift @{$vref2};
	}
    }
    return \@both;
}

# remove least key ( and associated vals) given 2 or more dbs
sub get_least_key {
    my ($self, $key, $val, $dbhdl) = @_;
    my $DONE = 0;
    #return ($key->[0], ());
    for ( my $l = $#{$key}; $l > -1; $l-- ) {
	while ( defined($key->[$l]) && exists($HIDDEN{$key->[$l]}) ) {
	    ($key->[$l], $val->[$l]) = _get_next_keyval($dbhdl->[$l], $key->[$l]);
	}
	unless  ( defined($key->[$l]) ) {
	    splice(@{$key}, $l, 1);
	    splice(@{$val}, $l, 1);
	    splice(@{$dbhdl}, $l, 1);
	}
    }
    return (undef, undef) unless scalar(@{$key});
    return $self->get_key($key, $val, $dbhdl) unless $#{$key};

    my @least = (1, (0) x $#{$key}); 
    my $leastk = $key->[0]; # not == __COUNT__

    for my $i (1..$#{$key}) { # <-------- lt
	if ( _lt($key->[$i], $leastk, $self->{num_key}) ) {
	    @least = (0) x $i;
	    $least[$i] = 1;
	    $leastk = $key->[$i];
	} elsif ( _eq($key->[$i], $leastk, $self->{num_key}) ) { # <----------- eq
	    $least[$i] = 1;
	}
    }

    #print("My least key is $leastk and values ");

    my @vals = (); # array of packed values
    my $leastv;
    # load in next key, values 
    for my $l ( 0..$#least ) {
	if ( $least[$l] ) {
	    push @vals,  $val->[$l]; # <-- save these values
	    # get next keys/values 
	    while ( defined($key->[$l]) and _eq($key->[$l], $leastk, $self->{num_key}) ) {
		($key->[$l], $val->[$l]) = _get_next_keyval($dbhdl->[$l], $key->[$l]);
		while ( defined($key->[$l]) && exists($HIDDEN{$key->[$l]}) ) {
		    ($key->[$l], $val->[$l]) = _get_next_keyval($dbhdl->[$l], 
								$key->[$l]);
		}
		unless ( defined($key->[$l]) ) {
		    $DONE = 1;
		    last;
		} elsif ( _eq($key->[$l], $leastk, $self->{num_key}) ) {
		    push @vals,  $val->[$l];
		}
	    }
	}
    }
    $leastv = $#vals ? combine_vals(\@vals) : $vals[0];
    # remove db handle from array of handles if its DONE 
    if ( $DONE ) { # There has to be a simpler way to do this
	for (my $l = $#least; $l > -1  ; $l-- ) {
	    unless ( defined($key->[$l]) ) {
		splice(@{$dbhdl}, $l, 1);
		splice(@{$key}, $l, 1);
		splice(@{$val}, $l, 1);
	    }
	}
    }
    #print((join " ", unpack("w*", $leastv)), "\n");
    # return / set @least
    return ( $leastk, $leastv );
}

# remove least key, values from one handle
sub get_key {
    my ($self, $key, $val, $dbhdl) = @_; 
    my ($k,$v) = ($key->[0], $val->[0]);

    # skip past COUNT and MSGIDS tags
    while ( defined($key->[0]) && exists($HIDDEN{$key->[0]}) ) {
	($key->[0], $val->[0]) = defined($dbhdl->[0]) ? 
	    _get_next_keyval($dbhdl->[0],$key->[0]) : (undef, undef);
	($k,$v) = ($key->[0], $val->[0]);
    }
    unless ( defined($k) ) {
	undef($dbhdl);
	@{$key} = ();
	@{$val} = ();
	return (undef, undef);
    }
    #print("get_key got ", $k, "--", "@v", "--\n");
    
    # look ahead to see if next key is the same
    ($key->[0], $val->[0]) = defined($dbhdl->[0]) ? 
	_get_next_keyval($dbhdl->[0],$key->[0]) : (undef, undef);
    while ( defined($key->[0]) and !exists($HIDDEN{$key->[0]}) && 
	    _eq($k, $key->[0], $self->{num_key}) ) { # <----------- eq
	my $nv = [ $v, $val->[0] ];
	$v = $self->{lex_val} ? combine_lex($nv) : combine_vals($nv);
	($key->[0], $val->[0]) = defined($dbhdl->[0]) ? 
	    _get_next_keyval($dbhdl->[0],$key->[0]) : (undef, undef);
    }
    #print("And now its set to ", $key->[0], "\n-\n", $val->[0], "-&-\n");
    # skip past COUNT and MSGIDS tags
    while ( defined($key->[0]) && exists($HIDDEN{$key->[0]}) ) {
	($key->[0], $val->[0]) = defined($dbhdl->[0]) ? 
	    _get_next_keyval($dbhdl->[0],$key->[0]) : (undef, undef);
    }
    unless ( defined($key->[0]) ) {
	undef($dbhdl);
	@{$key} = ();
	@{$val} = ();
    }
    return ($k, $v);
}

#
# IO functions borrowed and adapted from PureMessage::Policy::Message;
#

sub _open {
    my $file = shift;
    open(my $rc, "+>", $file) or die "$$: Can't open $file";
    if ( DBG ) {
	my $fn = fileno($rc);
	if ( $fn > 30 ) {
	    my $info = (caller())[2];
	    print STDERR "$$ ## from line $info ## opened fd $fn\n"; 
        }
    }
    return $rc;
}

sub _close {
    my $fn = fileno($_[0]);
    my $rc = close($_[0]);
    if ( DBG ) {
	warn "$$ #### close(fd $fn) failed: $!" unless $rc;
	if ( $fn > 30 ) {
	    my $info = (caller())[2];
	    print STDERR "$$ ## from line $info ## closing fd $fn\n"; 
        }
    }
    return $rc;
}

sub _flock {
    my ($fh, $fl) = @_;

    if ( flock($fh, $fl) ) {
	if ( DBG ) {
	    my $info = (caller())[2];
	    my $fn = fileno($_[0]);
	    if ( $fn > 30 ) {
		my $info = (caller())[2];
		print STDERR "$$ ## from line $info ## locking fd $fn\n"; 
	    }
	}
	return 1;
    } else {
	return;
    }
}

package PureMessage::MessageStore::Dir::DB::MemDB;
# an in-memory structure for sorted data

use strict;
# uncomment the $self->{dirty} lines if you want to be safe !!!

sub LEXICAL { $a cmp $b };
sub NUMERIC { $a <=> $b };

sub new {
    my $invocant = shift;
    my %opts  = @_;
    my $class = ref($invocant) ||  $invocant;
    my $self  = { COUNT   => 0, 
		  NUM_KEY => delete $opts{num_key} || 0,
		  LEX_VAL => delete $opts{lex_val} || 0,
		  VERBOSE => delete $opts{verbose} || 0,
		  THIS    => 0,
		  VAL     => {},
		  ORDER   => [],
	      };

    die "MemDB constructor called with bad options: " . (join ", ", keys %opts) 
	if ( keys %opts );
    # ORDER is sorted array of keys
    bless $self, $class;

    return $self;
}
# want storing as hash, reading as sorted array of {key => [values]}

sub DESTROY {
    my $self = shift;
    @{$self->{ORDER}} = ();
    %{$self->{VAL}} = ();
    $self = {};
}

sub FIRSTKEY { # keys sorted lexically?
    my $self = shift;
    $self->{THIS} = 0;
    unless ( $self->{NUM_KEY} ) {
	$self->{ORDER} = [ sort keys %{$self->{VAL}} ]; # lexical is default
    } else {
	$self->{ORDER} = [ sort NUMERIC keys %{$self->{VAL}} ]; 
    }
    $self->{COUNT} = scalar(@{$self->{ORDER}});
    # print ("MemDB sorted ", $self->{COUNT}, " keys\n") if $self->{VERBOSE};
    $self->{dirty} = 0;
    return $self->{ORDER}[0];
}

# if this returns $key, VALUE =  $self->{VAL}{$key}
sub NEXTKEY {
    my ($self, $lkey) = @_;
    die "MemDB NEXTKEY called after STORE before FIRSTKEY at $lkey" if $self->{dirty};
    return ( $self->{COUNT} > ++$self->{THIS} ) ? $self->{ORDER}[$self->{THIS}] : undef;
}

# return a reference to sorted array of integers
sub FETCH {
    my ($self, $lkey) = @_;

    return $self->{VAL}{$lkey} if exists($self->{VAL}{$lkey});
    return;
}

# I ASSUME you will call FIRSTKEY after your LAST call to STORE and before NEXTKEY
sub STORE {
    shift->put(@_);
}

# same as STORE
sub put {
    my ($self, $k, $v) = @_;
    #print STDERR "called  put with $k and $v\n";# @{$v}, "\n";
    # ASSERT $k, $v are both defined and non-empty
    return 0 unless (length($v) && length($k)); # don't add empty values
    $self->{dirty} = 1;
    $k =~ s/\A__/___/; # escape hidden keys!
    my $count = 0;
    # if key exists, add to value array
    if ( exists($self->{VAL}{$k}) ) {
	$count = $self->_add_val($v, $self->{VAL}{$k});
    } else {
	$self->{VAL}{$k} = [$v];
	$count = 1;
    }
    $self->{COUNT} += $count;

    return $count;
    # ASSERT: NO empty values
    # ASSERT: $self->{VAL}{$k} values are sorted
}

# store faster / used with revdb
sub put_many {
    my ($self, $k, $v) = @_; # $v is array ref

    # ASSERT $k, $v are both defined and non-empty
    return 0 unless ((ref($v) eq "ARRAY") && length($k)); # don't add empty keys
    $k =~ s/\A__/___/;
    my $i = 0;
    my $val = $v->[$i++]; 

    return 0 unless (defined($val) && length($val));
    my $count = 0;
    
    # if key exists, add to value array
    if ( exists($self->{VAL}{$k}) ) {
	$count++ if $self->_add_val($val, $self->{VAL}{$k});
    } else {
	$self->{VAL}{$k} = [ $val ];
	$count++;
    }
    # now it exists, add the rest
    while ( $i < scalar(@{$v}) ) {
	$val = $v->[$i++]; 
	next unless ( defined($val) && length($val) );
	$count++ if $self->_add_val($val, $self->{VAL}{$k});
    }
    $self->{COUNT}  += $count;
    $self->{dirty} ||= $count;

    return $count;  # do we really want an _add_vals() method?
}

# delete a key/value IF it exists
sub del_many {
    my ($self, $k, $v) = @_; # $v is array ref

    # ASSERT $k, $v are both defined and non-empty
    return 0 unless ((ref($v) eq "ARRAY") && length($k)); # don't delete empty values
    $k =~ s/\A__/___/; # escape hidden keys!
    my $i = 0;
    my $val = $v->[$i++]; 
    return 0 unless (defined($val) && length($val));
    my $count = 0;

    # if key exists, REMOVE FROM value array
    if ( exists($self->{VAL}{$k}) && (ref($self->{VAL}{$k}) eq "ARRAY") && 
	 @{$self->{VAL}{$k}} ) {
	$count += $self->_rem_val($val, $self->{VAL}{$k});
	while ( $i < scalar(@{$v}) ) {
	    $val = $val = $v->[$i++]; # shift @{$v};
	    next unless ( defined($val) && length($val) );
	    $count++ if $self->_rem_val($val, $self->{VAL}{$k});
	}
	delete($self->{VAL}{$k}) unless @{$self->{VAL}{$k}};
	# @{$self->{VAL}{$k}} is not empty
    } else {
	# don't need to delete anything if the key/value doesn't exist
	#delete($self->{VAL}{$k}) if exists($self->{VAL}{$k});
	return 0;
    }

    $self->{COUNT}  -= $count; # where $self->{COUNT} is the number of values
    $self->{dirty} ||= $count;

    return $count;  # do we really want a _del_vals() method?
}

# delete a key/value IF it exists
sub del {
    my ($self, $k, $v) = @_;

    # ASSERT $k, $v are both defined and non-empty
    return 0 unless (length($v) && length($k)); # don't delete empty values
    $k =~ s/\A__/___/; # escape hidden keys!

    my $count = 0;
    if ( exists($self->{VAL}{$k}) && (ref($self->{VAL}{$k}) eq "ARRAY") && 
	 @{$self->{VAL}{$k}} ) {
	$count = $self->_rem_val($v, $self->{VAL}{$k});
	delete($self->{VAL}{$k}) unless @{$self->{VAL}{$k}};
	# @{$self->{VAL}{$k}} is not empty
    } else {
	# don't need to delete anything if the key/value doesn't exist
	return 0;
    }
    $self->{COUNT} -= $count; # where $self->{COUNT} is the number of values
    $self->{dirty} = 1 if $count;

    return $count;
}

# same as FETCH
sub get {
    my ($self, $lkey) = @_;

    return $self->{VAL}{$lkey} if exists($self->{VAL}{$lkey});
    return;
}

# returns min and max keys
sub get_range {
    my $self = shift;

    return () unless ($self->{dirty} || $self->{COUNT});

    if ( $self->{dirty} ) {
	unless ( $self->{NUM_KEY} ) {
	    $self->{ORDER} = [ sort keys %{$self->{VAL}} ]; # lexical is default
	} else {
	    $self->{ORDER} = [ sort NUMERIC keys %{$self->{VAL}} ]; 
	}
	$self->{dirty} = 0;
    }
    return ($self->{ORDER}[0], $self->{ORDER}[-1])
}

# remove a value from an array
# usually removing value near the end?
# @{$vref} is not empty
sub _rem_val {
    my ($self, $val, $vref) = @_;
    my ($lb, $ub) = (0, $#{$vref});
    my $mid; # = int(($ub - $lb)/2) + $lb;

    if ( $self->{LEX_VAL} ) {
	return 0 if ( ($val gt $vref->[$ub]) or ($val lt $vref->[$lb]) );
	for my $v ( $ub, $lb ) {
	    if ( $val eq $vref->[$v] ) {
		splice(@{$vref}, $v, 1);
		return 1;
	    }
	}
	while ( $lb < ($ub - 1) ) {
	    $mid = int(($ub - $lb)/2) + $lb; # DOESN'T work if $lb = $ub = 0 !!!
	    if ( $val le $vref->[$mid] ) {
		$ub = $mid;
		$lb++;
	    } else {
		$lb = $mid + 1;
		$ub--;
	    }
	    for my $v ( $ub, $lb ) {
		if ( $val eq $vref->[$v] ) {
		    splice(@{$vref}, $v, 1);
		    return 1;
		}
	    }
	}
    } else { # !$self->{LEX_VAL}
	return 0 if ( ($val > $vref->[$ub]) or ($val < $vref->[$lb]) );
	for my $v ( $ub, $lb ) {
	    if ( $val == $vref->[$v] ) {
		splice(@{$vref}, $v, 1);
		return 1;
	    }
	}
	while ( $lb < ($ub - 1) ) {
	    $mid = int(($ub - $lb)/2) + $lb;
	    if ( $val <= $vref->[$mid] ) {
		$ub = $mid;
		$lb++;
	    } else {
		$lb = $mid + 1;
		$ub--;
	    }
	    for my $v ( $ub, $lb ) {
		if ( $val == $vref->[$v] ) {
		    splice(@{$vref}, $v, 1);
		    return 1;
		}
	    }
	}
    }
    return 0;
}

# sort numerically or lexically - values usually appended
# return true if value added
sub _add_val {
    my ($self, $val, $vref) = @_;

    if ( $self->{LEX_VAL} ) {
	if ( $val gt $vref->[-1] ) {      # most likely case
	    push @{$vref}, $val;
	} elsif ( $val lt $vref->[0] ) { # next most likely
	    unshift @{$vref}, $val;
	} elsif ( ($val eq $vref->[-1]) or ($val eq $vref->[0]) ) {
	    # drop dups
	    return 0;
	} else { # binary search - least likely case
	    my($lb,$ub) = (0, $#{$vref});
	    my $mid = int($#{$vref}/2);
	    while ( $mid ne $lb ) {
		if ( $val gt  $vref->[$mid] ) {
		    $lb = $mid;
		} elsif ( $val lt $vref->[$mid] ) {
		    $ub = $mid;
		} else { # $val eq $vref->[$mid] found dup!
		    return 0;
		}
		$mid = int(($ub - $lb)/2) + $lb;
	    }
	    splice(@{$vref},$lb,0,$val);
	}
    }  # !$self->{LEX_VAL}
    elsif ( $val > $vref->[-1] ) {  # most likely case
	push @{$vref}, $val;
    } elsif ( $val < $vref->[0] ) { # next most likely 
	unshift @{$vref}, $val;
    } elsif ( ($val == $vref->[-1]) or ($val == $vref->[0]) ) {
	# drop dups
	return 0;
    } else { # binary search - least likely case
	my($lb,$ub) = (0, $#{$vref});
	my $mid = int($#{$vref}/2);
	while ( $mid != $lb ) {
	    if ( $val > $vref->[$mid] ) {
		$lb = $mid;
	    } elsif ( $val < $vref->[$mid] ) {
		$ub = $mid;
	    } else { # $val == $vref->[$mid] found dup!
		return 0;
	    }
	    $mid = int(($ub - $lb)/2) + $lb;
	}
	splice(@{$vref},$lb,0,$val);
    }

    return 1;
}

1;

#exit 0;

__END__
