package PureMessage::MessageStore::Dir::Index::cdb;

use strict;
use ActiveState::Config::Simple qw(config_parse_file);
use Carp qw(carp croak);

use Fcntl qw(O_CREAT O_RDWR O_RDONLY LOCK_SH LOCK_EX);
use POSIX qw(:sys_wait_h);
use PureMessage::MessageStore::Dir::DB;
use PureMessage::MessageStore::Dir::Fields;
sub FIELDS () { "PureMessage::MessageStore::Dir::Fields" }
use Time::HiRes qw (usleep);

sub LEXICAL { $a cmp $b }; # lexical sorting
sub NUMERIC { $a <=> $b }; 

sub new {
    my($class, $dir, %opts) = @_;

    my $writable  = delete $opts{writable} || 0;
    my $truncate  = delete $opts{truncate};
    my $verbose   = delete $opts{verbose};
    my $idir      = delete $opts{index_dir} || $dir;
    my $conf_file = delete $opts{config} || "index.conf";
    my $alt_conf  = delete $opts{alt_config};

    $conf_file = "$dir/$conf_file" unless $conf_file =~ m,^/,;
    my $conf = config_parse_file($conf_file);
    $conf ||= config_parse_file($alt_conf) if defined $alt_conf;
    $conf ||= {};
    $idir = $conf->{index_directory} if exists($conf->{index_directory});
    # $| = 1;
    # $verbose = 1;

    if ($^W) {
	for (keys %opts) {
	    carp("Unrecognized option '$_'");
	}
    }

    my $index;
    my $idx = {FIELD => {}, INDEX => [], mdb => {}};
    my %field;
    my $have_db = 0;
    my $i = 0;
    
    if ($index = $conf->{index}) {
	for my $iname (keys %$index) {
	    my $h = $index->{$iname};
	    if (exists $h->{field}) {
		my $f = $h->{field};
		$h->{field} = [$f] unless ref($f) eq "ARRAY";
	    }
	    else {
		$h->{field} = [$iname];
	    }
	    $h->{enabled}  = 1 unless exists($h->{enabled});
	    $h->{reverse_index} = 0 unless exists($h->{reverse_index});
	    if ( exists($h->{key_transform}) ) {
		unless ($h->{key_transform} eq "lowercase"  ) {
		    delete($h->{key_transform});
		}
	    }

	    # intepret bools - some default to false
	    for my $b ($h->{enabled}, $h->{numeric_keys}, $h->{reverse_index}) { 
		$b = 0 unless ( defined($b) && lc($b) =~ /^(yes|on|true|1)$/ );
	    } # some bools default to true
	    
	    next unless $h->{enabled};

	    for my $f (@{$h->{field}}) {
		die "Bad field '$f' for index '$iname' in $conf_file"
		    unless FIELDS->field_ok($f);
		die "$f field saved in more than one index" if exists($field{$f});
		$field{$f} = $iname;
		$idx->{FIELD}{$iname}{$f} = $i;
	    }
	    $idx->{INDEX}[$i] = $h;
	    $h->{iname} = $iname; # save the name of the index
	    $i++; # $i identifies the index containing a field, NOT the field
	    # $index->{FIELD} is unique for a given index.conf but will be incorrect if
	    #  indexes are added or index order changed!  This requires a reindex.
	    $h->{db} =  PureMessage::MessageStore::Dir::DB->new( dir      => $idir,
								 db_name  => "i_$iname",
								 writable => $writable,
								 truncate => $truncate,
								 verbose  => $verbose,
								 num_key  => $h->{numeric_keys},
								 lex_val  => 0,
								 );
	    $have_db = 1;
	    next unless $h->{reverse_index};
	    # reverse index - (id=>[field=value])
	    $h->{rdb} =  PureMessage::MessageStore::Dir::DB->new( dir      => $idir,
								  db_name  => "r_$iname",
								  writable => $writable,
								  truncate => $truncate,
								  verbose  => $verbose,
								  num_key  => 1,
								  lex_val  => !$h->{numeric_keys},
								 );
	    # stats index - (#ids=>[value])?
#	    $h->{sdb} =  PureMessage::MessageStore::Dir::DB->new( dir      => $idir,
#								  db_name  => "s_$iname",
#								  writable => $writable,
#								  truncate => $truncate,
#								  verbose  => $verbose,
#								  num_key  => 1,
#								  lex_val  => 1,
#								 );
	}
	# master index - (id=>reverse index)
	$idx->{mdb}{db} =  PureMessage::MessageStore::Dir::DB->new( dir      => $idir,
								    db_name  => "i_mdb",
								    writable => $writable,
								    truncate => $truncate,
								    verbose  => $verbose,
								    num_key  => 1,
								    lex_val  => 0,
								    );
	
	$idx->{mdb}{enabled} = 1;
	$idx->{mdb}{field} = ["mdb"];
	$field{mdb} = "mdb";
	# get and/or save field ids to/from master?

	$index = {} unless $have_db;
    }
    else {
	$index = {};
    }
    undef($conf);
    foreach my $k (keys %{$index}) {
	$idx->{$k} = $index->{$k};
    }
    undef($index);

    bless {
	dir => $idir,
	index => $idx,
	field => \%field,
	rdonly => !$writable,
	verbose => $verbose,
    }, $class;
}

sub DESTROY {
    my $self = shift;
    # make sure database is closed
    delete $self->{index};
}

sub index {
    my($self, $msg, $forget) = @_;
    croak("Can't index message for readonly index '$self->{dir}'")
	if $self->{rdonly};
    
    my $id = $msg->id;
    my $count = 0;
    croak("Tried to index Message with NULL ID") 
	unless (defined($id) && length($id));
    croak("Tried to index Message with INVALID ID: $id") 
	unless ($id =~ /^\d+$/o);
    print "Indexing message $id\n" if $self->{verbose};
    my @mstr = ();
    my $mdb = $self->{index}{mdb}{db};
    my $field_num = $self->{index}{FIELD};
    for my $iname (keys %{$self->{index}}) { # each index name in index.conf
	next if ( $iname =~ /INDEX|FIELD/o );
	next if ( $iname eq "mdb" );
 	my $h = $self->{index}{$iname};
	next unless $h->{enabled};
	my $db  = $h->{db};
	my $rdb = $h->{rdb};
	my @rev = ();
	next unless $db;
	next unless $mdb;
	for my $field (@{$h->{field}}) { # each field in this index
	    for my $val (FIELDS->field($msg, $field)) {
		next unless (defined($val) && length($val) && ($val =~/\S+/));
		if ( exists($h->{key_transform}) ) {
		    $val = lc($val) if ( $h->{key_transform} eq "lowercase" );
		}
		if ($self->{verbose}) {
		    require Data::Dump;
		    print "  $iname.$field = ", Data::Dump::dump($val), "\n";
		}
		push @rev, $val; # $val is lexical?
		die "Attempt to index on unknown field: $field, need to reindex" 
		    unless defined($field_num->{$iname}{$field});
		unless ( @mstr && $mstr[-1] == $field_num->{$iname}{$field} ) {
		    push @mstr, $field_num->{$iname}{$field};
		}
		if ($forget) {
		    $count += $db->del($val, $id);
		} else {
		    $count += $db->put($val, $id);
		}
	    }
	}
	next unless (ref($rdb) eq "PureMessage::MessageStore::Dir::DB");
	if ( @rev ) {
	    @rev  = $h->{numeric_keys} ? (sort NUMERIC @rev) : (sort @rev);
	    if ($forget) {
		$count += $rdb->del_rev($id, \@rev);
	    } else {
		$count += $rdb->put_rev($id, \@rev);
	    }
	}
    }
    return $count unless @mstr;

    @mstr = sort NUMERIC @mstr; # is this sort slow?
    if ($self->{verbose}) {
	require Data::Dump;
	print "  master index for $id = ", Data::Dump::dump(\@mstr), "\n";
    }
    if ($forget) {
	$count += $mdb->del_rev($id, \@mstr);
    } else {
	$count += $mdb->put_rev($id, \@mstr);
    }
    return $count; # what is this counting?
}

sub done {
    my $self = shift;

    for my $iname (keys %{$self->{index}}) { # each index name in index.conf
	next if ( $iname =~ /INDEX|FIELD/o );
	next if ( $iname eq "mdb" );
 	my $h = $self->{index}{$iname};
	$h->{db}->done() if $h->{enabled};
	$h->{rdb}->done() if $h->{reverse_index};
    }
    $self->{index}{mdb}{db}->done();
}

sub forget {
    my($self, $msg) = @_;
    $self->index($msg, 1);
}

# should this be a re-index kind of function with optional forgetting?
# called as add_new($cb, $progress_cb, $min_id, $max_id);
sub add_new {
    my ($self, $cb, $p_cb, $min, $max) = @_;
    # &$cb(@id)
    my $records = 0;
    my $checked = 0;
    my $msgs = 0;
    my ($kvals, $dvals, $nid, $old_total, $new_total, $num_ids);

    my $mdb = $self->{index}{mdb}{db};
    ($old_total) = $mdb->get_count();
    $new_total = $max - $min;
    $new_total = 0 unless ($new_total > 0);
    $num_ids = int($new_total/100) || 1; # just to make &p_cb pretty
    if ( $p_cb ) {
	&$p_cb(0.0); 
	print "  checking $new_total messages";
    }
    return "NYI";
}

# called as forget_old($cb, $progress_cb, $min_id, $max_id);
sub forget_old { # scan_index, 
    my ($self, $cb, $p_cb, $min, $max) = @_;
    my $records = 0;
    my $checked = 0;
    my $msgs = 0;
    my ($kvals, $dvals, $nid, $total, $num_ids);
    
    my $mdb = $self->{index}{mdb}{db};
    ($total) = $mdb->get_count();
    $num_ids = int($total/100) || 1; # just to make &p_cb pretty
    if ( $p_cb ) {
	&$p_cb(0.0); 
	print "  checking $total messages";
    }
    # create a new master "deleted" index if ids to delete
    my $dmdb = PureMessage::MessageStore::Dir::DB->new( dir      => $self->{dir},
						        db_name  => "d_mdb",
						        truncate => 1,
						        verbose  => $self->{verbose},
						        num_key  => 1,
							); # has numeric values
    do {
	($kvals, $nid) = $mdb->get_keys($num_ids, $nid, $min, $max);
	$checked += scalar(@{$kvals});
	foreach my $info ( @{$kvals} ) {
	    my ($id, $ind) = %{$info}; # $id => [$index_num, ...]
	    next if ( -e &{$cb}($id) ); # next if this message in quarantine
	    next unless ( @{$ind} );
	    $dmdb->put_many($id, $ind);
	    &$p_cb($checked/$total) if $p_cb; 
	}
	&$p_cb($checked/$total) if $p_cb;
	undef($kvals);
    } while ( defined($nid) ); # found all ids we need to remove
    $dmdb->flush(1);
    my ($number) = $dmdb->get_count();
    print "\nChecked $checked messages, found $number to remove\n";

    if ( $number ) {
	$checked = 0;
	if ( $p_cb ) {
	    &$p_cb(0.0); 
	    print "  removing records";
	}
	do {
	    my $gone = [];
	    #($dvals, $nid) = $dmdb->get_keys($num_ids, $nid, $min, $max);
	    ($dvals, $nid) = $dmdb->get_keys($num_ids, $nid);
	    $checked += scalar(@{$dvals});
	    my %nid = ();
	    foreach my $info ( @{$dvals} ) {
		my ($id, $ind) = %{$info}; # $id => [$iname_num, ...]
		# put into $iname_num => [ids] structure
		foreach my $iname_num ( @{$ind} ) {
		    if ( ref($gone->[$iname_num]) eq "ARRAY" ) {
			push @{$gone->[$iname_num]}, $id;
		    } else {
			$gone->[$iname_num] = [ $id ];  
		    }
		}
		$msgs++;
	    }
	    #print("\nchecked= $checked - gone->[0] has ", scalar(@{$gone->[0]}), " entries\n");
	    # get info as [index]{key => [id,...]}
	    foreach my $iname_num ( 0..$#{$gone} ) {
		# $gone->[$iname_num] = [$id, ...]
		#print "working with gone->[$iname_num]\n";
		my $dele = [];
		my $rdb = $self->{index}{INDEX}[$iname_num]{rdb};
		my $db  = $self->{index}{INDEX}[$iname_num]{db};
		# Need a method HERE to use if we have no reverse index!!!
		# next unless $rdb; # $gone->[$iname_num] = [id, ...]
		my $rvals;
		if ( ref($rdb) eq "PureMessage::MessageStore::Dir::DB" ) {
		    $rvals = $rdb->get_multi_keys($gone->[$iname_num]);
		} else { # this is the slower way
		    $rvals = $db->get_multi_vals($gone->[$iname_num]);
		}
		#print "\nready to remove ", scalar(@{$rvals}), " records\n";
		# $rvals 
		foreach my $info ( @{$rvals} ) {
		    my ($id, $vref) = %{$info};
		    foreach my $key ( @{$vref} ) {
			if ( defined($key) && length($key) && ($key =~ /\S+/) ) {
			    if ( exists($dele->[$iname_num]{$key}) ) {
				push @{$dele->[$iname_num]{$key}}, $id;
			    } else {
				$dele->[$iname_num]{$key} = [ $id ];  
			    }
			}
		    } # now delete from rdb
		    $records += $rdb->del_many($id, $vref) 
			if ( ref($rdb) eq "PureMessage::MessageStore::Dir::DB" );
		}
		# delete this stuff
		foreach my $key ( keys %{$dele->[$iname_num]} ) {
		    my $rem += $db->del_many($key, $dele->[$iname_num]{$key});
		    $records += $rem;
		    #print "removed $rem/$records with $key from $iname_num\n";
		}
	    }
	    foreach my $info ( @{$dvals} ) {
		my ($id, $ind) = %{$info}; # $id => [$iname_num, ...]
		$records += $mdb->del_many($id, $ind); # remove from master index
	    }
	    undef($dvals);
	    &$p_cb($msgs/$number) if $p_cb; 
	} while ( defined($nid) );
	$mdb->flush(1);
    }
    $mdb->flush(1);
    my ($new_count) = $mdb->get_count();
    print "Removed $records records from $msgs messages. Now there are $new_count in quarantine.\n";
    # now delete $dmdb
    $dmdb->remove_me();
    undef($dmdb);

    return ($msgs, $records);
}

sub find {
    my($self, $field, $value, $op, $ref) = @_;
    my $idx = $self->{field}{$field};
    croak("No index for field '$field'") unless $idx;
    print "Using index '$idx' for '$field'\n" if $self->{verbose};

    my $db = $self->{index}{$idx}{db};
    unless ($db) {
    	print "Index '$idx' not yet created\n" if $self->{verbose};
	return;
    }
    my $ids = $self->{index}{$idx}{db}->get($value);
    $ids = [] unless (ref($ids) eq 'ARRAY');

    if ( defined($op) and defined($ref) ) {
	print "doing $op\n" if $self->{verbose};
	
	$ref = [$ref] unless ( ref($ref) eq 'ARRAY' );
	if ( $op =~ /^andnot_$/ ) {
	    $ids = $self->{index}{$idx}{db}->andnot_vals($ids, $ref);

	} elsif ( $op =~ /^and_$/ ) {
	    $ids = $self->{index}{$idx}{db}->and_vals($ids, $ref);

	} elsif ( $op =~ /^or_$/ ) {
	    $ids = $self->{index}{$idx}{db}->or_vals($ids, $ref);
	}
    }

    return wantarray ? @{$ids} : $ids;
}

# find keys given an id and [field_name(s)], from a reverse index
sub find_rev {
    my($self, $id, $field_ref, $op, $ref) = @_;
    # note that $op, $ref are probably meaningless in this context!!!
    # will decide if its useful when I have time to think about it
    my($result);
    foreach my $field ( @{$field_ref} ) {
	my $idx = $self->{field}{$field};
	croak("No index for field '$field'") unless $idx;
	print "Using index '$idx' for '$field'\n" if $self->{verbose};

	my $rdb = $self->{index}{$idx}{rdb};
	unless ( ref($rdb) eq "PureMessage::MessageStore::Dir::DB" ) {
	    print "Reverse Index '$idx' not yet created\n" if $self->{verbose};
	    return;
	}

	my $vals = $self->{index}{$idx}{rdb}->get($id);
	$vals = [] unless (ref($vals) eq 'ARRAY');
	
	if ( defined($op) and defined($ref) ) {
	    print "doing $op\n" if $self->{verbose};
	    
	    $ref = [$ref] unless ( ref($ref) eq 'ARRAY' );
	    if ( $op =~ /^andnot_$/ ) {
		$vals = $self->{index}{$idx}{rdb}->andnot_vals($vals, $ref);
		
	    } elsif ( $op =~ /^and_$/ ) {
		$vals = $self->{index}{$idx}{rdb}->and_vals($vals, $ref);
		
	    } elsif ( $op =~ /^or_$/ ) {
		$vals = $self->{index}{$idx}{rdb}->or_vals($vals, $ref);
	    }
	}
	push @{$result}, $vals;
    }

    return wantarray ? @{$result} : $result;
}

sub update_db {
    my($self) = @_;
    for my $iname (keys %{$self->{index}}) {
	next if ( $iname =~ /INDEX|FIELD/o );
        my $h = $self->{index}{$iname};
        next unless $h->{enabled};
        my $db = $h->{db};
        $db->update_db("i_$iname");
        print "Index '$iname' updated to cdb format with ", $db->get_count(), 
              " messages indexed\n" if $self->{verbose};
    }
}

sub range {
    # get/set numerical min and max to be returned by find
    # my($self, $field, $min, $max) = @_;
    # this works if ids are sorted in date order
    # currently this is numerical order
    my($self, %opts) = @_;
    my $field    = delete $opts{field};
    my $min_id   = delete $opts{min_id};
    my $max_id   = delete $opts{max_id};
    # these dates are useful if date indexes exist to calculate ids
    my $min_date = delete $opts{min_date};
    my $max_date = delete $opts{max_date};
    #my $save_ids = delete $opts{save_range} || 0;
    if ($^W) {
	for (keys %opts) {
	    carp("Unrecognized option '$_'");
	}
    } 

    # set id range values if they were provided
    $self->{min_id} = $min_id if defined($min_id); 
    $self->{max_id} = $max_id if defined($max_id);

    # get NOT set
    unless ( defined($min_id) or defined($max_id) or
	     defined($min_date) or defined($max_date) ) { 
	# figure out min and/or max if we can, return undef otherwise

	#return ($self->{min_id}, $self->{max_id});
    }

    # die unless we have a date index!

    #die "NYI";
    return ($min_id, $max_id);
}

sub indexed_fields {
    my $self = shift;
    my @f = keys %{$self->{field}};
    return @f unless wantarray;
    return sort @f;
}

sub indexed {
    my($self, $field) = @_;
    return !!$self->{field}{$field};
}

sub reaper {
    my $zombie;
    our %Kid_Status; # store exit status?
    # $zombies = 0;
    while ( ($zombie = waitpid(-1, WNOHANG)) != -1 ) {
	$Kid_Status{$zombie} = $?;
    }
}

sub dump {
    my $self = shift;
    #print STDERR "my opts are @_\n";
    my %opts = @_;
    my $fields = $opts{fields};
    my $form = $opts{form};
    my $index_type = lc($opts{type});
    $index_type ||= "forward";
    my @itype = split ",", $index_type;
    my $count = 0;
    my @indices = (defined($fields) && (ref($fields) eq "ARRAY") && @{$fields}) ? 
	           @{$fields} : sort grep !/INDEX|FIELD/o, (keys %{$self->{index}});
    $form ||= "keyval_dump";
    my ($dump_rdb, $dump_mdb, $dump_fdb) = ( grep $_ eq "all", @itype );
    $dump_rdb ||= ( grep $_  eq "reverse", @itype );
    $dump_mdb ||= ( grep $_  eq "master",  @itype );
    $dump_fdb ||= ( grep $_  eq "forward", @itype );

    if ( $dump_mdb ) {
	print "master = (\n" unless ( $form eq "key" );
	$self->{index}{mdb}{db}->dump($form);
	print ");\n" unless ( $form eq "key" );
	$count++;
    }

    for my $iname (@indices) {
	next if ( $iname eq "mdb" );
	print "\n" if $count++;
	print "$iname = (\n" unless ( $form eq "key" );
	unless ($self->{index}{$iname}{enabled}) {
	    print " # DISABLED INDEX\n);\n";
	    next;
	}

	$self->{index}{$iname}{db}->dump($form) if $dump_fdb;
	if ( $dump_rdb ) {
	    print ");\nreverse $iname = (\n";
	    $self->{index}{$iname}{rdb}->dump($form);
	}
	
	print ");\n" unless ( $form eq "key" );
    }
}

# get the count of messages indexed
sub get_count {
    my $self = shift;
    my $mdb = $self->{index}{mdb}{db};
    my ($ids, $fields) = $mdb ? $mdb->get_count() : ();
    return $ids;
}

sub gen_report {
    my $self = shift;
    my %args = @_;
    my ($index, $keys, $ids, $ooky, $ooid, $dupky, $dupid, $num_msg, $num_key);
 
    my $flush    = exists($args{flush}) ? delete($args{flush}) : 0;
    my $chk_ordr = exists($args{chk_ordr}) ? delete($args{chk_ordr}) : 0;
    my $missing  = exists($args{find_missing}) ? delete($args{find_missing}) : 0;
    # get count and message ids
    my @indices = (exists($args{indices}) && @{$args{indices}}) ? @{$args{indices}} : 
	sort grep !/INDEX|FIELD/, (keys %{$self->{index}});
    # print("My indices are: ", (join ", ", @indices), "\n");

    foreach my $a (keys %args) {
	warn "$$ cdb::gen_report: Useless argument $a=" . $args{$a};
    }

    my $report;
    if ( $chk_ordr ) {
	$report = select(STDOUT);
	$~ = "FULL";
	$^ = "FULL_TOP";
	select($report);
    } else {
	$report = select(STDOUT);
	$~ = "BRIEF";
	$^ = "BRIEF_TOP";
	select($report);
    }

    my $mdb;
    if ( $flush ) {
	$mdb = PureMessage::MessageStore::Dir::DB->new( dir      => $self->{dir},
							db_name  => "i_mdb",
							writable => 1,
							flush    => $flush,
							verbose  => $self->{verbose},
							num_key  => 1,
						      );
    } else {
	$mdb = $self->{index}{mdb}{db};
    }
    ($keys, $ids) = $mdb->get_count();
    $num_msg = $keys;

    for $index ( @indices ) {
	next if ( $index eq "mdb" );
	if ( exists($self->{index}{$index}) && 
	     exists($self->{index}{$index}{numeric_keys}) ) {
	    $num_key = $self->{index}{$index}{numeric_keys};
	} else {
	    $num_key = 0;
	}
	my $db;
	if ( $flush ) {
	    $db = PureMessage::MessageStore::Dir::DB->new( dir      => $self->{dir},
							   db_name  => "i_$index",
							   writable => 1,
							   flush    => $flush,
							   verbose  => $self->{verbose},
							   num_key  => $num_key,
							  );
	} else {
	    $db = $self->{index}{$index}{db};
	}
	($keys, $ids) = $db->get_count();
	($ooky, $ooid, $dupky, $dupid) = $chk_ordr ? $db->chk_ordr($chk_ordr) : (0) x 4;
	write STDOUT;
    }
    $index = "master";
    ($keys, $ids) = $mdb->get_count();
    my ($missing_ids, $m_ids) = $missing ? $mdb->find_missing() : ([], 0);
    ($ooky, $ooid, $dupky, $dupid) = $chk_ordr ? $mdb->chk_ordr($chk_ordr) : (0) x 4;
    write STDOUT;
    if ( my $m = scalar(@{$missing_ids}) ) {
	open(FILE, ">/tmp/missing_ids.txt") or die "Can't open /tmp/missing_ids.txt: $!";
	print FILE <<"MISSING";
$m MISSING IDS FROM RANGE OF $m_ids IDS:
@{$missing_ids}
MISSING
    close(FILE);
    }

    # write report
    format FULL_TOP =

Total Messages Indexed = @##########
                         $num_msg

FIELD NAME                   KEYS   MESSAGE IDS   OO KYS   OO IDS  DUP KYS  DUP IDS
--------------------|-------------|-------------|--------|--------|--------|--------|
.

    format BRIEF_TOP =

Total Messages Indexed = @##########
                         $num_msg

FIELD NAME                   KEYS   MESSAGE IDS  
--------------------|-------------|-------------|
.

    format FULL =
@<<<<<<<<<<<<<<<<<< | @########## | @########## | @##### | @##### | @##### | @##### |
$index,               $keys,        $ids,         $ooky,   $ooid,   $dupky,  $dupid
.

    format BRIEF =
@<<<<<<<<<<<<<<<<<< | @########## | @########## |
$index,               $keys,        $ids,        
.

}

1;
