package PureMessage::MessageStore::Dir::Index::findex;

use strict;
use ActiveState::Config::Simple qw(config_parse_file);
use Carp qw(carp croak);

use Fcntl qw(O_CREAT O_RDWR O_RDONLY LOCK_SH LOCK_EX);
use Digest::MD5 qw(md5_base64);
use File::Path qw(mkpath rmtree);
sub NUM_LEVELS() { 2 }		# change for deeper nesting of index tree

use PureMessage::MessageStore::Dir::Fields;
sub FIELDS () { "PureMessage::MessageStore::Dir::Fields" }

sub new {
    my($class, $dir, %opts) = @_;

    my $writable = delete $opts{writable};
    my $truncate = delete $opts{truncate};
    my $verbose = delete $opts{verbose};
    my $conf_file = delete $opts{config} || "index.conf";
    my $alt_conf = delete $opts{alt_config};

    $conf_file = "$dir/$conf_file" unless $conf_file =~ m,^/,;
    my $conf = config_parse_file($conf_file);
    $conf ||= config_parse_file($alt_conf) if defined $alt_conf;
    $conf ||= {};

    if ($^W) {
	for (keys %opts) {
	    carp("Unrecognized option '$_'");
	}
    }

    my $index;
    my %field;

    if ($index = $conf->{index}) {
	for my $iname (keys %$index) {
	    my $h = $index->{$iname};
	    if (exists $h->{field}) {
		my $f = $h->{field};
		$h->{field} = [$f] unless ref($f) eq "ARRAY";
	    }
	    else {
		$h->{field} = [$iname];
	    }
	    
	    $h->{enabled} = 1 unless exists $h->{enabled};

	    # intepret bools
	    for my $b ($h->{ordered}, $h->{enabled}) {
		$b = defined($b) && lc($b) =~ /^yes|on|true|1$/;
	    }

	    next unless $h->{enabled};

	    for my $f (@{$h->{field}}) {
		die "Bad field '$f' for index '$iname' in $conf_file"
		    unless FIELDS->field_ok($f);
		push(@{$field{$f}}, $iname);
	    }

	    if ($writable) {
		my $db = _index_dir($dir, $iname);
		rmtree($db) if $truncate; # XXX implement using rename and find
		mkpath($db); # will croak if $db doesn't exist and can't be created
		$h->{db} = $db;
	    }
	}
    }
    else {
	$index = {};
    }

    bless {
	dir => $dir,
	index => $index,
	field => \%field,
	rdonly => !$writable,
	verbose => $verbose,
    }, $class;
}

sub DESTROY {
    my $self = shift;
    # make sure database is closed
    delete $self->{index};
}

sub done {
}

sub index {
    my($self, $msg, $forget) = @_;
    croak("Can't index message for readonly index '$self->{dir}'")
	if $self->{rdonly};

    my $id = $msg->id;
    print "Indexing message $id\n" if $self->{verbose};
    for my $iname (keys %{$self->{index}}) {
	my $h = $self->{index}{$iname};
	next unless $h->{enabled};
	my $db = $h->{db};
	for my $field (@{$h->{field}}) {
	    for my $val (FIELDS->field($msg, $field)) {
		next unless defined($val);
		if ($self->{verbose}) {
		    require Data::Dump;
		    print "  $iname.$field = ", Data::Dump::dump($val), "\n";
		}
		my $hdir = _hval_dir($val);
		if ($forget) {
		    unlink("$db/$hdir/$id");
		    rmdir("$db/$hdir");
		}
		else {
		    unless (symlink($msg->filename, "$db/$hdir/$id")) {
		        mkpath("$db/$hdir");
			symlink($msg->filename, "$db/$hdir/$id");
		    }
		}
	    }
	}
    }
}

sub forget {
    my($self, $msg) = @_;
    $self->index($msg, 1);
}

sub _index_dir {
    my($dir, $index) = @_;
    $dir = $dir->{dir} if ref($dir);
    # we use a separate directory for all findexes, so that
    # it can be conveniently mounted as a different filesystem type
    "$dir/fi_ix/$index";
}

sub _hval_dir {
    my $value = shift;
    my $hval = md5_base64($value);
    $hval =~ tr{/}{_};
    my $hdir = substr($hval, 0, NUM_LEVELS);
    $hdir =~ s,(.),$1/,g;
    return "$hdir$hval";
}

sub find {
    my($self, $field, $value) = @_;
    my $idx = $self->{field}{$field};
    croak("No index for field '$field'") unless $idx;
    for my $iname (@$idx) {
	next if $self->{index}{$iname}{ordered};
	$idx = $iname;
	last;
    }
    $idx = $idx->[0] if ref($idx);
    print "Using index '$idx' for '$field'\n" if $self->{verbose};
    my $db = $self->_index_dir($idx);
    my $hdir = _hval_dir($value);
    if (opendir(my $d, "$db/$hdir")) {
        return grep !/^\./, readdir($d);
    }
    return;
}

sub range {
    my($self, $field, $min, $max) = @_;
    die "NYI";
}

sub indexed_fields {
    my $self = shift;
    my @f = keys %{$self->{field}};
    return @f unless wantarray;
    return sort @f;
}

sub indexed {
    my($self, $field) = @_;
    return !!$self->{field}{$field};
}

sub dump {
    my $self = shift;
    require Data::Dump;
    require File::Find;
    my $count;
    my @indices = @_ ? @_ : sort keys %{$self->{index}};
    for my $iname (@indices) {
	my $ordered = $self->{index}{$iname}{ordered};
	print "\n" if $count++;
	print $ordered ? '@' : '%';
	print "$iname = (\n";

	if (!$self->{index}{$iname}{enabled}) {
	    print " # DISABLED INDEX\n);\n";
	    next;
	}

	my $db = $self->_index_dir($iname);
	File::Find::find({
	    wanted => sub {
	        #print "[$_]\n";
		return unless -d $_;
		return if /^\./;
		return if /^[+\w]\z/;
		$File::Find::prune = $File::Find::prune = 1;
		return unless /^[+\w]{22}\z/;
		# XXX could find out actual $key by looking inside
		# the first message file
		my $key = Data::Dump::dump($_);
		if (opendir(my $ID, $_)) {
		    print " $key => [",
			  join(", ", sort(grep !/^\./, readdir($ID))),
			  "],\n";
		    closedir $ID;
		}
		else {
		    print "Can't opendir '$_', skipping: $!";
		}
	    }
	}, $db);
	print ");\n";
    }
}

1;
