package PureMessage::MessageStore::MessageSet;

use strict;
use Carp qw(croak carp);
use PureMessage::MessageStore::MessageSet::Row;
use PureMessage::MessageStore::MessageSet::QueryParser qw(qp_gen_id qp_gen_expr);

sub DEFAULT_LIMIT { 20 }

# Override in derrived sub-classes.
sub _compile_query;

eval {
    require ActiveState::StopWatch;
    ActiveState::StopWatch->import(qw(start_watch read_watch));
};
#warn $@ if $@;

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my %opts = @_;
    if (my $id = $opts{id}) {
	for (qw(where order_by group_by fields)) {
	    if (exists $opts{$_}) {
		delete $opts{$_};
		carp "Argument '$_' ignored when id specified";
	    }
	}
	return $class->restore(\%opts);
    }

    my $execute = 1;
    $execute = delete $opts{execute} if exists $opts{execute};
    $opts{limit} = DEFAULT_LIMIT unless defined $opts{limit};

    # disable the result set limit if grouping or sorting is enabled
    $opts{limit} = 0 if (defined $opts{order_by} && @{$opts{order_by}} || 
                         defined $opts{group_by} && @{$opts{group_by}});

    my $self = bless \%opts, $class;
    $self->execute if $execute;

    return $self;
}

sub execute {
    my $self = shift;

    my $store = $self->{message_store};
    $self->{watch} = start_watch() if defined &start_watch;
    $self->{stat} = {};
    $self->_compile_query;
    $self->_fetch_next_rows;


    if ($self->{watch}) {
	my $t = read_watch($self->{watch});
	$self->{stat}{sort_time} = $t;
	$store->log("INFO", "mset-sort times: $t");
    }

    return scalar(@{$self->{rows}});
}


sub _fetch_next_rows {
    my $self = shift;
    my $query = $self->{query_info};
    use Data::Dumper;
    my $match_sub = $query->{match_sub};
    my $watch = $query->{watch};
    #my $more_rows = 1;
    $query->{found_count} = 0;
    return undef unless $query->{more_rows};
    if ($query->{idx_ids}) {
	
	# Scan through the messages picked up by the index. Only look at 
	# messages that we've not yet seen.
	my $start = $query->{scan_count};
	my $end = @{$query->{idx_ids}} - 1;
	foreach my $id (@{$query->{idx_ids}}[$start .. $end]) {
   
	    my $msg = $self->{message_store}->message($id);
	    unless($msg) {
		$query->{scan_count}++;
		next;
	    }
	    eval {&$match_sub($msg)};
	    last if ($@ =~ /query limit reached/);
	    croak $@ if $@; 
	}
	$query->{more_rows} = 0 if $query->{scan_count} >= scalar @{$query->{idx_ids}};
    }
    else {
	# scan database
	# Increment the min_id closer to the max_id each scan. If in
	# reverse mode then decrement the max_id closer to the min_id.
	my $min = $query->{min_id};
	my $max = $query->{max_id};
	if ($query->{'reverse'}) {
	    $max = $query->{max_id} < $query->{last_id} ?
		$query->{max_id} : $query->{last_id};
	}
	else {
	    $min = $query->{min_id} > $query->{last_id} ?
		$query->{min_id} : $query->{last_id};
	}
	
	eval {
	    $self->{message_store}->message_scan($match_sub, $self->{folder}, 
		$min, $max, $query->{'reverse'})
	};
	croak $@ if $@ && $@ !~ /query limit reached/;
	$query->{more_rows} = 0 unless $@;
    }

    if ($watch) {
	my $t = read_watch($watch);
	$self->{stat}{scan_time} = $t;
	$self->{message_store}->log("INFO", "mset-scan times: $t");
	$watch = start_watch();
    }
    $self->{stat}{scan_msg} = $self->{rstat_scan_msg};
    $self->{message_store}->log("INFO", "mset-scan count: ${$self->{rstat_scan_msg}}")
	if exists $self->{rstat_scan_msg};
    
    my @extract_fields = @{$self->{x_fields}};
    if (@{$self->{rows}}) {
	# build an index map
	my %xfield;
	for (my $i = 0; $i < @extract_fields; $i++) {
	    $xfield{$extract_fields[$i]} = $i;
	}

	# colapse rows
	if ($self->{group_by}) {
	    _sort($self->{rows}, $self->{group_by}, \%xfield, $self->{verbose});
	    my @x = map $xfield{$_}, @{$self->{group_by}};
	    _collapse($self->{rows}, \@x);
	}

	# sort them
	if ($self->{order_by}) {
	    _sort($self->{rows}, $self->{order_by}, \%xfield, $self->{verbose});
	}

	# throw away fields not requested
	my @fields = @{$self->{fields} || []};
	$self->{num_fields} = @fields;
	if (@extract_fields && join(",", @extract_fields) ne join(",", @fields)) {
	    for my $row (@{$self->{rows}}) {
		my @x = splice(@$row, 0, @extract_fields);
		splice(@$row, 0, 0, @x[map $xfield{$_}, @fields]);
	    }
	}
    }
    else {
	$self->{num_fields} = 0;
    }
    
    return $query->{more_rows};
}

# Grouping messages involves packing messages in a group into a
# single row.  This is achieved by tacking on message IDs to
# the end of the first row of that group.  
#
# For example:
#   [qw(a b c alpha 1000)]
#   [qw(a x d alpha 1002)]
#   [qw(q x e alpha 2003)]
#   [qw(q x f beta  2005)]
#   [qw(q y f beta  2006)]
#
# Grouped on the 4th (index=3) column gives:
#   [qw(a b c alpha 1001 1002 2003)]
#   [qw(q x f beta  2005 2006     )]
#
# Assumptions:
# * rows contain the id in the last column
# * rows are sorted by the grouping columns
# * rows are not already grouped
# * the grouping values do not contain \000.
#
# $rows is modified in-place 
sub _collapse {
    my ($rows,$columns) = @_;
    my ($key,$last_key);
    my ($row,$last_row);
    my @new_rows;

    while (my $row = shift @$rows) {
	my $key = join("\0",@$row[@$columns]);
	if ($last_row && $key eq $last_key) {
	    # Push the grouped message-id onto the previous row
	    push(@$last_row, $row->[-1]);
	}
	else {
	    $last_key = $key;
	    $last_row = $row;
	    push @new_rows, $row;
	}
    }

    @$rows = @new_rows;
    return $rows;
}

sub _sort {
    my($rows, $sort_by, $x_fields, $verbose) = @_;
    my @cmp;
    for my $f (@$sort_by) {
	my $desc = ($f =~ s/\s+desc$//);
	my($a, $b) = $desc ? ('$b', '$a') : ('$a', '$b');
	for ($a, $b) {
	    $_ .= "->[$x_fields->{$f}]";
	}
	my $cmp = "cmp";
	$cmp = '<=>' if $f eq "m_spam_probability" || $f =~ /_size$/;
	push(@cmp, "$a $cmp $b");
    }
    my $sort = "@\$rows = sort { " . join(" || ", @cmp) . " } @\$rows;\n";
    print $sort if $verbose;

    # don't want noise if we happen to sort fields containing NULL values
    no warnings 'uninitialized';
    eval $sort;
    die $@ if $@;

    return;
}

sub _file {
    my $self = shift;
    $self->{message_store}->dir . "/mset/" . $self->{id} . ".sto";
}

sub restore {
    my($class, $opts) = @_;
    my $self = bless { id => $opts->{id}, message_store => $opts->{message_store} }, $class;
    require Storable;
    my $tmp = Storable::retrieve($self->_file);
    $tmp->{message_store} = $opts->{message_store};
    $tmp->_compile_query;
    return $tmp;
}

sub rows {
    my $self = shift;
    return undef unless $self->{rows};  # not executed yet
    scalar(@{$self->{rows}});
}

sub where    { shift->{where} }
sub fields   { shift->{fields}   || [] }
sub group_by { shift->{group_by} || [] }
sub order_by { shift->{order_by} || [] }

sub release {
    my $self = shift;
    my $file = $self->{id} ? $self->_file : "";
    delete $self->{message_store};
    delete $self->{rows};
    unlink($file) if $file;
}

sub save {
    my $self = shift;
    my $id = shift || $self->{id} || die "NYI: make up some numeric id";
    require Storable;
    $self->{id} = $id;
    my $file = $self->_file;
    my $store = delete $self->{message_store};  # avoid storing store

    # don't store the match sub
    my $ms = delete $self->{query_info}{match_sub} 
	if exists $self->{query_info}{match_sub};
    eval {
	$id = "" unless Storable::nstore($self, $file);
    };
    $self->{message_store} = $store;
    $self->{query_info}{match_sub} = $ms if $ms;
    die $@ if $@;
    return $id;
}

sub id {
    my $self = shift;
    return $self->{id};
}

# The fetch() function will request that more rows be added to the result set
# if the index is greater than the current number of matched rows. fetch()
# will return undef if there are no more rows that match and the index is out
# of bounds.
#
# In array context the actual number of rows returned may be less than 
# requested if there are an insuficient number of rows that match.
#
# XXX This needs to be reflected in the POD
#
sub fetch {
    my($self, $i, $n) = @_;
    $n ||= 1;
    my $nrows = @{$self->{rows}};
    return unless $nrows;

    # Fetch a row from the end if $i < 0
    # DEPRECATED in 4.0.4. The message set my not have the full list
    # of rows at this point.
    # $i += $rows if $i < 0;
    croak "Cannot fetch with a negative index." if $i < 0;

    # Fetch more rows until we have enough to satisfy the index or
    # the index + row count or we run out of rows.
    while ($i >= $nrows || $i + $n > $nrows) {
	my $more = $self->_fetch_next_rows;
	$nrows = @{$self->{rows}};
	last unless $more;
    }
    return undef if $i >= $nrows;

    my $num_fields = $self->{num_fields};
    if (wantarray) {
	$n = $nrows - $i if $i + $n > $nrows;
	return map PureMessage::MessageStore::MessageSet::Row::_make($_, $num_fields),
	           @{$self->{rows}}[$i .. ($i + $n - 1)];
    }
    else {
	croak("Multiple rows can't be returned in scalar context")
	    if $n > 1;
	return PureMessage::MessageStore::MessageSet::Row::_make($self->{rows}[$i],
								 $num_fields);
    }
}

#
# Match sub
#

sub _generate_match_sub {
    my($self, $expr, $x_fields) = @_;
    my @code;
    push(@code, <<'EOT');
#line 0 "match_sub (generated)"
no warnings 'uninitialized';
sub {
    my $m = shift;
    #print "Visit $id\n";
    $stat_scan_msg++;
    $$query_info{scan_count}++;
EOT

    if ($expr) {
	push(@code, "    return unless ");
	push(@code, qp_gen_expr($expr));
	push(@code, ";\n");
    }
    
    push(@code, <<'EOT');
    $$query_info{last_id} = $m->id;
    if ($$query_info{limit} && ++$$query_info{found_count} > $$query_info{limit}) {
	#print "Reached the limit at $$query_info{found_count}\n";
	$$query_info{scan_count}--;
	croak "query limit reached";
    }
EOT

    push(@code, '    push(@$rows, ');
    if (@$x_fields) {
	push(@code, '[');
	for (@$x_fields) {
	    push(@code, 'scalar(', qp_gen_id($_), '), ');
	}
	push(@code, '$m->id]');
    }
    else {
	push(@code, '$m->id');
    }
    push(@code, ");\n");
    push(@code, "}\n");
    return join("", @code);
}


1;

