package PureMessage::MessageStore::MessageSet::QueryParser;
use strict;
use warnings;
use base 'Exporter';

our @EXPORT_OK = qw(
    extract_ands
    where_parse
    qp_str_val
    qp_id_val
    qp_gen_id
    qp_gen_expr
    );



# Alias the public interface subs to the internal ones as they are really just
# the same (for now).
*extract_ands = \&_extract_ands;
*where_parse  = \&_where_parse;
*qp_gen_expr  = \&_gen_expr;
*qp_gen_id    = \&_gen_id;
*qp_id_val    = \&_id_val;
*qp_str_val   = \&_str_val;

sub _str_val {
    my ($val, $field) = @_;
    die "$field must be compared with string, not $val->[0]" unless $val->[0] eq "str";
    return $val->[1];
}

sub _id_val {
    my $val = shift;
    die "id must be compared with a number, not $val->[0]" unless $val->[0] eq "num";
    return $val->[1];
}

sub _gen_id {
    my $id = shift;
    
    # XXX
    # Use the MessageStore::Dir::Fields package to get the field codes.
    # There should be a more flexable way to specify this.
    require PureMessage::MessageStore::Dir::Fields;
    return PureMessage::MessageStore::Dir::Fields->field_code($id, '$m') ||
	   die "Bad field name '$id'";
}

sub _gen_expr {
    my $expr = shift;
    for my $op ($expr->[0]) {
	return join(" ", "!(", _gen_expr($expr->[1]), ")")
	    if $op eq "not";
	return join(" ", "(", _gen_expr($expr->[1]), $op, _gen_expr($expr->[2]), ")")
	    if $op eq "and" || $op eq "or";
	return "1" if $op eq "like"    && $expr->[2][1] eq "%";  # special case
	return "0" if $op eq "notlike" && $expr->[2][1] eq "%";  # special case
	return 'do { my @v = ' . _gen_id($expr->[1]) . '; scalar(@v) }' if $op eq "any";
 
	if ($expr->[3]) {
	    my @code;
	    push(@code, 'do { my @v = ',
		        _gen_id($expr->[1]),
		        '; my $found = 0; for (@v) { $found++, last if ');
	    if ($expr->[2][0] eq "null") {
		push(@code, _gen_null_expr(undef, $op));
	    }
	    else {
		push(@code, '$_ ', _gen_op_val($op, $expr->[2]));
	    }
	    push(@code, ' }');
	    if ($expr->[3] eq 'any') {
		push(@code, ' @v && $found');
	    }
	    elsif ($expr->[3] eq 'all') {
		push(@code, ' @v && !$found');  # test is negated for 'all'
	    }
	    else {
		die "Assert: Unexpected modified $expr->[3]";
	    }
	    push(@code, ' }');
	    return join("", @code);
	}
	else {
	    return _gen_null_expr($expr->[1], $op) if $expr->[2][0] eq "null";
	    return join(" ", "(", _gen_id($expr->[1]), _gen_op_val($op, $expr->[2]), ")");
	}
    }
}

sub _gen_null_expr {
    my($id, $op) = @_;
    die "Can't compare NULL using $op" unless $op eq "==" || $op eq "!=";
    return ($op eq "==" ? "!" : "") . "defined(" . _gen_id($id) . ")";
}


my %STR_OP = (
    '==' => 'eq',
    '<'  => 'lt',
    '>'  => 'gt',
    '<=' => 'le',
    '>=' => 'ge',
    '!=' => 'ne',
);

sub _gen_op_val {
    my($op, $val) = @_;
    if ($op =~ /^(not)?(r?)like$/) {
	return ($1 ? "!~" : "=~") . " " .
	       ($2 ? _gen_re($val->[1]) : _gen_re_glob($val->[1]));
    }
    die "Unknown operator '$op'" unless exists $STR_OP{$op};
    if ($val->[0] eq "str") {
	my $str = $val->[1];
	$str =~ s/\\/\\\\/g;
	$str =~ s/@/\\@/g;
	$str =~ s/\$/\\\$/g;
	$str =~ s/\"/\\\"/g;
	$str =~ s/\r/\\r/g;
	$str =~ s/\n/\\n/g;
	$str =~ s/\t/\\t/g;
	return qq($STR_OP{$op} "$str");
    }

    return "$op $val->[1]";
}

sub _gen_re_glob {
    my $str = quotemeta(shift);
    $str =~ s/_/./g;
    $str =~ s/\\%/.*/g;
    $str = "^$str" unless $str =~ s/^\.\*//;
    $str .= "\\z" unless $str =~ s/\.\*\z//;
    return qq(/$str/i);
}

sub _gen_re {
    my $str = shift;
    $str =~ s,([\@\$/]),\\$1,g;
    return qq(/$str/);
}

#
# Syntax tree extraction and manipulation
#
sub _extract_ands {
    my $expr = shift;
    return unless $expr;
    if ($expr->[0] eq "and") {
	return ($expr->[1], _extract_ands($expr->[2]))
    }
    else {
	return $expr;
    }
}


#
# The WHERE expression parser
#

my %WHERE_KEYWORD   = map { $_ => 1 } qw(and or not any all like rlike is between in null);
my %WHERE_VALID_OP  = map { $_ => 1 } qw(== < > <= >= != like rlike between is);
my %WHERE_VALID_CO  = map { $_ => 1 } qw(and or);

my %NEGATED_OP = (
    '==' => '!=',
    '!=' => '==',
    '<'  => '>=',
    '>'  => '<=',
    '<=' => '>',
    '>=' => '<',
    'like' => 'notlike',
    'rlike' => 'notrlike',
    'between' => 'notbetween',
);

sub _where_parse {
    my $where = shift;
    return undef unless defined($where) && $where =~ /\S/;
    my $tokens = [$where];
    my $expr = _where_parse_expr($tokens);
    my $t = _where_get_token($tokens);
    die "Syntax error: $t->[0] when expected EOF"
	if $t->[0] ne "eof";
    return $expr;
}

sub _where_parse_expr {
    my $tokens = shift;
    my $term = _where_parse_term($tokens);
    my $t = _where_get_token($tokens);
    return [$t->[0], $term, _where_parse_expr($tokens)]
	if $WHERE_VALID_CO{$t->[0]};
    _where_unget_token($tokens, $t);
    return $term;
}

sub _where_parse_term {
    my $tokens = shift;
    my $t = _where_get_token($tokens);
    if ($t->[0] eq "(") {
	my $expr = _where_parse_expr($tokens);
	_where_get_token($tokens)->[0] eq ")" or die "Missing ')'";
	return $expr;
    }
    if ($t->[0] eq "not") {
	return ["not", _where_parse_term($tokens)];
    }

    my $mod;
    if ($t->[0] eq "any" || $t->[0] eq "all") {
	$mod = $t->[0];
	$t = _where_get_token($tokens);
    }
    
    if ($t->[0] eq "id") {
	my $neg;
	my $op = _where_get_token($tokens);
	if ($op->[0] eq "not") {
	    $neg++;
	    $op = _where_get_token($tokens);
	}
	if ($mod && $mod eq "all") {
	    $neg = !$neg;
	}
	unless ($WHERE_VALID_OP{$op->[0]}) {
	    if ($mod && $mod eq "any") {
		# any foo
		return ["any", $t->[1]];
	    }
	    else {
		die "Syntax error: got $op->[0] when expected operator";
	    }
	}
	if ($op->[0] eq "is") {
	    my $tmp = _where_get_token($tokens);
	    if ($tmp->[0] eq "not") {
		$op->[0] = "!=";
	    }
	    else {
		$op->[0] = "==";
		_where_unget_token($tokens, $tmp);
	    }
	}

	if ($neg) {
	    if (my $tmp = $NEGATED_OP{$op->[0]}) {
		$op->[0] = $tmp;
	    }
	    else {
		die "Can't negate operator $op->[0]";
	    }
	}

	if ($op->[0] =~ /^(not)?between$/) {
	    die "\U$mod\E BETWEEN not implemented yet" if $mod;
	    my $from = _where_parse_value($tokens);
	    my $and = _where_get_token($tokens);
	    die "Syntax error: 'and' expected" unless $and->[0] eq "and";
	    my $to = _where_parse_value($tokens);
	    die "NULL can not be used as operand of BETWEEN"
		if $from->[0] eq "null" || $to->[0] eq "null";
	    return ["and", [">=", $t->[1], $from], ["<=", $t->[1], $to]]
		if $op->[0] eq "between";
	    return ["or", ["<", $t->[1], $from], [">", $t->[1], $to]];
	}

	return [$op->[0], $t->[1], _where_parse_value($tokens), ($mod ? $mod : ())];
    }
    die "Syntax error: $t->[0] in term";
}

sub _where_parse_value {
    my $tokens = shift;
    my $t = _where_get_token($tokens);
    return $t if $t->[0] eq "null" || $t->[0] eq "num" || $t->[0] eq "str";
    die "Syntax error: $t->[0] when expected value";
}

sub _where_get_token {
    my $tokens = shift;
    return pop(@$tokens) if @$tokens > 1;
    for ($tokens->[0]) {
	s/^\s+//;

	return ["eof"]
	    unless length;

	# identifiers are sometimes keywords
	if (s/^([a-zA-Z]\w*)//) {
	    my $id = lc($1);
	    return [$WHERE_KEYWORD{$id} ? $id : ("id", $id)];
	}

	return ["num", $1]
	    if s/^([-+]?(?:\.\d+|\d+(?:\.\d*)?)(?:[eE][-+]?\d+)?)//;

	# double char equal operators
	return [$1]
	    if s/^([<>!=]=)//;

	# some aliased operators
	return ["!="]
	    if s/^<>//;
	return ["=="]
	    if s/^=//;
	return ["and"]
	    if s/^&&//;
	return ["or"]
	    if s/^\|\|//;
	return ["not"]
	    if s/^!//;

	# strings can use single or double quotes and the quotes
	# themselves can be escaped by doubling them.
	if (s/^\"([^\"]*(?:""[^\"]*)*)\"//) {
	    my $str = $1;
	    $str =~ s/""/\"/g;
	    return ["str", $str];
	}
	if (s/^\'([^\']*(?:''[^\']*)*)\'//) {
	    my $str = $1;
	    $str =~ s/''/\'/g;
	    return ["str", $str];
	}

	# as a last resort grab a single char
	return [substr($_, 0, 1, "")];
    }
}

sub _where_unget_token {
    my $tokens = shift;
    push(@$tokens, @_);
}


1;

