package PureMessage::MessageStore::pmdb;

use strict;
use base 'PureMessage::MessageStore::Dir';
use Carp qw(croak carp);
use PureMessage::MessageFile ();
use PureMessage::MessageInfo ();
use ActiveState::I18N::UTF8 qw(utf8_length);
use DBI;

sub new {
    my $class = shift;
    my $dir = shift;
    my %opts = @_;

    my $create = delete $opts{create};
    my $verbose = delete $opts{verbose} || 0;
    my $store_class = delete $opts{class};
    my $dirmode = delete $opts{dirmode};
    my $logger = delete $opts{logger};
    my $msg_writable => delete $opts{message_writable} || 0;
    my $dsn = delete $opts{dsn};
    my $location = delete $opts{location} || "localhost";
    my $connect = delete $opts{connect};
    my $store_id = delete $opts{store_id} || "pmx_quarantine";

    my $default_max_body = 2000;
    my $default_raw_msg = 4000;
    my $max_body;
    my $raw_msg;
    if (%opts) {
	if ($logger) {
	    $logger->log("WARNING", "Unknown message store option '$_'") for keys %opts;
	}
	else {
	    carp "Unknown option '$_'" for keys %opts;
	}
    }

    if ($create) {
	for (qw(new cur tmp mset)) {
	    next if -d "$dir/$_";
	    $class->_mk_path("$dir/$_", $verbose, $dirmode) ||
		croak("Can't create directory '$dir/$_': $!");
	}
	$class->init_cfg_file(dir => $dir,
                              verbose => $verbose)
            unless (-f "$dir/db.conf");

    }
    else {
	if (-d $dir) {
	    my @missing;
	    for (qw(new cur tmp mset)) {
		push(@missing, $_) unless -d "$dir/$_";
	    }
	    if (@missing) {
		croak("Malformed database directory '$dir': missing subdirs: " .
		      join(",", map qq('$_'), @missing));
	    }
	    croak("Malformed database directory '$dir': missing db.conf")
		unless -f "$dir/db.conf";
	}
	else {
	    croak("No such directory '$dir'");
	}
    }

    my $conf = $class->SUPER::get_config($dir);
    $location = $conf->value("location");
    my $location_id = $conf->value("location_id");
    $max_body = $conf->value("max_body_size");
    my $insert_rule_hits = $conf->value("insert_rule_hits");
    $insert_rule_hits = 1 unless defined $insert_rule_hits;
    $max_body = $default_max_body unless defined $max_body;
    if ($max_body > $default_max_body) {
        $max_body = $default_max_body;
        $logger->log("WARNING",
            "Configuration parameter 'max_body_size' is larger than the " .
            "maximum value allowed; using the default ($default_max_body)")
            if $logger;
    }
    $raw_msg = $conf->value("raw_msg_size");
    $raw_msg = $default_raw_msg unless defined $raw_msg;
    if ($raw_msg > $default_raw_msg) {
        $raw_msg = $default_raw_msg;
        $logger->log("WARNING",
            "Configuration parameter 'raw_msg_size' is larger than the " .
            "maximum value allowed; using default ($default_raw_msg)")
            if $logger;
    }

    my $max_retries = $ENV{PMX_IDX_MAX_RETRIES} || 25;
    my $self = bless {
	pmxdir_dir => $dir,
	pmxdir_dsn => $dsn,
	pmxdir_store_id => $store_id,
	pmxdir_verbose => $verbose,
	pmxdir_dirmode => $dirmode,
	pmxdir_logger => $logger,
	pmxdir_writable => $msg_writable,
	pmxdir_concurrent_ok => 1,
	pmxdir_disk_used_delta => 0,
	pmxdir_num_msg_delta => 0,
	pmxdir_folder_id => { cur => 0 },
	pmxdir_reason_id => {},
	pmxdir_sth => {},  # cached statement handles
	pmxdir_location => $location,
	pmxdir_location_id => $location_id,
	pmxdir_max_body => $max_body,
	pmxdir_max_raw => $raw_msg,
	pmxdir_insert_rule_hits => $insert_rule_hits,
	pmxdir_commit_period => 1000,

        # for msg indexing retry mechanism
        pmxdir_max_retries => $max_retries,
        pmxdir_retry_needed => 0,
    }, $class;

    $self->_db_connect() if $connect;
    return $self;
}

# establish new db connection
sub connect_new {
    my ($self, $connect_cached) = @_;
    my $dsn = $self->{pmxdir_dsn};
    my $dbh;

    # NOTE: We ensure that all database handles open using exactly the same
    # connection arguments so that DBI handles can be shared among multiple
    # packages.  See the following docs for more details.
    # http://perl.apache.org/docs/1.0/guide/databases.html#Opening_Connections_With_Different_Parameters

    my $connect_method = $connect_cached ? "connect_cached" : "connect";
    $dbh = DBI->$connect_method($dsn, "", "",
			{
			    RaiseError => 0,
			    PrintError => 0,
			    AutoCommit => 0,
			});
    $dbh->{RaiseError}  = 1;
    $dbh->{PrintError}  = 0; 
    $dbh->{AutoCommit}  = 0; 
    $dbh->{HandleError} = undef;
    return $dbh;
}

# keeping this global prevents this variable from going through the
# expensive initialization; this helps in CGI environment
my $pmdb_cf;

sub _db_connect {
    my $self = shift;
    my $store_id = $self->{pmxdir_store_id};
    my $dsn = $self->{pmxdir_dsn};
    if (not $dsn) {
        require PureMessage::Database::Config;
        $pmdb_cf = PureMessage::Database::Config->new() unless $pmdb_cf;
        $dsn = $pmdb_cf->store_value($store_id, "dsn");
        $self->{pmxdir_dsn} = $dsn;
    }

    my $connect_cached = $ENV{MOD_PERL} ? 0 : 1;
    return $self->{pmxdir_dbh} = $self->connect_new($connect_cached);
}

sub DESTROY {
    my $self = shift;
    delete $self->{pmxdir_sth};  # close all cache statement handles
    if (my $dbh = $self->{pmxdir_dbh}) {
	$dbh->disconnect;
    }
    $self->_status_cache_sync;
}

sub store_id {
    "pmdb:" . shift->{pmxdir_dir};
}

sub location { shift->{pmxdir_location} }

sub add_index {
    my $self = shift;
    croak("Custom indexes not supported by the pmdb backend");
}

# Validates config parameters required for indexing; returns false if
# indexing can not proceed with this configuration.
sub _params_valid {
    my $self = shift;
    my $dsn      = $self->{pmxdir_dsn};
    my $dir      = $self->{pmxdir_dir};
    my $location = $self->{pmxdir_location};
    my $location_id = $self->{pmxdir_location_id};

    # location must exist and be a scalar value
    if ((not $location) or (ref $location)) {
        $self->log("ERROR",
                   "Missing or invalid parameter 'location' in $dir/db.conf");
        return 0;
    }

    # location, location_id may have default values only if we're
    # co-located with the DBMS
    if ( ($dsn =~ /host=/) and
         ($dsn !~ /\Qhost=$location\E/) ) {

        if ($location eq "localhost") {
            $self->log("ERROR",
                       "Invalid parameter 'location' in $dir/db.conf: " .
                       "unable to use default value ('localhost') when dsn is set to '$dsn'");
            return 0;
        }

        if ($location_id == 0) {
            $self->log("ERROR",
                       "Invalid parameter 'location_id' in $dir/db.conf: " .
                       "unable to use default value (0) when dsn is set to '$dsn'");
            return 0;
        }
    }

    return 1;
}

# This is called from meta_index() before beginning any indexing
# transactions. We need to make sure that we can connect to the 
# database before proceeding. Conceptually speaking, database
# is the index, so just return the db handle here. Croaks if parameters
# required for indexing are invalid.
sub _new_index { 
    my $self = shift;
    $self->_db_connect unless $self->{pmxdir_dbh};
    croak "Invalid configuration; see PMX log for more information"
        unless $self->_params_valid();
    return $self->{pmxdir_dbh};
}

sub get_user_views {
    my $self = shift;
    my $addr = shift;

    $self->_db_connect unless $self->{pmxdir_dbh};

    warn "Requesting multiple views in scalar context"
	if !wantarray && @_ > 1 && $^W;

    my $basep = "PureMessage::MessageStore::pmdb::UserView";
    require PureMessage::MessageStore::Dir::UserView;
    return PureMessage::MessageStore::Dir::UserView->new( 
	base_package => $basep, store => $self, address => $addr, 
	%{$_[0]}) unless wantarray;
    
    my @views;
    foreach my $v (@_) {
	push(@views, PureMessage::MessageStore::Dir::UserView->new(
	    base_package => $basep, store => $self, address => $addr, %$v));
    } 
    return @views;
}

sub _get_reason_id {
    my($self, $reason) = @_;
    my $id;
    if (defined $reason) {
        $reason = lc($reason);
        $id = $self->{pmxdir_reason_id}{$reason};
        
        # not eval-ing the db access here since the call of this sub is
        # eval-ed itself
        unless (defined $id) {
            my $dbh = $self->{pmxdir_dbh};
            ($id) = $dbh->selectrow_array(
                "SELECT reason_id FROM m_reason WHERE name = ?", undef, $reason);
            
            unless (defined $id) {
                $dbh->do(
                    "INSERT INTO m_reason (name) VALUES (?)", undef, $reason);
                ($id) = $dbh->selectrow_array(
                    "SELECT reason_id FROM m_reason WHERE name = ?", undef, $reason);
            }
            $self->{pmxdir_reason_id}{$reason} = $id;
        }
    }
    return $id;
}

# Wrapper for the _do_index(); encapsulates the retry and commit logic
# note that the reason for all of this complexity is the lack of support
# for savepoint in PGSQL 7.x. As soon as we migrate to 8.x this should
# be streamlined.
sub _index {
    my ($self, $message) = @_;
    my $message_id = $message->id;
    eval { $self->_do_index($message) };
    if (!$@) {
        if ($self->_push_indexed($message)) {
            $self->_flush_indexed();
        }
    }
    else {
        eval {
            my $do_die = !($self->{pmxdir_skip_message} or $self->{pmxdir_retry_needed});
            $self->_push_indexed($message) if ($self->{pmxdir_retry_needed});
            $self->_recover();
            $self->_flush_indexed();

            # die only if we encountered a real error
            if ($do_die) {
                die "Error while attempting to index message $message_id";
            }
        };
    }

    if ($@) {
        my $msg = "Unsuccessful recovery while attempting to index message $message_id: $@ ";
        $msg .= "Maximum count of attempts reached; giving up.";
        $self->log("ERROR", $msg);
        die $msg;
    }
}

# iterates through the list of successfully indexed but uncommitted
# messages (that we had to rollback due to the subsequent error) and
# indexes them again.
sub _recover {
    my $self = shift;
    my $attempt = 1;

    $self->{pmxdir_dbh}->rollback();

    my $max_retries = $self->{pmxdir_max_retries};
    while ($attempt <= $max_retries) {
        foreach my $id (@{$self->{pmxdir_indexed}}) {
            my $message = $self->message($id);
            eval {$self->_do_index($message)};
        }
        if ($@) {
            $self->{pmxdir_dbh}->rollback();
            $attempt++;
        }
        else {
            last;
        }
    }

    if ($attempt > $max_retries) {
        my $msg = "Unsuccessful recovery while meta-indexing: $@";
        $self->log("ERROR", $msg);
        die ($msg);
    }
}

# Adds a reference to the message to the list of successfully indexed
# (but not yet committed) messages; Returns true if the queue is full;
# false otherwise.
sub _push_indexed {
    my ($self, $message) = @_;
    push @{$self->{pmxdir_indexed}}, $message->id;
    return ($self->{pmxdir_commit_period} <= scalar @{$self->{pmxdir_indexed}});
}

sub _flush_indexed {
    my $self = shift;
    $self->{pmxdir_dbh}->commit();
    $self->{pmxdir_indexed} = [];
}

# called at the end of meta-indexing to commit the uncommitted insertions
# (if the count of meta-indexed messages is not evenly divisible by the
# count of messages per transaction).
sub done {
    my $self = shift;
    eval {$self->_flush_indexed();};
    if ($@) {
        $self->log("ERROR", "Error occurred while attempting the final commit: $@");
        eval {$self->{pmxdir_dbh}->rollback()};
    }
}

# does the actual indexing of the message
sub _do_index {
    my($self, $message) = @_;

    $self->_db_connect unless $self->{pmxdir_dbh};

    my $dbh = $self->{pmxdir_dbh};
    require PureMessage::Message::ComputedFields;
    my $message_id = $message->id;
    my $max_msg_cmpnnt_len = 255;
    
    eval {
        my $reason_id = $self->_get_reason_id($message->minfo_value("Reason"));
	croak "Undefined reason identifier for message " .
              $message->minfo_value("Global-ID")
	    unless $reason_id;

	# Get the basic message information (should all be scalars)
	my $m_global_id = $message->minfo_value("Global-ID");
	croak "Can't index a message that doesn't have a Global-ID." 
	    unless $m_global_id;
	my $h_message_id = $message->header("Message-ID");
	my $queue_id = $message->minfo_value("Queue-ID");
	my $subject = $message->c_subject_utf8;
	my $subject_body = $message->c_subject_body_utf8;
	my $spam_prob = $message->minfo_value("Spam-Probability") || '0.0';
	my ($env_from_local, $env_from_dom) = 
	    split_address($message->minfo_value("Envelope-From") || '');
	my $h_from = $message->header("From");    
	my ($h_from_local, $h_from_dom) = 
	    split_address(_clean_mail_addresses($h_from) || '');
	my $h_to = $message->header("To");    
	my $m_milter_host = $message->minfo_value("Milter-Host");
	my $m_relay = $message->minfo_value("Relay");
	my $message_size = $message->message_size();
	my $header_size = $message->header_size();
	my $location = $self->{pmxdir_location};
        my $direction = (lc($message->minfo_value('Direction')) eq 'outbound') ? 1 : 0;
	
	# Truncate values that are too long
	for ($h_message_id, $subject, $env_from_local, $env_from_dom, 
	    $h_from_local, $h_from_dom, $h_from, $h_to, $m_milter_host,
	    $m_relay, $queue_id) {

	    next unless defined;

	    # make sure they fit
            if (length($_) > $max_msg_cmpnnt_len) {
                substr($_, $max_msg_cmpnnt_len) = "";
            }
	}
        
	# Insert the data into the message table.
	my $insert_message = $self->{pmxdir_sth}{insert_message} ||=
	    $dbh->prepare("INSERT INTO message (m_global_id,
                                                location_name,
                                                m_queue_id,
                                                h_message_id,
                                                m_date,
                                                c_subject_utf8,
                                                c_subject_body_utf8,
                                                reason_id, 
						env_from_local,
						env_from_domain,
						h_from_local,
						h_from_domain,
						h_from,
						h_to,
						m_milter_host,
						m_relay,
						m_spam_probability,
						m_message_size,
						m_header_size,
                                                outbound,
                                                reason)
                           VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?,
                                   ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)");
	$insert_message->execute($m_global_id,
				 $location,
				 $queue_id,
				 $h_message_id,
				 scalar $message->c_qdate_iso,
				 $subject,
				 $subject_body,
				 $reason_id,
				 $env_from_local,
				 $env_from_dom,
				 $h_from_local,
				 $h_from_dom,
				 $h_from,
				 $h_to,
				 $m_milter_host,
				 $m_relay,
				 $spam_prob,
				 $message_size,
				 $header_size,
                                 $direction,
                                 lc($message->minfo_value("Reason")),
				);

	# Get the new (database) message id
	my $get_cur_id = $self->{pmxdir_sth}{get_cur_id} ||=
	    $dbh->prepare("select id from message where m_global_id = ?");
	$get_cur_id->execute($m_global_id);
	my ($cur_id) = $get_cur_id->fetchrow_array;
	$get_cur_id->finish;
	
        # as we go along, we accumulate the per-message recipient info
        # into this hash
        my %recipient;

	# "calculated" recipients
	foreach my $addr ($message->c_to) {
            $addr = _clean_mail_addresses($addr);
            $recipient{$addr}{is_calculated} = 1;
	}

	# envelope recipients
        my $env_rcpt_cnt = 0;
	foreach my $addr ($message->minfo_value("Envelope-To")) {
            $env_rcpt_cnt++;
            $addr = _clean_mail_addresses($addr);
            $recipient{$addr}{is_envelope} = 1;
	}

	$self->{pmxdir_logger}->log("WARNING",
                                    "No envelope recipients for message " .
                                    $message->minfo_value("Global-ID"))
            if ($self->{pmxdir_logger} and (0 == $env_rcpt_cnt));

	# mapped recipients
	if (my @mrecip = $message->minfo_value("Mapped-Recipients")) {
	    foreach my $r (@mrecip) {
		my $addr = _clean_mail_addresses($r);
                $recipient{$addr}{is_mapped} = 1;
	    }
	}

	# per-user message disposition
	if (my @enduser = $message->minfo_value("enduser")) {
	    foreach my $action (@enduser) {
		my ($act, $addr) = split(" ", $action, 2);
		$addr = _clean_mail_addresses($addr);
                $recipient{$addr}{action} = action_numeric($act)
                    if exists $recipient{$addr};
	    }
	}

	if ($self->{pmxdir_insert_rule_hits}) {
            $self->do_insert_rule_hits(message => $message,
                                       db_id => $cur_id,
                                       commit_needed => 0);
        }

        $self->_store_recipient_info(\%recipient, $cur_id,
                                     $message->minfo_value("Global-ID"), $reason_id);

        if($direction) {
            $self->_store_group_info($cur_id, $message->minfo_value("Sender-Group"));
        }
        else {
            my @mapped_recipients = $message->minfo_value("Mapped-Recipients");
            my %mapped_recipients_map = map { ($_, 1) } @mapped_recipients;
            my @recipient_group = $message->minfo_value("Recipient-Group");

            for (my $i = 0; $i < scalar(@recipient_group); $i++) {
                my ($key) = $recipient_group[$i] =~ /([^|]+)/;
                delete($recipient_group[$i]) unless exists($mapped_recipients_map{$key});
            }

            $self->_store_group_info($cur_id, @recipient_group);
        }
    };

    $self->{pmxdir_retry_needed} = 0;
    $self->{pmxdir_skip_message} = 0;

    if ($@) {
	my $err = $@;

	# When reindexing the quarantine it is possible that messages that
	# are already in the database will be added to the index_log. This
	# will cause the insert to fail when the global_id unique key is
	# violated. This is not normally an error (therefore we don't retry).
	if ($err =~ /violates unique constraint "unq_message_m_global_id"/) {
            my $msg = "Skipping index of message $message_id:"
		      . " message already exists in the database";
            $self->{pmxdir_skip_message} = 1;
	    $self->log("DEBUG", $msg);
            die $msg;
	}

        # in these cases we've hit a race condition and we want to retry
	elsif ($err =~ /deadlock detected/) {
            $self->{pmxdir_retry_needed} = 1;
            my $msg = "deadlock detected; retrying message $message_id";
	    $self->log("DEBUG", $msg);
            die $msg;
        }
	elsif ($err =~ /due to missing recipient/) {
            $self->{pmxdir_retry_needed} = 1;
            my $msg = "recipient dissapeared; retrying message $message_id";
	    $self->log("DEBUG", $msg);
            die $msg;
        }
	elsif ($err =~ /violates unique constraint "unq_recipient"/) {
            $self->{pmxdir_retry_needed} = 1;
            my $msg = "unq_recipient constraint violation; " .
                      "retrying message $message_id";
	    $self->log("DEBUG", $msg);
            die $msg;
        }

        # unknown error, no retry
	else {
	    die "Index $message_id failed: $err";
	}
    }
}

# interacts with the DBMS => needs to be evaled
sub do_insert_message_body {
    my($self, %opts) = @_;
    my $message = $opts{message};
    my $commit_needed = $opts{do_commit};
    my $cur_id = $opts{cur_id};
    my $global_id = $opts{global_id};
    my $dbh = $self->{pmxdir_dbh};

    my $c_body = $message->cleaned_body();
    substr($c_body, utf8_length($c_body, $self->{pmxdir_max_body})) = ""
        if length($c_body) > $self->{pmxdir_max_body};

    my $raw_msg = $message->raw_body() || "";
    substr($raw_msg, $self->{pmxdir_max_raw}) = "" 
        if length($raw_msg) > $self->{pmxdir_max_raw};

    # some DBMS drivers can't handle null characters in strings
    $raw_msg =~ s/\0//g;

    if ($cur_id) {
        my $stmt = $dbh->prepare('INSERT INTO message_body
                                  (message_id, c_body, raw_msg)
                                  values(?, ?, ?)');
        $stmt->execute($cur_id, $c_body, $raw_msg);
    }
    else {
        my $stmt = $dbh->prepare('INSERT INTO message_body
                                  (message_id, c_body, raw_msg)
                                  select id, ?, ?
                                  from message where m_global_id = ?');
        $stmt->execute($c_body, $raw_msg, $global_id);
    }

    if ($commit_needed) {
        $dbh->commit();
    }
}

# interacts with the DBMS => needs to be evaled
sub do_insert_rule_hits {
    my ($self, %opts) = @_;
    my $message = $opts{message};
    my $cur_id = $opts{db_id};
    my $do_commit = $opts{do_commit};
    my $interactive = $opts{interactive};
    my $global_id = $message->minfo_value("Global-ID");
    unless ($cur_id) {
        ($cur_id) = $self->{pmxdir_dbh}->selectrow_array(
            "select id from message where m_global_id = '$global_id'"
        );
    }

    # in interactive mode we don't build the in-memory rule-to-id map
    my $do_build_map = $interactive ? 0 : 1;
    my @hits = $message->minfo_value("Spam-Hit");
    if (scalar @hits) {
        $self->_store_qrntn_rule_hits($cur_id,
                                      $global_id,
                                      $self->_get_reason_id('spam'),
                                      \@hits,
                                      $do_build_map);
    }
    @hits = ($message->minfo_value("Virus-Id"));
    if (scalar @hits) {
        $self->_store_qrntn_rule_hits($cur_id,
                                      $global_id,
                                      $self->_get_reason_id('virus'),
                                      \@hits,
                                      $do_build_map);
    }
    if ($do_commit) {
        $self->{pmxdir_dbh}->commit();
    }
}

sub is_message_body_inserted {
    my ($self, $global_id) = @_;
    my $sql = <<EOT;
select 1 where exists (
    select 'x' from message, message_body
    where message.id = message_body.message_id and message.m_global_id = ?
)
EOT
    my $dbh = $self->{pmxdir_dbh};
    my $stmt = $dbh->prepare($sql);
    my $body_inserted = 0;
    $stmt->execute($global_id);
    $stmt->bind_columns(\$body_inserted);
    $stmt->fetch();
    return $body_inserted;
}

sub rule_hits_inserted {
    my ($self, $global_id) = @_;
    my $sql = <<EOT;
select 1 where exists (
    select 'x' from message, quarantine_rule_hit
    where message.id = quarantine_rule_hit.message_id and message.m_global_id = ?
)
EOT
    my $dbh = $self->{pmxdir_dbh};
    my $stmt = $dbh->prepare($sql);
    my $hits_inserted = 0;
    $stmt->execute($global_id);
    $stmt->bind_columns(\$hits_inserted);
    $stmt->fetch();
    return $hits_inserted;
}

# translates textual action codes to numbers; we're ignoring the first
# arg if it's a reference, so that the sub may be invoked through either
# "classic" or OO syntax
sub action_numeric {
    my $action;
    $action = shift if ref($action = shift);
    return 2 if ($action =~ /^approve/);
    return 3 if ($action =~ /^delete/);
    return 4 if ($action =~ /^forward/);
    return undef;
}

# translates numeric action codes to text
sub action_string {
    my $action;
    $action = shift if ref($action = shift);
    return "approve" if ($action == 2);
    return "delete" if ($action == 3);
    return "forward" if ($action == 4);
    return undef;
}

# we're ignoring the first arg if it's a reference, so that the sub
# may be invoked through either "classic" or OO syntax
sub split_address {
    my $addr;
    $addr = shift if ref($addr = shift);
    my $max_addr_cmpnnt_len = 255;
    my ($local, $domain) = split(/\@/, $addr, 2);
    $local = "" unless defined $local;
    $domain = "" unless defined $domain;
    substr($local, $max_addr_cmpnnt_len) = ""
        if length($local) > $max_addr_cmpnnt_len;
    substr($domain, $max_addr_cmpnnt_len) = ""
        if length($domain) > $max_addr_cmpnnt_len;
    return ($local, $domain);
}

# maps the specified quarantine rule patterns into corresponding numeric ids
# (tests show that this is overly expensive to do on the fly on the server)
sub _get_qrntn_rule_ids {
    my $self = shift;
    my $rule_numeric_ref = shift;
    my $rule_type = shift;
    my $rule_text_ref = shift;
    my $do_build_rule_map = shift;
    my %mapped;
    my $rule_id_ref;
    my $rule_map_ref = {};
    my $insert_stmt;
    my $select_stmt;

    @$rule_numeric_ref = ();

    # initialize cache, if needed
    if ($do_build_rule_map && !($rule_map_ref = $self->{pmxdir_rule_map})) {
        $rule_map_ref = $self->_get_qrntn_rule_map();
        $self->{pmxdir_rule_map} = $rule_map_ref;
    }
    $rule_map_ref->{$rule_type} = {} unless ($rule_map_ref->{$rule_type});
    $rule_id_ref = $rule_map_ref->{$rule_type};

    # Try to map the text pattern into numeric id based on the cached info;
    # if this fails then the pattern is not in the db yet => insert it
    # into the db, update cache and map it then.
    foreach my $pattern (@$rule_text_ref) {
        my $id;
        if ($id = ${$rule_id_ref}{$pattern}) {
            next if $mapped{$id};
            push @$rule_numeric_ref, $id;
            $mapped{$id} = 1;
        }
        else {
            # "double-checked db locking" pattern courtesy of Mischa S.
            my $dbh = $self->{pmxdir_dbh};
            unless ($insert_stmt) {
                $insert_stmt = $dbh->prepare(<<"EOT");
insert into quarantine_rule (reason_id, rule_text)
select ?::integer, ?::varchar
where not exists (
    select 'x' from quarantine_rule
    where reason_id = ? and rule_text = ?
)
EOT
                $select_stmt = $dbh->prepare(<<"EOT");
select id from quarantine_rule where reason_id = ? and rule_text = ?
EOT
            }

            # this is not going to happen very often
            $dbh->do("lock table quarantine_rule in exclusive mode");
            $insert_stmt->execute($rule_type, $pattern, $rule_type, $pattern);
            $select_stmt->execute($rule_type, $pattern);
            ($id) = $select_stmt->fetchrow_array();
            ${$rule_id_ref}{$pattern} = $id;
            push @$rule_numeric_ref, $id;
            $mapped{$id} = 1;
        }
    }
}

# retrieves the text-to-numeric rule map from the database and returns a
# reference to the map structure, which is a hash of references to hashes,
# with one hash for each rule type (or quarantine reason, if you prefer)
sub _get_qrntn_rule_map {
    my $self = shift;
    my %map;
    my ($id, $rule_type, $rule_text);
    my $dbh = $self->{pmxdir_dbh};
    my $stmt = $dbh->prepare("select id, reason_id, rule_text from quarantine_rule");
    $stmt->execute();
    $stmt->bind_columns(\$id, \$rule_type, \$rule_text);
    while ($stmt->fetch()) {
        $map{$rule_type} = {} unless $map{$rule_type};
        $map{$rule_type}->{$rule_text} = $id;
    }
    return \%map;
}

# this subroutine does not eval the DBMS access because it's evaled itself
# XXX PGSQL-specific code here in fromless selects
sub _store_qrntn_rule_hits {
    my $self = shift;
    my $cur_id = shift;
    my $global_id = shift;
    my $reason_id = shift;
    my $hits_ref = shift;
    my $do_build_rule_map = shift;
    my @rule_id;

    $self->_get_qrntn_rule_ids(\@rule_id, $reason_id, $hits_ref, $do_build_rule_map);
    my $hit_union =
        join " union all ",
        map  "select $cur_id as message_id, $_ as quarantine_rule_id", @rule_id;
    my $dbh = $self->{pmxdir_dbh};
    $dbh->do(<<"EOT");
insert into quarantine_rule_hit select * from ($hit_union) hit_union
EOT
}

sub _store_group_info {
    my ($self, $cur_id, @groups) = @_;
    my $dbh = $self->{pmxdir_dbh};

    return unless @groups;

    my $stmt = $dbh->prepare(<<EOT);
INSERT INTO message_group (message_id, group_name) VALUES (?, ?)
EOT

    my %group_names = ();
    for (@groups) {
        $_ =~ /.+\|(.+)/;
        $group_names{$1} = 1 if $1;
    }

    for my $group_name (keys %group_names) {
        $stmt->execute($cur_id, $group_name);
    }
}

# this subroutine also does not eval the DBMS access because it's evaled itself
sub _store_recipient_info {
    my $self = shift;
    my $recipient_ref = shift;
    my $cur_id = shift;
    my $global_id = shift;
    my $reason_id = shift;
    my $max_addr_cmpnnt_len = 255;
    my $dbh = $self->{pmxdir_dbh};

    my $ins_msg_recip_sth = $self->{pmxdir_sth}{insert_msg_recipient} ||=
        $dbh->prepare(<<"EOT");
insert into message_recipient
(recipient_id, message_id, is_mapped, is_calculated, is_envelope, reason_id, action)
select id, ?, ?, ?, ?, ?, ?
from recipient where local = ? and domain = ?
EOT
    my $ins_recip_sth = $self->{pmxdir_sth}{insert_recipient} ||=
        $dbh->prepare('insert into recipient (local, domain)
                       values (?, ?)');

    # see the above comments about race condition in
    # _store_qrntn_rule_hits(); the same thing applies here.
    while (my ($recipient, $info) = each %$recipient_ref) {
        my ($local, $domain) = split_address($recipient);

        my $recipient_inserted = 0;

        PER_RECIPIENT: while (1) {
            # the is_... columns can't be null
            $ins_msg_recip_sth->execute($cur_id,
                                        $info->{is_mapped} || 0,
                                        $info->{is_calculated} || 0,
                                        $info->{is_envelope} || 0,
                                        $reason_id,
                                        $info->{action},
                                        $local,
                                        $domain);

            last PER_RECIPIENT if $ins_msg_recip_sth->rows();

            if (not $recipient_inserted) {
                $ins_recip_sth->execute($local, $domain);
                $recipient_inserted = 1;
                next PER_RECIPIENT;
            }

            # We did insert the recipient, but it's not there;
            # this can happen only if another client deleted it
            # so we have a race condition. We'll let the retries
            # deal with it.
            croak "Unable to index message: $global_id " .
                  "due to missing recipient $recipient";
        }
    }
}

# Just remove the message from the DBI index
sub _forget {
    my ($self, $msg, $idx) = @_;
    $self->_trash_message($msg);
}

sub reindex {
    my($self, %opts) = @_;
    $self->SUPER::reindex(%opts);
    $self->purge_msg_data() if $opts{'purge'};
}

sub status {
    my $self = shift;

    # this collects the status info for the file system based quarantine
    # storage
    my $status_href = $self->SUPER::status();
    
    my $quarantine_type = 'pmdb';
    # this message count is an approximation; it should be good enough
    # for the purpose, though
    $self->_db_connect unless $self->{pmxdir_dbh};
    my $dbh = $self->{pmxdir_dbh};
    my ($indexed_message_count) = $self->{pmxdir_dbh}->selectrow_array(<<"EOT");
select cast (reltuples as integer) from pg_class where relname = 'message'
EOT
    $status_href->{"$quarantine_type#num_msg"} = $indexed_message_count;

    # if we can access the db admin utility we'll assume we're running
    # on the DBMS server and try to add the db disk utilization info to
    # the status; we'll silence and ignore any errors since the admin
    # utility exists only on the db server
    eval {
        my $saved_handler = $SIG{__WARN__};
        $SIG{__WARN__} = sub {};
        my ($disk_util) = grep /disk/, `pmx-database status`;
        $SIG{__WARN__} = $saved_handler;
        if ($disk_util) {
            chomp $disk_util;
            my ($discarded, $kbytes) = split /:\s+/, $disk_util;
            $status_href->{"$quarantine_type#disk_used"} = $kbytes;
        }
    };

    return $status_href;
}

sub remote_message {
    my $self = shift;
    return map scalar($self->message($_)), @_ if wantarray;

    die "Can't return multiple messages in scalar context" if @_ > 1;
    my $id = shift;
    my $filename = join("/", $self->{pmxdir_dir}, "cur", $self->_id_fname($id));

    require PureMessage::MessageStore::pmdb::Message;
    my $m = PureMessage::MessageStore::pmdb::Message::_new($self, $id,
						           $filename,
						           $self->{pmxdir_writable},
						          );
    return $m if $m;
    return undef;
}

sub dbi_prepare {
    my $self = shift;

    $self->_db_connect unless $self->{pmxdir_dbh};

    my $sth = eval{ $self->{pmxdir_dbh}->prepare(@_) };
    if ($@) {
	croak $@;
    }
    return $sth;
}

sub dbi_handle {
    my $self = shift;

    $self->_db_connect unless $self->{pmxdir_dbh};

    $self->{pmxdir_dbh};
}

sub remote_message_count {
    my $self = shift;
    my $folder = shift; # XXX folder is ignored in the global context
    $folder = "cur" unless defined $folder;

    $self->_db_connect unless $self->{pmxdir_dbh};

    my $dbh = $self->{pmxdir_dbh};
    my $msg_cnt = $self->{pmxdir_sth}{message_count} ||= 
	$dbh->prepare("SELECT COUNT(*) FROM message");
    $msg_cnt->execute;
    my($count) = $msg_cnt->fetchrow_array();
    return $count;
}

sub empty_trash {
    my $self = shift;
    return unless grep $_ eq "trash", $self->folders;
    $self->message_scan(sub {
	my $m = shift;
	_unlink_message_file($self, $m->filename);
	$self->_trash_message($m);
    }, "trash");
}

sub _unlink_message_file {
    my $self = shift;
    require ActiveState::DiskUsage;
    for my $f (@_) {
	my $du = ActiveState::DiskUsage::du($f);
	unlink($f) || die "Can't unlink '$f': $!";
	$self->{pmxdir_disk_used_delta} -= $du;
	$self->{pmxdir_num_msg_delta}--;
    }
}

# This should only ever be called from _forget() which is only called from
# meta_index(). The message object here is the fake one created when 
# reading remove lines from the index_log. Returns the count of deleted
# messages (can be 1 or 0);
sub _trash_message {
    my($self, $msg) = @_;

    $self->_db_connect unless $self->{pmxdir_dbh};

    my $dbh = $self->{pmxdir_dbh};
    my $gid = $msg->minfo_value("Global-ID");
    my $msgs_deleted = 0;

    # message deletion will cascade to dependent data
    my $del_message = $self->{pmxdir_sth}{delete_message} ||=
	$dbh->prepare("DELETE FROM message WHERE m_global_id = ?");
	    
    if ($gid) {
        eval {
            $del_message->execute($gid);
            $msgs_deleted = $del_message->rows();

	    # deletions are still on one-per-transaction basis, but they
	    # can be interspersed with inserions therefore with committing
	    # the deletion we commit the insertions too, so we need to empty
	    # the queue of the uncommitted but successfully indexed messages.
            # (which will result in committing the index operations as well)
            $self->_flush_indexed();

            if ($msgs_deleted) {
                $self->log("DEBUG", "Message $gid removed");
            }
            else {
                # We should handle this better by ensuring that we dont try to
                # delete messages that we know dont exist in DB.
                $self->log("DEBUG",
                "Message $gid could not be found in DB");
            }
        };
    }
    if ($@) {
	$dbh->rollback;
	die $@;
    }
    return $msgs_deleted;
}

# removes all messages and all dependent data from the DBMS-based quarantine;
# may take a long time and be very destructive
sub purge_msg_data {
    my $self = shift;
    $self->_db_connect() unless $self->{pmxdir_dbh};
    my $dbh = $self->{pmxdir_dbh};
    eval { 
        my $stmt = $dbh->prepare("delete from message
                                  where message.location_name = ?");
    
        $stmt->execute($self->{pmxdir_location});
        $dbh->commit();
    };
    if ($@) {
	$dbh->rollback;
	$self->log("WARNING", "Error while removing message data: $@");
    }
}

sub message_set {
    my $self = shift;
    require PureMessage::MessageStore::pmdb::MessageSet;

    $self->_db_connect unless $self->{pmxdir_dbh};


    my $rv = eval {
	PureMessage::MessageStore::pmdb::MessageSet->new(
	    verbose => $self->{pmxdir_verbose},
	    @_,
	    message_store => $self);
    };
    croak $@ if $@;
    return $rv;				   
}

sub _clean_mail_addresses {
    my $v = shift;
    require Mail::Address;
    local $Mail::Address::VERBOSE = 0; # silence "Unmatched () ... " warnings
    my @addr = grep length, map lc($_->address), Mail::Address->parse($v);
    return @addr if wantarray;
    return $addr[0];
}

sub _clean_detail {
    my $addr = shift;
    $addr =~ s/\+[^@]*//;
    return $addr;
}

sub message {
    my $self = shift;
    return map scalar($self->message($_)), @_ if wantarray;

    $self->_db_connect unless $self->{pmxdir_dbh};

    my $id = shift;
    unless ($self->id_is_local($id)) {

	# construct the message object from the db, but first
	# check if we have the right id and get one if we don't
        # (if what we have is the old style global id we need to get
        # the current one)
        if ($id =~ /^gid_/) {
            $id = $self->global_from_gid($id);
        }
	require PureMessage::MessageStore::pmdb::Message;
	return PureMessage::MessageStore::pmdb::Message->new($self, $id);
    }

    # or use the Dir::Message object for a local message
    return $self->SUPER::message($id);
}

# Constructs a reduced message object from the data supplied as arguments;
# useful for digesting purposes; returns undef on failure.
sub message_lite {
    my $self = shift;
    my $msg_data = shift;
    require PureMessage::MessageStore::pmdb::Message;
    return PureMessage::MessageStore::pmdb::Message->new_lite($msg_data);
}

# Returns the value the global keys can be sorted on. We want the order of
# quarantining, so we use the surrogate key, which will provide sufficiently
# accurate order.
sub sort_key {
    my $self = shift;
    my $id = shift;
    $self->_db_connect unless $self->{pmxdir_dbh};
    my $dbconn_handle = $self->{pmxdir_dbh};
    my $numeric;
    eval {
        my $slct_handle = $dbconn_handle->prepare("select id from message
                                                  where m_global_id = ?");
        $slct_handle->execute($id);
        $slct_handle->bind_columns(\$numeric);
        $slct_handle->fetch();
    };
    if ($@) {
        $dbconn_handle->rollback();
        $self->log("ERROR",
                   "Error while retrieving sort key for message $id: $@");
    }
    return $numeric;
}

# retrieves (and returns) the current message global id based on the
# old style global id ('gid_...'). Lookup is performed in the migrated 4.x data.
sub global_from_gid {
    my $self = shift;
    my $gid = shift;
    my $global_id;
    $self->_db_connect unless $self->{pmxdir_dbh};
    my $dbconn_handle = $self->{pmxdir_dbh};
    eval {
        my $slct_handle = $dbconn_handle->prepare(<<"EOT");
select message.m_global_id
from message, pmx4_location
where message.m_global_id like (pmx4_location.local_id || '%')
and message.location_name = pmx4_location.host
and pmx4_location.m_global_id = ?
EOT
        $slct_handle->execute($gid);
        $slct_handle->bind_columns(\$global_id);
        $slct_handle->fetch();
    };
    if ($@) {
        $dbconn_handle->rollback();
        $self->log("ERROR",
                   "Error while retrieving global id based on old style gid for message $gid: $@");
    }
    return $global_id;
}

# Does lookup in the migrated pmx4.x data and maps the specified surrogate key
# into the old-style global id. This is necessary only for approvals
# through 4.x-generated digests in 5.x environment.
sub gid_from_surrogate_key {
    my $self = shift;
    my $id = shift;
    my $gid;
    $self->_db_connect unless $self->{pmxdir_dbh};
    my $dbconn_handle = $self->{pmxdir_dbh};
    eval {
        my $slct_handle = $dbconn_handle->prepare(
            "select m_global_id from pmx4_location where message_id = ?"
        );
        $slct_handle->execute($id);
        $slct_handle->bind_columns(\$gid);
        $slct_handle->fetch();
    };
    if ($@) {
        $dbconn_handle->rollback();
        $self->log("ERROR",
                   "Error while retrieving old style global id for message $id: $@");
    }
    return $gid;
}

sub get_release_queue {
    my $self = shift;
    my $verbose = shift || 0;
    unless ($self->{release_queue}) {
        require PureMessage::MessageStore::pmdb::ReleaseQueue;
        $self->{release_queue} =
            PureMessage::MessageStore::pmdb::ReleaseQueue->new(
                store => $self, verbose => $verbose
            );
    }
    return $self->{release_queue};
}

sub active_edge_servers {
    my $self = shift;
    $self->_db_connect unless $self->{pmxdir_dbh};
    my $dbh = $self->{pmxdir_dbh};

    my $sql = <<EOT;
select hostname from cfg_location 
  where location_name in 
    (select distinct location_name from message)
EOT
    # this will return an arrayref of arrayrefs
    my $data = $dbh->selectall_arrayref($sql);
    my @edges = map { @$_ } @$data;
    return \@edges;
}

1;

__END__

=head1 NAME

PureMessage::MessageStore::pmdb - File system storage with DBI indices

=head1 DESCRIPTION

The directory format is inspired by the I<maildir> format
designed for I<qmail>; <http://cr.yp.to/proto/maildir.html>.  There is
a single file per message, and each of these files live in
sub-directories of the quarantine directory.

A quarantine directory will contain at least the following
sub-directories.  There might be more:

=over

=item F<tmp>

This directory is used for temporary files.  Messages that might
potentially be stored are collected here first.  If the message
is to be stored, then it renamed so that it ends up in the F<new>
directory.  If not it is unlinked.  Old files left in the F<tmp>
directory can be safely deleted.

=item F<new>

Messages that are to be stored by PureMessage end up here.
The file name is ugly and long to make sure different milter
processes generate unique names.

=item F<cur>

When C<pmx-qman> starts up it moves new messages to this folder.  At
the same time it renames the messages so that the file names are plain
sequential integer numbers (like mh uses).

=back

=head2 Message file format

There is a single file per stored message.  The file contains the raw
e-mail message containing headers and a body, followed by the meta
information.

The meta information is a plain text block appended to the
message.  The size of this block is determined by reading the last 5
bytes of the file.  The last of these 5 bytes will be "\n".  It can be
ignored.  The 4 before should be interpreted as an ASCII hex number.

The internal format of the meta information block is
similar to that of email headers.  The first line contains the string
"-- \n" and should be ignored.

=head1 COPYRIGHT

Copyright (c) 2002 ActiveState Corp.  All rights reserved.

=head1 SEE ALSO

L<PureMessage::MessageStore>

=cut

