package PureMessage::MessageStore::pmdb::Message;

# Encapsulates operations on message  objects put together from the data
# read from the database. These message object are typically incomplete;
# they lack either a part or the entire message body, and are typically
# useful only for digesting or header previewing.

# XXX The big problem with this module is its circular dependency on
# the pmdb.pm module. This has to be rectified.

use strict;
use Carp qw(croak);

my $msg_sql = <<EOT;
select
    message.id as m_db_id,
    message.m_global_id as id,
    m_global_id,
    m_queue_id,
    h_message_id,
    m_date as c_date_iso,
    m_date,
    c_subject_utf8,
    c_subject_body_utf8,
    message.reason_id as reason_id,
    m_reason.name as m_reason,
    m_spam_probability,
    h_to,
    m_message_size,
    m_header_size,
    date_part('epoch', message.m_date) as c_qdate_time,
    h_from_local || '\@' || h_from_domain as h_from,
    env_from_local || '\@' || env_from_domain as env_from,
    env_from_local || '\@' || env_from_domain as m_envelope_from,
    location_name as location,
    m_milter_host,
    m_relay
from
    message
    join m_reason on message.reason_id = m_reason.reason_id
EOT

sub get_msg_sql {
    return $msg_sql;
}

# Constructor that builds a message read from the database. "Limited
# retrieval" mode populates only core part of the message structure,
# thus minimizing the DBMS access. If limited retrieval is OK, caller
# needs to indicate that explicitely. Returns the reference to the
# message structure, or undef if the message can not be retrieved.
sub new {
    my $class = shift;
    my ($store, $id, $limited_retrieval) = @_;

    my $sth = $store->{pmxdir_sth}{select_message} ||=
	_mk_query($store);
    $sth->execute($id);
    my %row;
    $sth->bind_columns( \( @row{ @{$sth->{NAME_lc} } } ));

    my %msg;
    my @c_from;
    unless ($sth->fetch()) {
        return undef;
    }

    push(@c_from, $row{env_from});
    push(@c_from, $row{h_from});
    %msg = %row;
    $msg{c_from} = \@c_from;
    
    # initialize storage for multiple-valued attributes
    my @m_envelope_to;
    $msg{m_envelope_to} = \@m_envelope_to;
    my @c_to;
    $msg{c_to} = \@c_to;
    my @m_mapped_recipients;
    $msg{m_mapped_recipients} = \@m_mapped_recipients;
    my @m_spam_hit;
    $msg{m_spam_hit} = \@m_spam_hit;
    my @m_virus_id;
    $msg{m_virus_id} = \@m_virus_id;

    my $location = $msg{location};

    # XXX some of the SQL statements that follow are not cached on the
    # client side; perhaps they should be

    # retrieve multiple-valued attributes in a new statement handle
    # XXX note the PGSQL-specific syntax in casting the parameter
    my $multi_val_sth;

    $multi_val_sth = $store->dbi_handle()->prepare(
        "select recipient.local || '\@' || recipient.domain,
                message_recipient.is_mapped,
                message_recipient.is_calculated,
                message_recipient.is_envelope
         from (recipient join message_recipient
               on recipient.id = message_recipient.recipient_id)
         where message_recipient.message_id = ?::bigint");
    $multi_val_sth->execute($msg{m_db_id});
    while (my ($recipient, $is_mapped, $is_calculated, $is_envelope) =
           $multi_val_sth->fetchrow_array()) {
        push(@m_mapped_recipients, $recipient) if $is_mapped;
        push(@c_to, $recipient) if $is_calculated;
        push(@m_envelope_to, $recipient) if $is_envelope;
    }
    $multi_val_sth->finish();

    # retrieve the non-core data, if needed
    if (not $limited_retrieval) {
        $multi_val_sth = $store->dbi_handle()->prepare(
            "select quarantine_rule.rule_text, m_reason.name
             from ((quarantine_rule_hit join quarantine_rule
                  on quarantine_rule_hit.quarantine_rule_id = quarantine_rule.id)
                  join m_reason on quarantine_rule.reason_id = m_reason.reason_id)
             where message_id = ?::bigint");
        $multi_val_sth->execute($msg{m_db_id});
        while (my ($this_hit, $this_reason) = $multi_val_sth->fetchrow_array()) {
            push(@m_spam_hit, $this_hit) if $this_reason eq "spam";
            push(@m_virus_id, $this_hit) if $this_reason eq "virus";
        }
        $multi_val_sth->finish();
    }
    
    my $self = bless {
	pmxdir_store => $store,
	pmxdir_id    => $id,
	pmxdir_sth   => $sth,
	pmxdir_msg   => \%msg,
	pmxdir_location => $location,
    }, $class;
    return $self;
}

# This is an alternative constructor which accepts message attributes
# as parameters (as opposed as reading them from the db). Returns undef
# if any of the required parameters is missing.
sub new_lite {
    my $class = shift;
    my $msg_data = shift;

    # non-mandatory attributes
    for my $key (qw(subject header_from envelope_from subject_body)) {
        unless ($msg_data->{$key}) {
            $msg_data->{$key} = '';
        }
    }

    unless ($msg_data->{spam_probability}) {
        $msg_data->{spam_probability} = 0;
    }

    # mandatory attributes
    for my $key (qw(id m_date reason recipient size)) {
        unless (defined ($msg_data->{$key})) {
            return undef;
        }
    }

    my $msg = {};
    $msg->{pmxdir_id} = $msg_data->{id};
    $msg->{pmxdir_msg} = { m_global_id => $msg_data->{id},
                           env_from => $msg_data->{envelope_from},
                           c_subject_utf8 => $msg_data->{subject},
                           c_subject_body_utf8 => $msg_data->{subject_body},
                           c_date_iso => $msg_data->{m_date},
                           h_from => $msg_data->{header_from},
                           m_envelope_from => $msg_data->{envelope_from},
                           m_spam_probability => $msg_data->{spam_probability},
                           m_date => $msg_data->{m_date},
                           m_reason => $msg_data->{reason},
                           m_message_size => $msg_data->{size},
                           m_envelope_to => [$msg_data->{recipient}],
                         };
    bless $msg, $class;
    return $msg;
}

sub append_envelope_recipient { 
    my $self = shift;
    my $recipient = shift;
    push @{$self->{pmxdir_msg}->{m_envelope_to}}, $recipient;
}

sub _mk_query { 
    my $store = shift;
    my $db = $store->dbi_handle;

    #XXX remove the unnecessary env-from pseudo col, but which one that is?
    my $sql = "$msg_sql where message.m_global_id = ?";
    print "$sql\n" if $ENV{PMX_pmdb_SQL_DEBUG};
    return $db->prepare($sql);
}

sub id {
    shift->{pmxdir_id};
}

sub message_store {
    shift->{pmxdir_store};
}

sub folder { "cur" }

sub trash {
    my $self = shift;
    $self->get_local_message->trash;
}

sub message_size {
    shift->{pmxdir_msg}{m_message_size};
}

sub header_size {
    shift->{pmxdir_msg}{m_header_size};
}

sub c_from { 
    my $self = shift;
    return @{$self->{pmxdir_msg}{c_from}} if wantarray;
    return $self->{pmxdir_msg}{c_from}[0];
}

sub c_to {
    my $self = shift;
    return @{$self->{pmxdir_msg}{c_to}} if wantarray;
    return $self->{pmxdir_msg}{c_to}[0];
}

sub c_subject_utf8 {
    my $self = shift;
    return $self->{pmxdir_msg}{c_subject_utf8};
}

sub c_reason_csv {
    my $self = shift;
    return $self->{pmxdir_msg}{m_reason};
}

# Returns a concatenation of subject and first few chars of the message
# body; if not part of the object already, it needs to be read from the db.
# This function is used for presentation purposes only so HTML formatting
# will be removed.
sub c_subject_body_utf8 {
    my $self = shift;
    my $subject_body = $self->{pmxdir_msg}{c_subject_body_utf8};

    unless ($subject_body) {

        # We are probably dealing with message that was indexed before the
        # subject_body column was added => fall back to subject only + separator
        $subject_body = $self->c_subject_utf8() || '';
        $subject_body .= ' << ';
    }
    return $subject_body;
}

sub c_qdate_time {
    my $self = shift;
    return $self->{pmxdir_msg}{c_qdate_time};
}

sub c_date_iso {
    my $self = shift;
    return $self->{pmxdir_msg}{c_date_iso};
}

sub minfo_value {
    my ($self, $key, $val) = @_;

    warn "Updating minfo not yet supported." if $val;
    $key =~ s/-/_/g;
    $key = lc("m_$key") unless $key =~ /^m_/;
    my $rv = $self->{pmxdir_msg}{$key};
    
    return undef unless defined $rv;
    $rv = [$rv] unless ref($rv) eq "ARRAY";
    return @{$rv} if wantarray;
    return $rv->[0];
}

sub minfo_items {
    my $self = shift;
    my @items;
    foreach my $item ("Queue-Id", "Date", "Spam-Probability", "Envelope-To",
	"Mapped-Recipients", "Spam-Hit", "Global-Id", "Virus-Id") {

	push(@items, map { $item => $_ } $self->minfo_value($item))
            if $self->minfo_value($item);
    }
    push(@items, "Envelope-From", $self->{pmxdir_msg}{env_from})
	if $self->{pmxdir_msg}{env_from};
    push(@items, "Relay", $self->{pmxdir_msg}{m_relay})
        if $self->{pmxdir_msg}{m_relay};
    push(@items, "Milter-Host", $self->{pmxdir_msg}{m_milter_host})
        if $self->{pmxdir_msg}{m_milter_host};
    push(@items, "Reason", $self->{pmxdir_msg}{m_reason})
        if $self->{pmxdir_msg}{m_reason};
    
    return @items;
}

sub header {
    my ($self, $header) = @_;
    $header =~ s/-/_/g;
    $header = lc("h_$header") unless $header =~ /^h_/;
    my $rv = $self->{pmxdir_msg}{$header};

    return undef unless defined $rv;
    $rv = [$rv] unless ref($rv) eq "ARRAY";
    return @{$rv} if wantarray;
    return $rv->[0];
}

# XXX Should depricate the $negate flag as it shouldn't be
# needed anymore (just use "new" for the action).
sub user_actions {
    my ($self, $recip, $action, $negate, $refresh_aliases) = @_;
    
    my $to = ref($recip) ? $recip : [$recip];
    $action = "new" if $negate;

    my %user_actions;
    my $store = $self->{pmxdir_store};
    my $id = $self->{pmxdir_msg}{m_db_id};
    my $dbh = $store->dbi_handle;
    require PureMessage::MessageStore::pmdb;

    #XXX The PGSQL-style explicit casting into bigint should go away
    # with eventual upgrade to PGSQL 7.5 where this issue is fixed
    my $get_action_sth = $dbh->prepare(
	'select action from message_recipient
	 where message_id = ?::bigint
         and recipient_id in (
             select id from recipient where local = ? and domain = ?
                                                      and is_mapped = 1
         )');
    my $set_action_sth = $store->dbi_handle->prepare(
	'update message_recipient set action = ?
	 where message_id = ?::bigint
         and recipient_id in (
             select id from recipient where local = ? and domain = ?
                                                      and is_mapped = 1
         )');

    # We'll look first for an existing mapping and only if we can't
    # find one we'll rebuild the alias map (which is very time consuming)
    foreach my $addr (@$to) {
	my ($local, $domain) =
            PureMessage::MessageStore::pmdb::split_address($addr);
	if ($action) {

	    # By default the action is new(undef).
	    my $action_id;
	    if ($action eq "approve") {
		$action_id = 2;
	    }
	    elsif ($action eq "delete") {
		$action_id = 3;
	    }
	    eval {
		$set_action_sth->execute($action_id, $id, $local, $domain);
                unless ($set_action_sth->rows()) {
                    my $aliases = $store->get_alias_map($refresh_aliases);
                    my $tx_addr = $aliases->transform($addr);
                    ($local, $domain) =
                        PureMessage::MessageStore::pmdb::split_address($tx_addr);
                    $set_action_sth->execute($action_id, $id, $local, $domain);
                }
	    };
	    if ($@) {
		my $err = $@;
		$dbh->rollback;
		croak $err;
	    }
	    $dbh->commit;
	    $user_actions{$addr} = $action;
	}
	else {
	    $get_action_sth->execute($id, $local, $domain); 
	    my @result = $get_action_sth->fetchrow_array;
            unless (scalar @result) {
                my $aliases = $store->get_alias_map($refresh_aliases);
                my $tx_addr = $aliases->transform($addr);
                ($local, $domain) =
                    PureMessage::MessageStore::pmdb::split_address($tx_addr);
                $get_action_sth->execute($id, $local, $domain);
	        @result = $get_action_sth->fetchrow_array;
            }
            
            my $action;
	    my $action_id = $result[0];
	    if ($action_id) {
		if ($action_id == 2) {
		    $action = "approve";
		}
		elsif ($action_id == 3) {
		    $action = "delete";
		}
	    }
	    $user_actions{$addr} = $action;
        }
    }
    return %user_actions if wantarray;
    return $user_actions{$to->[0]};
}

sub is_local {
    my $self = shift;
    $self->_debug("location: $self->{pmxdir_location}");
    return ($self->{pmxdir_location} eq $self->{pmxdir_store}->location);
}


sub locations { 
    my $self = shift;
    return [ $self->{pmxdir_location} ];
}

sub local_id {
    my $self = shift;
    my $store = $self->{pmxdir_store};
    my $global_id = $self->{pmxdir_msg}->{id};
    return $store->local_from_global($global_id);
}

# Will use the local id to load the Dir message object for the message
# in this location (if necessary).
sub get_local_message {
    my $self = shift;
    my $local_id = $self->local_id();
    return $self->{pmxdir_store}->message($local_id);
}

sub _debug {
    my $self = shift;
    my $log = $self->{pmxdir_store}->{pmxdir_logger};
    $log->log("DEBUG", @_);
}

sub cleaned_body {
    my $self = shift;
    my $db = $self->{pmxdir_store}->dbi_handle;
    my $id = $self->{pmxdir_msg}{id};
    my $sql = 'select message_body.c_body
               from message_body join message 
               on message_body.message_id = message.id
               where message.m_global_id = ?';
    my $sth = $db->prepare($sql);
    $sth->execute($id);
    my ($body) = $sth->fetchrow_array;
    return $body;
}

sub raw_body {
    my $self = shift;
    my $db = $self->{pmxdir_store}->dbi_handle;
    my $id = $self->{pmxdir_msg}{id};
    my $sql = 'select message_body.raw_msg
               from message_body join message
               on message_body.message_id = message.id
               where message.m_global_id = ?';
    my $sth = $db->prepare($sql);
    $sth->execute($id);
    my ($body) = $sth->fetchrow_array;
    return $body;
}

# Mimics the "file read" interface for messages stored in DBMS
# by retrieving raw message data in chunks. Accepts buffer reference
# and max chunk length as arguments.
sub read {
    croak("Wrong number of arguments") if @_ != 3;
    my $self = shift;
    my $max_chunk_len = $_[1];
    my $delivered = $self->{delivered} ? $self->{delivered} : 0;
    $self->{raw_body} = $self->raw_body() unless $self->{raw_body};
    my $total = length($self->{raw_body});
    my $chunk_len = $delivered + $max_chunk_len > $total ?
                    $total - $delivered : $max_chunk_len;
    $_[0] = substr($self->{raw_body}, $delivered, $chunk_len);
    $self->{delivered} += $chunk_len;
    if (0 == $chunk_len) {
        delete $self->{delivered};
        delete $self->{raw_body};
    }
    return $chunk_len;
}

sub message_groups {
    my $self = shift;
    my $db = $self->{pmxdir_store}->dbi_handle;
    my $id = $self->{pmxdir_msg}{id};
    my $sql = 'select group_name from message_group, message 
                 WHERE message_group.message_id = message.id 
                   AND message.m_global_id = ?';
    my $sth = $db->prepare($sql);
    $sth->execute($id);
    my @groups;
    while(my ($foo) = $sth->fetchrow_array()){
        push @groups, $foo;
    }
    return \@groups;
}

sub is_outbound {
    my $self = shift;
    my $db = $self->{pmxdir_store}->dbi_handle;
    my $id = $self->{pmxdir_msg}{id};
    my $sql = 'select outbound from message 
                 WHERE message.m_global_id = ?';
    my $sth = $db->prepare($sql);
    $sth->execute($id);
    return $sth->fetchrow_array;
}

1;
