package PureMessage::MessageStore::pmdb::ReleaseQueue;

use strict;

# This module is comprised of routines for manipulating the DBMS
# implementation of enqueued message release requests.
#
# Error handling: Routines in this module log errors and return
# statuses. We're trying to not croak() or die() here...
#
# Transaction processing: This module uses the same db handle as the
# one used for message store (i.e. quarantine index) operations.

my $capability = "Centralized Queue";

sub new {
    my $class = shift;
    my %opts = @_;
    my $msg_store = $opts{store};
    my $self = bless {
        store     => $msg_store,
        db_handle => $msg_store->{pmxdir_dbh},
        logger    => $msg_store->{pmxdir_logger},
        verbose   => $opts{verbose},
    }, $class;

    return $self;
}

# enqueues new release request; returns true if the request was
# successfully enqueued; otherwise false
sub submit_request {
    my $self = shift;
    my %opts = @_;
    my @param_error;
    my ($global_id, $local, $domain, $request_type);
    my $count_inserted = 0;
    my $action_numeric;

    # check parameters
    ($global_id = $opts{global_id})
        or push @param_error, 'missing parameter "global_id"';
    ($request_type = $opts{request_type})
        or push @param_error, 'missing parameter "request_type"';
    ($local = $opts{local} or $request_type eq 'delete' or
     $request_type eq 'approve')
        or push @param_error, 'missing parameter "local"';
    if ($request_type) {
        ($action_numeric = $self->{store}->action_numeric($request_type))
            or push @param_error, 'invalid parameter "request_type"';
    }
    $domain = $opts{domain};

    # ensure that into null columns we insert nulls and not zero-length strings
    $local = undef unless $local;
    $domain = undef unless $domain;

    if (scalar @param_error) {
        my $errors = join '; ', @param_error;
        $self->_log("ERROR",
                    "Invalid parameters for submit_request(): $errors");
        return 0;
    }

    $self->{ins_rqst_sql} ||= <<"EOT";
insert into qrntn_action_request
    (location_name, m_global_id, action, local, domain)
select message.location_name, ?, ?, ?, ? from message where m_global_id = ?
EOT
    eval {
        my $msg = "submitting request: $global_id, $request_type, $global_id";
        $msg .= ", $local" if $local;
        $msg .= ", $domain" if $domain;
        $self->_vprint($msg);
        my $prep_stmt = $self->{db_handle}->prepare($self->{ins_rqst_sql});
        $prep_stmt->execute($global_id, $action_numeric, $local, $domain, $global_id);
        $count_inserted = $prep_stmt->rows();
        $self->{db_handle}->commit();
    };
    if ($@) {
        # if we already have such a request it's not an error
        $self->_log("ERROR", "Error while inserting release request: $@")
            unless ($@ =~ /violates unique constraint/);
        $self->{db_handle}->rollback();
    }
    return $count_inserted;
}

# Retrieves all pending requests for the specified mail processing
# host; if request one or more types are specified as a parameter then
# only the requests of these types will be retrieved. Returns reference
# to an array of hash references. Hashes are keyed by:
# host, global_id, request_type, local, domain, attempt_count.
sub get_requests {
    my $self = shift;
    my $host = shift;
    my $request_type = shift;
    my $result_set_ref = [];
    my $sql = <<"EOT";
select location_name as host, m_global_id as global_id,
       action as request_type_numeric, local, domain, attempt_count
from qrntn_action_request
where location_name = ?
EOT

    # bild action filter, if needed
    if ($request_type) {
        $request_type = [$request_type] unless ref $request_type;
        my @action = map {$self->{store}->action_numeric($_)} @$request_type;
        my $action_clause = "and action in (" .  join(', ', @action) . ")\n";
        $sql .= $action_clause;
    }

    $sql .= "order by submission_time";

    eval {
        my $stmt = $self->{db_handle}->prepare($sql);
        $stmt->execute($host);
        while (my $hash_ref = $stmt->fetchrow_hashref()) {

            # outside of this module we're working with decoded actions
            $hash_ref->{request_type} =
                $self->{store}->action_string($hash_ref->{request_type_numeric});
            push @$result_set_ref, $hash_ref;
        }
    };
    if ($@) {
        $self->{logger}->log("ERROR",
                             "Error while retrieving release request: $@");
        $self->{db_handle}->rollback();
    }
    return $result_set_ref;
}

# Deletes requests specified as a reference to an array of hash
# references; hashes are keyed by:
# host, global_id, request_type_numeric, local, domain.
# Returns the count of deleted requests. All requests get deleted in
# a single transaction.
sub delete_requests {
    my $self = shift;
    my $request_arr_ref = shift;
    my $deleted_count = 0;
    my ($local, $domain);
    eval {

        # the actual SQL for deletion can take one of three forms;
        # prepare them ouside of the loop
        my $stmt_full = $self->{db_handle}->prepare(<<"EOT");
delete from qrntn_action_request
where m_global_id = ? and action = ? and local = ? and domain = ?
EOT
        my $stmt_null_domain = $self->{db_handle}->prepare(<<"EOT");
delete from qrntn_action_request
where m_global_id = ? and action = ? and local = ? and domain is null
EOT
        my $stmt_null_local_domain = $self->{db_handle}->prepare(<<"EOT");
delete from qrntn_action_request
where m_global_id = ? and action = ? and local is null and domain is null
EOT
        foreach my $hash_ref (@$request_arr_ref) {
            my @sql_arg = ($hash_ref->{global_id},
                           $hash_ref->{request_type_numeric});
            my $stmt;
            if ($hash_ref->{local} and $hash_ref->{domain}) {
                push @sql_arg, $hash_ref->{local};
                push @sql_arg, $hash_ref->{domain};
                $stmt = $stmt_full;
            }
            elsif ($hash_ref->{local}) {
                push @sql_arg, $hash_ref->{local};
                $stmt = $stmt_null_domain;
            }
            else {
                $stmt = $stmt_null_local_domain;
            }

            $stmt->execute(@sql_arg);
            $deleted_count += $stmt->rows();
        }
        $self->{db_handle}->commit();
    };
    if ($@) {
        $self->{logger}->log("WARNING",
                             "Error while deleting release request: $@");
        $self->{db_handle}->rollback();
        $deleted_count = 0;
    }
    return $deleted_count;
}

# Increments attempt counters for the specified requests (see the
# delete_requests() function for the input format). All counts incremented
# in a single transaction. Returns the count of incremented counters.
sub inc_attempt_counts {
    my $self = shift;
    my $request_arr_ref = shift;
    my $updated_count = 0;
    my ($local, $domain);
    eval {

        # the actual SQL for incrementing can take one of three forms;
        # prepare them ouside of the loop
        my $stmt_full = $self->{db_handle}->prepare(<<"EOT");
update qrntn_action_request set attempt_count = attempt_count + 1
where m_global_id = ? and action = ? and local = ? and domain = ?
EOT
        my $stmt_null_domain = $self->{db_handle}->prepare(<<"EOT");
update qrntn_action_request set attempt_count = attempt_count + 1
where m_global_id = ? and action = ? and local = ? and domain is null
EOT
        my $stmt_null_local_domain = $self->{db_handle}->prepare(<<"EOT");
update qrntn_action_request set attempt_count = attempt_count + 1
where m_global_id = ? and action = ? and local is null and domain is null
EOT
        foreach my $hash_ref (@$request_arr_ref) {
            my @sql_arg = ($hash_ref->{global_id},
                           $hash_ref->{request_type_numeric});
            my $stmt;
            if ($hash_ref->{local} and $hash_ref->{domain}) {
                push @sql_arg, $hash_ref->{local};
                push @sql_arg, $hash_ref->{domain};
                $stmt = $stmt_full;
            }
            elsif ($hash_ref->{local}) {
                push @sql_arg, $hash_ref->{local};
                $stmt = $stmt_null_domain;
            }
            else {
                $stmt = $stmt_null_local_domain;
            }

            $stmt->execute(@sql_arg);
            $updated_count += $stmt->rows();
        }
        $self->{db_handle}->commit();
    };
    if ($@) {
        $self->{logger}->log("WARNING",
                             "Error while incrementing attempt counts: $@");
        $self->{db_handle}->rollback();
        $updated_count = 0;
    }
    return $updated_count;
}

# Does table/storage structure maintenance.
sub maintain_queue {
    my $self = shift;
    $self->{db_handle}->{AutoCommit} = 1;
    $self->{db_handle}->do("vacuum qrntn_action_request");
    $self->{db_handle}->{AutoCommit} = 0;
}

sub _vprint {
    my $self = shift;
    return unless $self->{verbose};
    my $message = shift;
    print "$capability: $message\n";
}

sub _log {
    my $self = shift;
    $self->{store}->log(@_);
}

1;
