package PureMessage::MessageStore::pmdb::UserView;
use base 'Exporter';
use strict;
use warnings;

# This package encapsulates a couple of routines used in the optimized
# EUWI processing. The routines are invoked using the non-OO style of
# interface. These routines perform DB (centralized index) access, and
# may die, therefore they should be evaled by the caller.
# This package uses the caller's logger object (if specified) to log
# errors.

our @EXPORT_OK = qw(get_view_sizes get_msg get_user_msgs);
my @COLS = qw(qid m_enduser c_reason_csv h_from h_to m_date:tzt
              c_subject_utf8 m_spam_probability);

# returns the count of quarantined messages for each of the per-user
# message states for the specified user
sub get_view_sizes {
    my %opts = @_;
    my $address = $opts{address};
    my $logger = $opts{logger};
    my $index = $opts{store};
    my $reasons = $opts{reasons};
    my $this_sub = "get_view_counts()";
    unless ($address and $index) {
        my $error_msg = "$this_sub: invalid parameters; unable to proceed";
        _log($logger, "ERROR", $error_msg);
        die($error_msg);
    }
    my $dbh = $index->dbi_handle();
    my $approved_numeric = $index->action_numeric('approve');
    my $deleted_numeric = $index->action_numeric('delete');
    my $view_size;
    my $reasons_predicate = _get_reasons_predicate($reasons);
    my $sql = <<"__SQL__";
select  coalesce(sum(case when action is null then 1 end),0) as visible,
        coalesce(sum(case when action = $approved_numeric then 1 end),0) as approved,
        coalesce(sum(case when action = $deleted_numeric then 1 end),0) as deleted
from    message_recipient
join    recipient on recipient.id = recipient_id
where   is_mapped = 1
and     ((local = ?::text or local ~ ?::text) and domain = ?::text)
and     $reasons_predicate
__SQL__

    my ($local, $domain) = @$address;
    eval {
        $view_size = $dbh->selectall_arrayref($sql, undef, $local, "^$local\\+",
                                              $domain);
    };
    if ($@) {
        my $error_msg = "$this_sub: unable to retrieve view counts for " .
                        ${$address}[0] . "\@" . ${$address}[1] . ": $@"; 
        _log($logger, "ERROR", $error_msg);
        die($error_msg);
    }
    my %view_size;
    $view_size{visible} = $view_size->[0]->[0];
    $view_size{approve} = $view_size->[0]->[1];
    $view_size{delete} = $view_size->[0]->[2];
    return \%view_size;
}

# retrieves meta-data for the specified list of messages from the db
sub get_msg {
    my %opts = @_;
    my $message_ids = $opts{message_ids};
    my $index = $opts{store};
    my $address = $opts{address};
    my $logger = $opts{logger};
    my $this_sub = "get_msg_data()";
    unless ($address and $index and $message_ids) {
        my $error_msg = "$this_sub: invalid parameters; unable to proceed";
        _log($logger, "ERROR", $error_msg);
        die($error_msg);
    }
    my $dbh = $index->dbi_handle();
    my $id_pseudo_params = substr(',?' x @$message_ids, 1);
    my $sql = <<"__SQL__";
select message.m_global_id as qid, local||'\@'||domain as m_enduser, 
       m_reason.name as c_reason_csv, h_from,  h_to,
       m_date::timestamp as m_date, c_subject_utf8 , m_spam_probability
from   message 
join   message_recipient on message_id = message.id
join   recipient on recipient.id = recipient_id
join   m_reason on m_reason.reason_id = message.reason_id
where  ((local = ?::text or local ~ ?::text) and domain = ?::text)
and    message_recipient.is_mapped = 1
and    message.m_global_id in ($id_pseudo_params)
__SQL__

    my ($local, $domain) = @$address;
    # Transform an array of arrays into an array of hashes.
    my @data;
    eval {
        @data =  map {
            my $row = $_;
            my %h = map {$COLS[$_] => $row->[$_]} 0..$#COLS;
            \%h;
        } @{$dbh->selectall_arrayref($sql, undef, $local, "^$local\\+", $domain, 
                                     @$message_ids)};
    };
    if ($@) {
        my $error_msg = "$this_sub: unable to retrieve meta data for " .
                        "message(s): " . join(', ', @$message_ids) . ": $@";
        _log($logger, "ERROR", $error_msg);
        die($error_msg);
    }
    return @data;
}

# retrieves a page of message data for the specified user
sub get_user_msgs {
    my %opts = @_;
    my $index = $opts{store};
    my $address = $opts{address};
    my $logger = $opts{logger};
    my $view = $opts{view};
    my $start = $opts{start};
    my $end = $opts{end};
    my $sort_expr = $opts{'sort'};
    my $desc = $opts{desc};
    my $reasons = $opts{reasons};

    my $this_sub = "get_user_msgs()";

    unless ($address and $index) {
        my $error_msg = "$this_sub: invalid parameters; unable to proceed";
        _log($logger, "ERROR", $error_msg);
        die($error_msg);
    }

    $sort_expr = 'qid' unless $sort_expr;
    $sort_expr = 'sortable_mail_address(lower(h_from))' if $sort_expr =~ /h_from/;
    $sort_expr = 'lower(c_subject_utf8)' if $sort_expr eq 'c_subject_utf8';

    my $offset  = $start || 0;
    $end = $offset + 999 unless defined $end;
    my $limit   = $end-$offset+1;
    if ($limit <= 0) {
        my $error_message = "$this_sub: invalid start/end range: $start/$end";
        _log($logger, "ERROR", $error_message);
        die($error_message);
    }

    my $sort_order = $desc ? 'DESC' : '';

    # "fast" means PG can sort and limit on message_recipient alone.
    my ($fast_sort, $slow_sort) = ('','');
    if ($sort_expr eq 'qid') {
        $sort_expr = 'message_recipient.message_id';
        $fast_sort = "ORDER BY $sort_expr $sort_order OFFSET $offset LIMIT $limit";
    }
    else {
        $slow_sort = "ORDER BY $sort_expr $sort_order OFFSET $offset LIMIT $limit";
    }

    my $action_predicate;
    if ((defined $view) and ($view eq 'delete')) {
        $action_predicate = '(message_recipient.action = 3)';
    }
    else {
        $action_predicate = '(message_recipient.action is null)';
    }

    my $dbh = $index->dbi_handle();
    my $reasons_predicate = _get_reasons_predicate($reasons);

    my $sql = <<"__SQL__";
select  message.m_global_id as qid, (local || '\@' || domain) as m_enduser,
        m_reason.name as c_reason_csv, h_from, h_to,
        (m_date::timestamp) as m_date,
        c_subject_utf8, m_spam_probability
from    message
join    m_reason using (reason_id)
join    (
    select  message_id, local, domain
        from    message_recipient
        join    recipient on recipient.id = recipient_id
        where   is_mapped = 1 and $action_predicate
        and     ((local = ?::text or local ~ ?::text) and domain = ?::text)
        and     $reasons_predicate
        $fast_sort
) as x on x.message_id = message.id
$slow_sort
__SQL__

    #print "$sql\n";

    my @result;

    my ($local, $domain) = @$address;
    # transforms an array of arrays into an array of hashes
    eval {
        @result = map {
            my $row = $_;
            $_ = {map {$COLS[$_] => $row->[$_]} 0..$#COLS}
        } @{$dbh->selectall_arrayref($sql, undef, $local, "^$local\\+", $domain)};
    };

    if ($@) {
        my $error_msg = "$this_sub: unable to retrieve messages for " .
                        "user " . ${$address}[0] . "\@" . ${$address}[1] .
                        ": " . $@;
        _log($logger, "ERROR", $error_msg);
        die($error_msg);
    }

    return @result;
}

sub _log {
    my $logger = shift;
    if ($logger) {
        $logger->log(@_);
    }
    else {
        print STDERR join ' ', @_, '\n';
    }
}

sub _get_reasons_predicate {
    my $reasons = shift;

    my $predicate;
    if ($reasons) {
        if (scalar @$reasons) {
            $predicate = "(message_recipient.reason_id in (" . 
                         "select reason_id from m_reason where name in ('" .
                         join("', '", @$reasons) .
                         "')))";
        }
        else {
            # empty reason list => never match
            $predicate = "(1 = 0)";
        }
    }
    else {
        # no reason list => always match
        $predicate = "(0 = 0)";
    }

    return $predicate;
}

return 1;
