package PureMessage::MessageStore::pmdb::UserView::all;
use strict;
use warnings;
use Carp;
use PureMessage::MessageStore::UserView::Row;

# The 'all' view acts as a base class for the 'visible', 'delete' and 'approve'
# views.

# SQL for retrieval of database ids (surrogate keys) of the messages
# of interest.
my $base_get_ids_sql = <<'EOT';
select message_recipient.message_id
from message_recipient
join recipient on message_recipient.recipient_id = recipient.id
join m_reason on message_recipient.reason_id = m_reason.reason_id
where recipient.local = ? and recipient.domain = ?
and message_recipient.is_mapped > 0
EOT

# SQL for sorting the retrieved list of ids.
my $base_sort_ids_sql = <<'EOT';
select message.id
from message
join message_recipient on message.id = message_recipient.message_id
join recipient on message_recipient.recipient_id = recipient.id
join m_reason on message.reason_id = m_reason.reason_id
EOT

# SQL for retrieving actual message data
my $base_get_msg_sql = <<'EOT';
select message.m_global_id as id, 'NA'::text as m_enduser,
       m_reason.name as m_reason, message.h_from, message.h_to,
       message.m_date, message.c_subject_utf8, message.m_spam_probability
from message
join m_reason on message.reason_id = m_reason.reason_id
EOT

# maps sort fields as passed by the caller into fully qualified db columns
my %valid_sort_column = (
    "c_reason_csv" => "m_reason.reason_id",
    "h_from"       => "message.h_from",
    "h_to"         => "message.h_to",
    "h_subject"    => "message.c_subject_utf8",
);

# View constructor retrieves the database ids into an array and stores
# the array into $self. Specified sorting criteria will _not_ be applied
# during this step, since we don't know how many rows we're dealing with yet.
sub new {
    my $p = shift;
    my $class = ref($p) || $p;

    my %data    = @_;
    my $state    = delete $data{state};
    my $addr    = delete $data{address};
    my $name    = delete $data{name};
    my $store   = delete $data{store};
    my $reasons = delete $data{reasons};
    my $base_pkg = delete $data{base_package};
    my $limit = delete $data{limit} || 0; # 0 is unlimited
    my $offset = delete $data{offset} || 0; # 0 is first row
    my $sort_cols = delete $data{sort_cols} || [];
    
    if ($^W) {
	warn "Invalid option $_ for the view $name."
	    foreach keys %data;
    }

    warn "Invalid option 'store' for the view $name." unless $store;
    
    my $self = bless {
	store    => $store,
	name     => $name,
	address  => $addr,
    }, $class;

    # if state has been provided, we have everything we need...
    return $self->_restore($state) if $state;

    $self->{limit} = $limit;
    $self->{offset} = $offset;

    # By default, sorting according to the specified sorting criteria
    # is required. Caller may decide to change this if the row count
    # is prohibitive.
    $self->{do_sort} = 1;

    # By default, ids will be retrieved in the reverse order of indexing
    # (i.e. insertion into the db), which is close enough to the reverse
    # order of quarantining. The actual sorting criteria, which may or
    # may not be applied later are only evaluated and stored to $self;
    my $default_order_by = 'message_recipient.message_id desc';
    $self->{current_order_by} = $default_order_by;
    $self->{default_order_by} = $default_order_by;

    my $order_by = "";
    $order_by = $self->_get_order_by_clause() || $default_order_by;

    $self->{order_by} = $order_by;

    my $action_filter = "";
    $action_filter = "AND " . $self->_eu_filter() if $self->_eu_filter();

    my $reason_filter = "";
    # reason is always lowercase in the db
    @$reasons = grep($_ = lc($_), @$reasons);
    $reason_filter = 'AND m_reason.name in (' . join(',', map {"'$_'"} @$reasons) . ')' if @$reasons;

    my $sql = "$base_get_ids_sql $action_filter $reason_filter " .
              "order by $default_order_by";
    print STDERR "SELECT (all): $sql \n" if $ENV{PMX_pmdb_SQL_DEBUG};
    print STDERR "reasons: @$reasons\n" if $ENV{PMX_pmdb_SQL_DEBUG};
    
    my $dbh = $self->{store}->dbi_handle;
    my $sth = $dbh->prepare($sql);
    my ($local, $domain) = split(/\@/, $addr, 2);
    $domain = "" unless defined $domain;

    # execute the statement and retrieve all the ids into an array that
    # we'll store into $self
    $sth->execute($local, $domain);
    my ($id, @id);
    $sth->bind_col(1, \$id);
    push @id, $id while $sth->fetch();
    $self->{ids} = \@id;
    return $self;
}

# returns SQL order by clause based on the input list of sort columns
sub _get_order_by_clause {
    my $self = shift;
    my $sort_cols = shift;
    my $order_by = "";
    foreach my $sort_col (@$sort_cols) {
        my $order = (($sort_col =~ s/^-//) ? 'desc' : 'asc');
        if (exists $valid_sort_column{$sort_col}) {
            $order_by .= "$valid_sort_column{$sort_col} $order,";
        }
    }
    chop $order_by if ($order_by);
    return $order_by;
}

sub _eu_filter { undef }

sub name { shift->{name} }

# the count of rows that this view may return in unconstrained mode
# (i.e. if the limit/offset are set to 0)
sub row_count {
    my $self = shift;
    return scalar @{$self->{ids}};
}

# caller can use this to avoid sorting of large datasets
sub set_sort_needed {
    my $self = shift;
    my $sort_needed = shift || 0; 
    $self->{do_sort} = $sort_needed;
}

# modifies the sorting criteria for the existing view
sub set_sort_criteria {
    my $self = shift;
    my $sort_cols = shift || [];

    # we'll set the desired sort order as caller wanted here; we'll
    # reset the current statement and offset _only_ if this desired sort
    # order will indeed take effect, however.
    my $order_by = $self->_get_order_by_clause($sort_cols);
    if ($order_by) {
        $self->{order_by} = $order_by;
        if ($self->{do_sort}) {
            $self->{offset} = 0;
            if (exists $self->{sth}) {
                $self->{sth}->finish();
                delete $self->{sth};
            }
        }
    }
}

# sets the limit and resets the data retrieval handle; does nothing if
# invalid limit specified
sub set_limit {
    my $self = shift;
    my $limit = shift || 0;
    my $old_limit = $self->{limit};
    if ($limit > 0) {
        $self->{limit} = $limit;
        if (($old_limit != $limit) and exists $self->{sth}) {
            $self->{sth}->finish();
            delete $self->{sth};
        }
    }
}

# sets the offset and resets the data retrieval handle; does nothing if
# invalid offset specified
sub set_offset {
    my $self = shift;
    my $offset = shift || 0;
    my $old_offset = $self->{offset};
    if (($offset >= 0) and ($offset < $self->row_count())) {
        $self->{offset} = $offset;
        if ($old_offset != $offset) {
            $self->{sth}->finish();
            delete $self->{sth};
        }
    }
}

# returns the internal state of the view
sub get_state {
    my $self = shift;
    my $state = {};
    $state->{limit} = $self->{limit};
    $state->{offset} = $self->{offset};
    $state->{do_sort} = $self->{do_sort};
    $state->{current_order_by} = $self->{current_order_by};
    $state->{default_order_by} = $self->{default_order_by};
    $state->{order_by} = $self->{order_by};
    $state->{ids} = $self->{ids};
    return $state;
}

# sets the internal state of the view based on the supplied state
sub _restore {
    my $self = shift;
    my $state = shift;
    $self->{limit} = $state->{limit};
    $self->{offset} = $state->{offset};
    $self->{do_sort} = $state->{do_sort};
    $self->{current_order_by} = $state->{current_order_by};
    $self->{default_order_by} = $state->{default_order_by};
    $self->{order_by} = $state->{order_by};
    $self->{ids} = $state->{ids};
    return $self;
}

# returns the next row of message data
sub fetch {
    my $self = shift;
    $self->_execute_query() unless ($self->{sth});
    my $sth = $self->{sth};
    my @row = $sth->fetchrow_array();
    return undef unless @row;

    my @fields = @{$sth->{NAME_lc}};
    shift @fields; # don't want to keep the id field
    return PureMessage::MessageStore::UserView::Row->new(
	fields => \@fields,
	id     => shift @row, # shift off the id
	data   => \@row,
    );
}

# composes and executes the "get message data" query
sub _execute_query {
    my $self = shift;
    if (($self->{do_sort}) and
        ($self->{order_by} ne $self->{current_order_by})) {
        $self->_sort_ids();
    }

    my $ids_ref = $self->{ids};
    my $limit = $self->{limit} || (scalar @$ids_ref);
    my $start = $self->{offset};
    my $end = $start + $limit - 1;
    $end = scalar @$ids_ref if ($end >= (scalar @$ids_ref));
    my $where_condition = ($start <= $end) ?
        "where message.id in (" .
            join (',', map {$_ . "::bigint"} @$ids_ref[$start .. $end]) . ")" :
        "where 0=1";

    # we have to modify the default "order by" slightly in this context
    my $order_by = ($self->{current_order_by} eq $self->{default_order_by}) ?
                   'message.id desc' : $self->{current_order_by};

    my $sql = "$base_get_msg_sql $where_condition order by $order_by";
    print STDERR "_execute_query: $sql \n" if $ENV{PMX_pmdb_SQL_DEBUG};

    my $dbh = $self->{store}->dbi_handle;
    my $sth = $dbh->prepare($sql);
    $sth->execute();
    $self->{sth} = $sth;
}

# sorts the message database ids according to the 
sub _sort_ids {
    my $self = shift;
    my $ids_ref = $self->{ids};

    # anything to do?
    return unless scalar @$ids_ref;

    my $where_condition = "where message.id in (" .
                          join (',', map {$_ . "::bigint"} @$ids_ref) . ")";
    my $sql = "$base_sort_ids_sql $where_condition order by $self->{order_by}";
    print STDERR "_sort_ids: $sql \n" if $ENV{PMX_pmdb_SQL_DEBUG};
    my $dbh = $self->{store}->dbi_handle;
    my $sth = $dbh->prepare($sql);
    $sth->execute();
    my ($id, @id);
    $sth->bind_col(1, \$id);
    my %fetched;
    while ($sth->fetch()) {
        push @id, $id unless exists $fetched{$id};
        $fetched{$id} = 1;
    }
    $self->{ids} = \@id;
    $self->{current_order_by} = $self->{order_by};
}

DESTROY {
    my $self = shift;
    delete $self->{sth} if $self->{sth};
}

1;
