package PureMessage::Preferences::db;
use strict;
use warnings;

use Carp;
use HTTP::Date qw(str2time time2iso);
use PureMessage::Config qw(%config);
use PureMessage::Resources::DBH qw(DBH);
use PureMessage::Preferences;
use base qw(PureMessage::Preferences);

sub TARGET() { '' }

sub new {       # (resource_name, instance_id)
    my $p = shift;
    my $class = ref($p) || $p;
    my %data = @_;
    $data{dbh} = DBH();
    croak "$class must be initialized with an instance_id"
        unless($data{instance_id});
    return bless \%data, $class;
}

my $SQL_GET_EU_LANGUAGE = <<EOT;
    select  k_value as language
    from    cfg_kv
    where   k_name = 'lang'
    and     instance_id = ?
    and     userkey = ?
EOT

my $SQL_SET_EU_LANGUAGE = <<EOT;
    update  cfg_kv set k_value = ?
    where   k_name = 'lang'
    and     instance_id = ?
    and     userkey = ?
EOT

my $SQL_ADD_EU_LANGUAGE = <<EOT;
    insert  into cfg_kv (instance_id, userkey, k_name, k_value)
    values  (?, ?, 'lang', ?)
EOT

sub lang {
    my ($self, $uid, $lang) = @_;

    if ($lang) {        ## Update database
        my $rows;
        croak "SET_EU_LANGUAGE query failed" unless $rows = $self->{dbh}->do(
            $SQL_SET_EU_LANGUAGE, undef, $lang, $self->{instance_id}, $uid
        );
        
        if(substr($rows, 0, 1) eq '0') {
            croak "ADD_EU_LANGUAGE query failed" unless $self->{dbh}->do(
                $SQL_ADD_EU_LANGUAGE,undef,$self->{instance_id}, $uid, $lang
            );
        }            
    }
    else {              ## Retrieve from DB.
        my $sth = $self->{dbh}->prepare($SQL_GET_EU_LANGUAGE);
        croak "GET_EU_LANGUAGE query failed" 
        unless $sth->execute($self->{instance_id}, $uid);
        ($lang) = $sth->fetchrow_array;
    }

    return $lang || $config{language};
}

my $SQL_GET_EU_EXPIRY = <<EOT;
    select  k_value as quarantine_expiry_days
    from    cfg_kv
    where   k_name = 'expire_delay'
    and     instance_id = ?
    and     userkey = ?
EOT

my $SQL_SET_EU_EXPIRY = <<EOT;
    update  cfg_kv set k_value = ?
    where   k_name = 'expire_delay'
    and     instance_id = ?
    and     userkey = ?
EOT

my $SQL_ADD_EU_EXPIRY = <<EOT;
    insert  into cfg_kv (instance_id, userkey, k_name, k_value)
    values  (?, ?, 'expire_delay', ?)
EOT

sub expire_delay {
    my ($self, $user, $tm) = @_;
    croak "Missing user id" unless $user;
    if ($tm) {        ## Update database
        my $rows;
        croak "SET_EU_EXPIRY query failed" unless $rows = $self->{dbh}->do(
            $SQL_SET_EU_EXPIRY, undef, "$tm", $self->{instance_id}, $user
        );
        
        if(substr($rows, 0, 1) eq '0') {
            croak "ADD_EU_EXPIRY query failed" unless $self->{dbh}->do(
                $SQL_ADD_EU_EXPIRY, undef, $self->{instance_id}, $user, "$tm"
            );
        }
    }
    else {              ## Retrieve from DB.
        my $sth = $self->{dbh}->prepare($SQL_GET_EU_EXPIRY);
        croak "GET_EU_EXPIRY query failed" 
        unless $sth->execute($self->{instance_id}, $user);
        ($tm) = $sth->fetchrow_array || '';
    }

    return $tm;
}

# (expire_delay_iso is really common code for all Preferences implementations)

sub expire_delay_iso {
    my ($self, $user, $isotm) = @_;
    croak "Missing user id" unless $user;
    if ($isotm) {
        $self->expire_delay($user, str2time($isotm));
    }
    else {
        my $delay = $self->expire_delay($user);
        $isotm = time2iso($delay) if $delay;
    }
    return $isotm || '';
}

my $SQL_GET_EXPIRE_MAX_DELAY = <<EOT;
    select  max(kv_value) as quarantine_expiry_days
    from    cfg_kv 
    where   k_name = 'expire_delay'
    and     instance_id = ?
    and     userkey in /*(?,...,?) appended by code below*/
EOT

sub expire_max_delay {
    my ($self,@users) = $@;

    my $sqlcmd = $SQL_GET_EXPIRE_MAX_DELAY . '(' . substr(',?' x scalar(@users),1) . ')';
    print "expire_max_delay:\n\t$sqlcmd\n\tusers: @users\n" if $ENV{PMX_pmdb_SQL_DEBUG};

    my $sth = $self->{dbh}->prepare($sqlcmd);

    croak "EXPIRE_MAX_DELAY query failed" 
    unless $sth->execute($self->{instance_id}, @users);

    my($max)  = $sth->fetchrow_array;
    return $max;
}

sub expire_max_delay_iso {
    my $self = shift;
    if (my $tm = $self->expire_max_delay(@_)) {
	return time2iso($tm);
    }
    return '';
}

# save's happen already
sub save { 1; }

sub dump {
    my ($self, $cb) = @_;

    my $sth = $self->{dbh}->prepare_cached(
        "SELECT * FROM cfg_kv WHERE instance_id = ? ORDER BY userkey"
    );
    
    # fun, fun, fun... Preferences callbacks expect to be called once
    # per user, not once per row, so we have to accumulate data.
    if($sth->execute($self->{instance_id})) {
        my $luser = undef;
        my %uinfo;
        while(my $row = $sth->fetchrow_hashref) {
            $luser = $row->{userkey} unless defined($luser);
            if(%uinfo && ($luser ne $row->{userkey})) {
                $cb->($luser, \%uinfo);
                %uinfo = ();
                $luser = $row->{userkey};
            }
            
            $uinfo{$row->{k_name}} = $row->{k_value};
        }
        if($luser and %uinfo) {
            # we have to catch the last user
            $cb->($luser,\%uinfo);
        }
    } else {
        die "database enumeration of preferences failed to execute: ",
            $sth->errstr;
    }
}

1;
