package PureMessage::Preferences::hdb;
use strict;
use warnings;

use Carp;
use HTTP::Date qw(str2time time2iso);
use PureMessage::HashDB;
use PureMessage::Config qw(expand_filename %config);
use PureMessage::MessageLog qw(decode_value join_line);
use File::Basename;
use File::Path;
use PureMessage::Preferences;
use base qw(PureMessage::Preferences);

sub TARGET() { 'var/userpref/preferences' }

sub new {
    my $p = shift;
    my $class = ref($p) || $p;
    my %data = @_;

    mkpath(dirname(expand_filename(TARGET)));
    $data{db} = PureMessage::HashDB->new(
	file => expand_filename(TARGET),
	%data);
    
    return bless \%data, $class;
}

sub lang {
    my ($self, $uid, $lang) = @_;
    my $ur = $self->_get_user_record($uid);
    if ($lang) {
	$ur->{lang} = $lang;
	$self->_put_user_record($uid, $ur);
    }
    return $ur->{lang} || $config{language};
}

sub expire_delay {
    my ($self, $user, $tm) = @_;
    croak "Missing user id" unless $user;
    
    my $ur = $self->_get_user_record($user);
    if ($tm) {
	$ur->{expire_delay} = $tm;
	$self->_put_user_record($user, $ur);
    }
    return $ur->{expire_delay} if $ur->{expire_delay};
    return '';
}

sub expire_delay_iso {
    my ($self, $user, $isotm) = @_;
    croak "Missing user id" unless $user;
    
    my $ur = $self->_get_user_record($user);
    if ($isotm) {
	my $tm = str2time($isotm);
	$ur->{expire_delay} = $tm;
	$self->_put_user_record($user, $ur);
    }
    return time2iso($ur->{expire_delay}) if $ur->{expire_delay};
    return '';
}

sub expire_max_delay {
    my $self = shift;
    my $max = 0;
    foreach my $user (@_) {
	my $ur = $self->_get_user_record($user);
	next unless $ur->{expire_delay};
	$max = $ur->{expire_delay} if 
	    $ur->{expire_delay} > $max;
    }
    return $max;
}

sub expire_max_delay_iso {
    my $self = shift;
    if (my $tm = $self->expire_max_delay(@_)) {
	return time2iso($tm);
    }
    return '';
}

sub save {
    my $self = shift;
    $self->{db}->save;
    delete($self->{db});
}

sub reload { shift->{db}->reload }

sub dump {
    my ($self, $cb) = @_;
    $self->{db}->enumerate(sub {
	my ($user, $line) = @_;
	&$cb($user, _parse_user_record($line));
    });
}

# User records are stored as a list of key/value pairs
# encoded using the MessageLog's Q encoding/decoding
sub _get_user_record {
    my ($self, $uid) = @_;
    my $val;
    my %user_record;
    return _parse_user_record($self->{db}->get($uid));
    return undef;
}

sub _parse_user_record {
    my $line = shift;
    return undef unless $line;

    my %user_record;
    my ($ts, @kv) = split(" ", $line);
    for (@kv) {
	my ($k, $v) = split("=", $_, 2);
	$v = decode_value($v);
	$user_record{$k} = $v;
    }
    $user_record{__last_mod} = $ts;
    return \%user_record;
}

sub _put_user_record {
    my ($self, $uid, $user_record) = @_;
    my $ts = delete $$user_record{__last_mod};
    my $val = join_line(time, %$user_record);
    $self->{db}->put($uid, $val);
    $$user_record{__last_mod} = $ts;
}

1;
